/*
*  dbs.h 
*
*  Copyright (C) 2012 Richard Christman 
*
*  This file is part of QuickSilver AAM.
*
*  QuickSilver AAM is free software; you can redistribute it and/or modify 
*  it under the terms of the GNU General Public License as published by the 
*  Free Software Foundation; either version 3 of the License, or (at your 
*  option) any later version.
* 
*  This program is distributed in the hope that it will be useful, but 
*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
*  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
*  for more details.
* 
*  You should have received a copy of the GNU General Public License along 
*  with this program; if not, see <http://www.gnu.org/licenses>.
* 
*  Additional permission under GNU GPL version 3 section 7
* 
*  If you modify QuickSilver AAM, or any covered work, by linking or 
*  combining it with CodeGear VCL (or a modified version of that library), 
*  containing parts covered by the terms of CodeGear's license, the licensors 
*  of QuickSilver AAM grant you additional permission to convey the 
*  resulting work.
*
*/

//---------------------------------------------------------------------------
#ifndef dbsH
#define dbsH
//---------------------------------------------------------------------------
#include <Classes.hpp>
#include <Controls.hpp>
#include <StdCtrls.hpp>
#include <ComCtrls.hpp>
class TDB; class TDBmanager; class TDBindex; class TDBfreeList; class TDBbuffer;
//---------------------------------------------------------------------------
//                         Q   S   2   5   1   7  12  25  216
#define DBFILEHEADER    "\x51\x53\x02\x05\x01\x07\x0C\x19\xD8"
#define DBDATE          "2007-12-25"
#define DBVERSION       "2.5.1"

TDB* __fastcall OpenDB(const AnsiString &FileName, bool InMemory = true);
TDB* __fastcall OpenDB(const AnsiString &FileName, TStream *Stream);
bool __fastcall CloseDB(TDB *DB);

class TDB : public TObject
{
protected:
   bool delstream;
   TStream *fstream;
   AnsiString ffilename;
   TDBindex *index[16];
   AnsiString __fastcall NextStep(int level, const AnsiString &path);

public:
   __property TStream *stream     = { read=GetStream };
   __property AnsiString filename = { read=GetFileName, write=SetFileName };

   bool modified, working, inmemory;
   AnsiString ziperrmsg, exceptmsg;
   int error, ziperror;
   TDBmanager *dbman;

   __fastcall virtual TDB(const AnsiString &FileName, bool InMemory);
   __fastcall virtual TDB(const AnsiString &FileName, TStream *Stream);
   __fastcall virtual~TDB();
   bool __fastcall virtual OpenDatabase();
   bool __fastcall virtual CloseDatabase();

   bool __fastcall GetText(const AnsiString &path, TStrings *text);
   bool __fastcall PutText(const AnsiString &path, TStrings *text);
   bool __fastcall GetBuffer(const AnsiString &path, TDBbuffer *buf);
   bool __fastcall PutBuffer(const AnsiString &path, void *buf, int length, bool SaveIndex = true);
   bool __fastcall PutBuffer(const AnsiString &path, TDBbuffer *buf, bool SaveIndex = true);
   //bool __fastcall GetStream(const AnsiString &path, TMemoryStream *stream);
   //bool __fastcall PutStream(const AnsiString &path, TMemoryStream *stream);
   bool __fastcall RenItem(const AnsiString &path, const AnsiString &newpath);
   bool __fastcall DelItem(const AnsiString &path);
   bool __fastcall DataExists(const AnsiString &path);
   bool __fastcall SwapKey(const AnsiString &path, const AnsiString &newkey);
   bool __fastcall FlagKey(const AnsiString &path, int flagsbyte, byte flag);
   bool __fastcall UnFlagKey(const AnsiString &path, int flagsbyte, byte flag);
   bool __fastcall KeyIsFlagged(const AnsiString &path, int flagsbyte, byte flag);
   // for renaming or moving without moving the data
   bool __fastcall MoveData(const AnsiString &Path, const AnsiString &NewPath);
   bool __fastcall GetDataOffset(const AnsiString &path, int &offset);
   bool __fastcall PutDataOffset(const AnsiString &path, int offset);
   bool __fastcall DelDataKey(const AnsiString &path);
   // memorystream only
   bool virtual __fastcall SaveToFile(const AnsiString &FileName = "");
   // other
   void __fastcall ShowFreeList(const AnsiString &caption = "Free List");
   void __fastcall GetMainIndex(TStrings *lines);
   TDBindex* __fastcall GetIndex(int level);
   void __fastcall SetFileName(const AnsiString &FileName);
   AnsiString __fastcall GetFileName();
   TStream* __fastcall GetStream();
   void __fastcall ShowDBError(int error);
};

class TDBmanager : public TObject
{
private:
   TDB *db;
   TStream *stream;
   TDBfreeList *freeList;	  // Free list block in memory.
   int startOfData;	         // Offset to first data in file.
   int endOfData;		// Offset to last data in file.
   bool delStream;
   bool locked;
public:
   int error;
   AnsiString filename;
   __fastcall virtual TDBmanager(TDB *DB, const AnsiString &Filename);
   __fastcall virtual TDBmanager(TDB *DB, TStream *Stream, const AnsiString &Filename);
   __fastcall ~TDBmanager();
   void __fastcall Init(); // does all construction except stream
   int  __fastcall start() const { return startOfData; }
   int  __fastcall endData() const  { return endOfData; }
   int  __fastcall Allocate(int);
   int  __fastcall AllocateInFreeList(int);
   int  __fastcall AllocateAtEnd(int spaceRequired);
   void __fastcall Deallocate(int);
   bool __fastcall SeekTo(int ofs);
   bool __fastcall WriteDat(void *dat, int size);
   bool __fastcall ReadDat(void *dat, int size);
   bool __fastcall CheckDBFileHeader();
   bool __fastcall WriteHeader();
   bool __fastcall WriteStartAndEnd();
   bool __fastcall ReadStartAndEnd();
   bool __fastcall WriteFreeList();
   bool __fastcall ReadFreeList();
   bool __fastcall AddToFreeList(int, int);
   bool __fastcall Decompress(TDBbuffer *buf);
   bool __fastcall Compress(TDBbuffer *buf);
   void __fastcall ShowFreeList(const AnsiString &caption = "debug");
};

class TDBindex : public TStringList
{
public:
   AnsiString name;
   TDBindex *parent;
   TDBmanager *dbm;
   int offset, block;
   __fastcall virtual TDBindex(TDBmanager *Fm, AnsiString Name="", TDBindex *Parent=NULL);
   __fastcall virtual~TDBindex();
   int  __fastcall Save();
   void __fastcall ShowIndex();
   int  __fastcall GetType(const AnsiString &key);
   int  __fastcall GetSize(const AnsiString &key);
   int  __fastcall GetOffset(const AnsiString &key);
   int  __fastcall SetOffset(const AnsiString &key, int Offset); // <-not to disk
   int  __fastcall PutOffset(const AnsiString &key, int Offset); // <-to disk
   void __fastcall SwapKey(const AnsiString &key, const AnsiString &newkey);
   void __fastcall FlagByte(AnsiString path, int flagsbyte, byte flag);
   void __fastcall UnFlagByte(AnsiString path, int flagsbyte, byte flag);
   bool __fastcall ByteIsFlagged(AnsiString path, int flagsbyte, byte flag);
   void __fastcall DelKey(const AnsiString &key); // but not data

   int  __fastcall GetBinary(const AnsiString &key, TDBbuffer* buf);
   // functions below just write data and update index in memory, but don't
   // save the index, or in the case of writeindex, don't cascade to parent..
   int  __fastcall PutBinary(const AnsiString &key, TDBbuffer *buf, bool SaveIndex=true);
   int  __fastcall WriteIndex();
   int  __fastcall GetFlags(const AnsiString &key);
   bool __fastcall IsIndexKey(int pos); //const AnsiString &key);
};

class TDBfreeList : public TObject {
private:
   TList *offset, *space;
public:
   int lastOffset, lastSpace;  //  assigned but not used ???
   __fastcall TDBfreeList();
   __fastcall virtual ~TDBfreeList();
   bool __fastcall AddFreeSpace(int, int);
   int  __fastcall Allocate(int);
   void __fastcall AttemptMerge(int);
   bool __fastcall Coalesce(int, int);
   bool __fastcall EnterFreeSpace(int, int);
   int  __fastcall getOffset(int i) { return (int)offset->Items[i]; }
   int  __fastcall getSpace(int i) { return (int)space->Items[i]; }
   bool __fastcall LoadFromStream(TStream *);
   bool __fastcall SaveToStream(TStream *);
   int  __fastcall Count();
};

class TDBbuffer : public TObject
{
public:
   char* data;
   int flags, length;
   __fastcall virtual TDBbuffer();
   __fastcall virtual TDBbuffer(const AnsiString &Data);
   __fastcall virtual TDBbuffer(char *Data, int Length = 0);
   __fastcall virtual TDBbuffer(void *Data, int Length);
   __fastcall virtual~TDBbuffer();
   void __fastcall SetData(char *Data, int Length);
   void __fastcall SetData(const AnsiString &Data);
   void __fastcall SetLength(int Length);
   void __fastcall SetFlag(int flag);
   bool __fastcall Flagged(int flag);
   void __fastcall ToggleFlag(int flag);
   void __fastcall ClearFlag(int flag);
   void __fastcall ClearFlags();
   void __fastcall ClearData();
   void __fastcall Clear();
   AnsiString __fastcall str();
   char* __fastcall c_str();
};

// AnsiString DBPS and DBFS allows more freedom when concatenating path or key
const AnsiString DBPS       = ''; // \255 path separator -- must be 1 char only
const       char DBPSc      = '';
const AnsiString DBFS       = ''; // \135 field separator -- must be 1 char only
const       char DBFSc      = '';
const AnsiString DBOS       = ''; // \254 Other key separator (utility)-- must be 1 char only
const       char DBOSc      = '';

const int DBFLAGSIZE        = 2;

// TDBbuffer flags
const int fNoflags          = 0;
const int fIndex            = 1;
const int fCompressed       = 2;
const int fUndefined3       = 4;
const int fUndefined4       = 8;
const int fUndefined5       = 16;
const int fUndefined6       = 32;
const int fUndefined7       = 64;
const int fUndefined8       = 128;

// DB item flags
const WORD mflUnused1       = 1;
const WORD mflUnused2       = 2;
const WORD mflUnused3       = 4;
const WORD mflUnused4       = 8;
const WORD mflUnused5       = 16;
const WORD mflUnused6       = 32;
const WORD mflUnused7       = 64;
const WORD mflCantuse1	    = 128;
const WORD mflUnused8       = 256;
const WORD mflUnused9       = 512;
const WORD mflUnused10      = 1024;
const WORD mflUnused11      = 2048;
const WORD mflUnused12      = 4096;
const WORD mflUnused13      = 8192;
const WORD mflUnused14      = 16384;
const WORD mflCantuse2      = 32768;
// start flags with this--1000000010000000 so there is no '\0' byte
const WORD mflwFlags        = 32896;

const byte bflags1 = 16;
const byte bflags2 = 17;
const byte wflags  = 16;

// data sizes
const int DBNIL             = -1;
const int INTSIZE           = sizeof(int);
const int PTRSIZE           = sizeof(void*);
const int OFSSIZE           = sizeof(int);
const int HDRSIZE           = sizeof(int);
const int FLGSIZE           = sizeof(int);
const int LENSIZE           = sizeof(int);

// database errors
const int DB_OK             = 0;
const int DB_WRITEERR       = 1;
const int DB_WRITEEXC       = 2;
const int DB_READERR        = 3;
const int DB_READEXC        = 4;
const int DB_SEEKERR        = 5;
const int DB_SEEKEXC        = 6;
const int DB_INITERR        = 7;
const int DB_FLWRITEERR     = 8;
const int DB_FLREADERR      = 9;
const int DB_INFLATEERR     = 10;
const int DB_DEFLATEERR     = 11;
const int DB_FILEHEADERERR  = 12;
const int DB_FAILEDINITERR  = 14;
const int DB_CANCELDECRYPT  = 15;
//---------------------------------------------------------------------------
//extern TDB *DB;
//---------------------------------------------------------------------------
#endif
