/*
*  gnupg.cpp 
*
*  Copyright (C) 2012 Richard Christman 
*
*  This file is part of QuickSilver AAM.
*
*  QuickSilver AAM is free software; you can redistribute it and/or modify 
*  it under the terms of the GNU General Public License as published by the 
*  Free Software Foundation; either version 3 of the License, or (at your 
*  option) any later version.
* 
*  This program is distributed in the hope that it will be useful, but 
*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
*  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
*  for more details.
* 
*  You should have received a copy of the GNU General Public License along 
*  with this program; if not, see <http://www.gnu.org/licenses>.
* 
*  Additional permission under GNU GPL version 3 section 7
* 
*  If you modify QuickSilver AAM, or any covered work, by linking or 
*  combining it with CodeGear VCL (or a modified version of that library), 
*  containing parts covered by the terms of CodeGear's license, the licensors 
*  of QuickSilver AAM grant you additional permission to convey the 
*  resulting work.
*
*/

//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include <sys\types.h>

#include "gnupg.h"
#include "global.h"
#include "utility.h"
#include "msgmans.h"
#include "pgpdpassdlgs.h"
#include "gpg4.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
// TGPGerrDlg *GPGerrDlg;
const int bufsize = 32768;
AnsiString gpgdatadir;
AnsiString gpgdir;
//---------------------------------------------------------------------------

AnsiString __fastcall GPGDir()
{
   //return GetShortName(AppDir()).LowerCase();
   return gpgdir;  // has already had GetShortName called
}
//---------------------------------------------------------------------------

AnsiString __fastcall GPGDataDir()
{
   //return GetShortName(AppDataDir()).LowerCase();
   return gpgdatadir; // has already had GetShortNameApplied
}
//---------------------------------------------------------------------------

AnsiString __fastcall GPGExe()
{
   return GPGDir() + "gpg.exe";
}
//---------------------------------------------------------------------------

AnsiString __fastcall GPGConf()
{
   return GPGDataDir() + "gpg.conf";
}
//---------------------------------------------------------------------------

AnsiString __fastcall GPGIdeaDll()
{
   return GPGDir() + "idea.dll";
}
//---------------------------------------------------------------------------

AnsiString __fastcall GPGIconvDll()
{
   return GPGDir() + "iconv.dll";
}
//---------------------------------------------------------------------------

/*AnsiString __fastcall GPGTempFile()
{
   static int c = 1;
   AnsiString filename = GPGTempFileBase();
   if (FileExists(filename + IntToStr(c)))
	  c++;
   return filename + IntToStr(c);
}*/
//---------------------------------------------------------------------------

AnsiString __fastcall GPGTempFile()
{
   return GPGDataDir() + "gpg.tmp";
}
//---------------------------------------------------------------------------

bool __fastcall GPGIsPGPMessage(TStrings *Message)
{
   int x;
   x = Message->IndexOf(BeginPGPMessage);
   if (x >= 0 && Message->IndexOf(EndPGPMessage) > x + 5)
	  return true;
   x = Message->IndexOf(BeginPGPSignedMessage);
   if (x >= 0 && Message->IndexOf(EndPGPSignature) > x + 5)
	  return true;
   return false;
}
//---------------------------------------------------------------------------

bool __fastcall GPGIsPGPSignedMessage(TStrings *Message)
{
   int x = Message->IndexOf(BeginPGPSignedMessage);
   return x >= 0 && Message->IndexOf(EndPGPSignature) > x;
}
//---------------------------------------------------------------------------

bool __fastcall GPGHasPGPSignatureVerification(TStrings *Message)
{
   int x = Message->IndexOf(PGPSignatureVerification);
   return x >= 0;
}
//---------------------------------------------------------------------------

AnsiString __fastcall GPGDefaultKeyID()
{
   AnsiString line, defkey;
   if (FileExists(GPGConf()))
   {
	  TStringList *cfg = new TStringList();
	  cfg->LoadFromFile(GPGConf());
	  for (int i = 0; i < cfg->Count; i++)
	  {
		 line = cfg->Strings[i].Trim();
		 if (line.LowerCase().Pos("default-key") == 1)
		 {
			defkey = line.SubString(12, 18);
			break;
		 }
	  }
   }
   return defkey.Trim();
}
//---------------------------------------------------------------------------

AnsiString __fastcall GPGTrustModel()
{
   AnsiString line, model;
   if (FileExists(GPGConf()))
   {
	  TStringList *cfg = new TStringList();
	  cfg->LoadFromFile(GPGConf());
	  for (int i = 0; i < cfg->Count; i++)
	  {
		 line = cfg->Strings[i].Trim();
		 if (line.LowerCase().Pos("trust-model") == 1)
		 {
			model = line.SubString(12, 18).Trim();
			break;
		 }
	  }
   }
   return model;
}

//---------------------------------------------------------------------------
// main decrypy/verify function
//---------------------------------------------------------------------------

/*bool __fastcall GPGProcessMessageDecryption(TForm *Owner, TStrings *Message)
{
   if (!checkgpgavailable())
	  return false;

   assert(Message->IndexOf(BeginPGPSignedMessage) >= 0 ||
		  Message->IndexOf(BeginPGPMessage) >= 0);

   if (Message->IndexOf(BeginPGPSignedMessage) >= 0)
   {
	  return GPGVerifyClearSigned2(Owner, Message);
   }

   bool tryall;
   int tries = 0, maxtries = 5;
   TStringList *keys = new TStringList();
   while (!keys->Count && tries < maxtries)
   {
	  if (!GPGGetEncryptionKeys2(Message, keys))
	  {
		 delete keys;
		 return false;
	  }
	  tries++;
   }
   if (!keys->Count)
   {
	  MsgBox(Owner,
		 "Problem! I've attempted to get the encryption keys    \n"
		 "from this message 5 times, with no luck! I don't know   \n"
		 "why. Sorry, the decryption failed.",
		  MT_ERROR, MB_OK);
	  delete keys;
	  return false;
   }

   AnsiString key;
   bool ok;
   tries = 0;
   maxtries = 3;
   while (true)
   {
	  ok = true;
	  //if (passphrase.IsEmpty())
		 passphrase = GetDecryptionPassphrase(Owner, keys, passphrase,tryall);
	  if (passphrase.IsEmpty())
	  {
		 delete keys;
		 return false;
	  }

	  if (!GPGDecryptMessage(Owner, Message, passphrase, tryall))
	  {
		 ok = false;
		 //passphrase = "";
		 tries++;
		 if (tries < maxtries)
			continue;
	  }
	  break;
   }
   //Owner->Perform(WM_FLASHDECRYPT, 0, 0);
   delete keys;

   return ok;
}*/
//---------------------------------------------------------------------------

/*bool __fastcall GPGDecryptMessage(TForm *Owner, TStrings *Message, AnsiString &Passphrase,
   bool TryAll, bool Quiet)
{
   assert(Message->IndexOf(BeginPGPMessage) >= 0);
   AnsiString cmdline = getcommandline(GPGdecrypt, GPGDecryptCmd);

  //((TForm*)Owner)->Perform(WM_FLASHDECRYPT, 1, (LPARAM)"decrypting message");
   TMemoryStream *input = new TMemoryStream();
   TMemoryStream *output = new TMemoryStream();
   TStringList *headers = new TStringList();
   getheaders(headers, Message);
   Message->SaveToStream(input);

   TGPGConsole2 *con = new TGPGConsole2(Owner);
   bool ok = con->Run(cmdline, Passphrase.c_str(), input, output);
   if (con->status->Text.Pos("DECRYPTION_OKAY"))
   {
	  FormatSigVerification(con->error);
	  if (con->error->Count)
	  {
		 headers->AddStrings(con->error);
		 headers->Add("*** BEGIN PGP DECRYPTED/VERIFIED MESSAGE ***");
		 headers->Add("");
	  }
	  else
	  {
		 headers->Add("*** BEGIN PGP DECRYPTED MESSAGE ***");
		 headers->Add("");
	  }
	  Message->LoadFromStream(output);
	  putheaders(Message, headers);
	  if (con->error->Count)
	  {
		 Message->Add("");
		 Message->Add("*** END PGP DECRYPTED/VERIFIED MESSAGE ***");
	  }
	  else
	  {
		 Message->Add("");
		 Message->Add("*** END PGP DECRYPTED MESSAGE ***");
	  }
	  ok = true;
   }
   else
   {
	  if (!Quiet)
		 ShowGPGError(Owner, con->error, con->status);
	  putheaders(Message, headers);
	  ok = false;
   }
   delete con;

   delete headers;
   delete output;
   delete input;

   //Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"Done!");
   Snooze(200);

   return ok;
}*/
//---------------------------------------------------------------------------

bool __fastcall GPGDecryptMessage(TForm *Owner, TStrings *Message, AnsiString &Passphrase,
   bool TryAll, bool Quiet)
{
   assert(Message->IndexOf(BeginPGPMessage) >= 0);
   AnsiString cmdline = getcommandline(GPGdecrypt, GPGDecryptCmd);
   setfield("%tryall%", TryAll ? " --try-all-secrets":"", cmdline);
   AnsiString tempfile = GPGTempFile();
   setfield("%filename%", tempfile, cmdline);

   TStringList *output = new TStringList();
   TStringList *headers = new TStringList();
   getheaders(headers, Message);
   Message->SaveToFile(tempfile);

   TGPGConsole *con = new TGPGConsole(Owner);
   bool ok = con->Run(cmdline, Passphrase.c_str(), NULL, output);
   if (ok || con->status->Text.Pos("DECRYPTION_OKAY"))
   {
	  FormatSigVerification(con->error);
	  Message->AddStrings(headers);
	  if (con->error->Count)
	  {
		 headers->AddStrings(con->error);
		 headers->Add("*** BEGIN PGP DECRYPTED/VERIFIED MESSAGE ***");
		 headers->Add("");
	  }
	  else
	  {
		 headers->Add("*** BEGIN PGP DECRYPTED MESSAGE ***");
		 headers->Add("");
	  }
	  Message->Text = output->Text;
	  putheaders(Message, headers);
	  if (con->error->Count)
	  {
		 Message->Add("");
		 Message->Add("*** END PGP DECRYPTED/VERIFIED MESSAGE ***");
	  }
	  else
	  {
		 Message->Add("");
		 Message->Add("*** END PGP DECRYPTED MESSAGE ***");
	  }
	  ok = true;
   }
   else
   {
	  if (!Quiet)
		 ShowGPGError(Owner, con->error, con->status);
	  putheaders(Message, headers);
   }
   delete con;

   DeleteFile(tempfile);
   delete headers;
   delete output;

   Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"Done!");
   Snooze(200);

   return ok;
}
//---------------------------------------------------------------------------

bool __fastcall GPGGetEncryptionKeys(TStrings *Message, TStrings *Keys)
{
   Keys->Clear();

   if (!checkgpgavailable())
	  return false;

   AnsiString cmdline = getcommandline(GPGlistencryptionkeys, GPGListEncryptionKeysCmd);
   TStringList *headers = new TStringList();
   getheaders(headers, Message);

   TStringList *output = new TStringList();
   TGPGConsole *con = new TGPGConsole(NULL);
   bool ok = con->Run(cmdline, NULL, Message, output);
   if (ok)
   {
	  AnsiString id, line;
	  for (int i = 0; i < con->status->Count; i++)
	  {
		 line = con->status->Strings[i];
		 if (line.Pos("ENC_TO") > 0)
		 {
			id = GetTok(2, line, ' ');
			assert(id.Length() == 16);
			Keys->Add("0x" + id.SubString(id.Length() - 7, 8));
		 }
	  }
	  for (int i = 0; i < con->error->Count; i++)
	  {
		 line = con->error->Strings[i];
		 if (line.Pos("passphrase") > 0)
			Keys->Add(line.SubString(20, 64).Trim());
	  }
   }
   else
   {
	  Snooze(200);
	  ShowGPGError(Screen->ActiveForm, con->error, con->status);
   }

   putheaders(Message, headers);

   delete con;
   delete output;
   delete headers;

   return ok;
}
//---------------------------------------------------------------------------

bool __fastcall GPGEncryptStringToStream(AnsiString &Passphrase, AnsiString &Input,
   TStream *Output)
{
   // max size of data = 1k
   if (!checkgpgavailable())
	  return false;

   AnsiString cmdline = getcommandline(GPGencryptstringtostream, GPGEncryptStringToStreamCmd),
	  filename = AppDB();

   TGPGConsole *con = new TGPGConsole(Screen->ActiveForm);
   bool ok = con->Run(cmdline, Passphrase, Input, Output);
   ZeroString(Passphrase);
   if (!ok)
   {
	  Snooze(200);
	  ShowGPGError(Screen->ActiveForm, con->error, con->status);
   }
   delete con;

   Snooze(200);

   //checkremovekeyringempties();
   return ok;
}
//---------------------------------------------------------------------------

bool __fastcall GPGDecryptStreamToString(AnsiString &Passphrase, TStream *Input,
   AnsiString &Output)
{
   // max size of data = 1k
   if (!checkgpgavailable())
	  return false;

   AnsiString cmdline = getcommandline(GPGdecryptstreamtostring, GPGDecryptStreamToStringCmd),
	  filename = AppDB();

   TGPGConsole *con = new TGPGConsole(Screen->ActiveForm);
   bool ok = con->Run(cmdline, Passphrase, Input, Output);
   ZeroString(Passphrase);
   if (!ok)
   {
	  Snooze(200);
	  ShowGPGError(Screen->ActiveForm, con->error, con->status);
   }
   delete con;

   Snooze(200);

   //checkremovekeyringempties();
   return ok;
}
//---------------------------------------------------------------------------

/*void __fastcall CheckFixQPInArmor(TStrings *msg)
{
   // check last 2 lines of armor
   AnsiString line;
   int p, idx = msg->IndexOf(EndPGPMessage);
   assert(idx > 3);
   // check next to last line
   line = msg->Strings[idx - 2];
   p = line.Pos("=3D");
   if (p > 0)
   {
	  line.Delete(p + 1, 2);
	  msg->Strings[idx - 2] = line;
   }
   // check last line
   line = msg->Strings[idx - 1];
   p = line.Pos("=3D");
   if (p > 0)
   {
	  line.Delete(p + 1, 2);
	  msg->Strings[idx - 1] = line;
   }
}*/
//---------------------------------------------------------------------------

/*bool __fastcall FixLine(AnsiString &line)
{
   int p, x = 1;
   bool fix = false;
   while ((p = line.SubString(x, line.Length()).Pos("=")) > 0)
   {
	  if (p == line.Length())
		 break;
	  if (line.SubString(p, 2) == "==")
	  {
		 p = p + 2;
		 continue;
	  }
	  try
	  {
		 char c = StrToInt("0x"+line.SubString(p + 1, 2))
		 sm(AnsiString(StrToInt("0x".line.SubString(p + 1, 2)))+"\n"+AnsiString(c)));
		 line.Delete(p, 3);
		 line.Insert(c, p);
		 if (!fix)
			fix = true;
	  }
	  catch(...)
	  {
		 // forgetaboutit!!!
	  }
	  x = p + 1;
   }
   return fix;
}
//---------------------------------------------------------------------------

void __fastcall FixFaultyArmor(TStrings *msg)
{
   bool skip = true;
   for (int i = 0; i < msg->Count; i++)
   {
	  AnsiString line = msg->Strings[0];
	  if (line == EndPGPMessage)
		 break;
	  if (line == BeginPGPMessage)
	  {
		 skipping = false;
		 continue;
	  }
	  if (skipping)
		 continue;

	  if (FixLine(line))
		 msg->Strings[i] = line;
   }
}*/
//---------------------------------------------------------------------------

AnsiString __fastcall getcommandline(const AnsiString &item, const AnsiString defvalue)
{
   TIniFile *ini = new TIniFile(AppIni());
   AnsiString cmdline = ini->ReadString("gpg", item, defvalue);
   delete ini;
   return GPGExe() + " " + cmdline.TrimLeft();
}
//---------------------------------------------------------------------------

bool __fastcall setfield(const AnsiString &field, const AnsiString &value,
   AnsiString &line)
{
   int x = line.Pos(field);
   if (x == 0)
      return false;
   line.Delete(x, field.Length());
   line.Insert(value, x);
   return true;
}
//---------------------------------------------------------------------------

bool __fastcall checkgpgavailable()
{
   bool ok = true;
   if (!FileExists(GPGExe()))
   {
	  MsgBox(Screen->ActiveForm,
		 "Problem! The file 'gpg.exe' is missing from my home directory.  \n"
		 "I can't complete the PGP operation! Reinstalling the program   \n"
		 "should fix this problem.",
		  MT_ERROR, MB_OK);
	  ok = false;
   }
   if (!FileExists(GPGIconvDll()))
   {
	  MsgBox(Screen->ActiveForm,
		 "Problem! The gpg file 'iconv.dll' is missing from my home directory.  \n"
		 "I can't complete the PGP operation! Reinstalling the program   \n"
		 "should fix this problem.",
		  MT_ERROR, MB_OK);
	  ok = false;
   }

   return ok;
}
//---------------------------------------------------------------------------

void __fastcall getheaders(TStrings *headers, TStrings *message)
{
   assert(headers != NULL);
   assert(message != NULL);
   while (message->Count)
   {
	  AnsiString line = message->Strings[0].Trim();
	  headers->Add(line);
	  message->Delete(0);
	  if (line.IsEmpty())
		  break;
   }
   assert(message->Count > 0);
}
//---------------------------------------------------------------------------

void __fastcall putheaders(TStrings *message, TStrings *headers)
{
   for (int i = 0; i < headers->Count; i++)
	  message->Insert(i, headers->Strings[i]);
}
//---------------------------------------------------------------------------

void __fastcall FormatSigVerification(TStrings *verified)
{
   assert(verified->Count);
   while (verified->Count && !verified->Strings[0].Pos("Signature made"))
	  verified->Delete(0);
   if (!verified->Count)
	  return;

   TStringList *veri = new TStringList();
   veri->Add("*** PGP SIGNATURE VERIFICATION ***");
   for (int i = 0; i < verified->Count; i++)
   {
	  AnsiString line = verified->Strings[i];
	  if (line.Pos("gpg:") == 1)
		 veri->Add("*** " + line.SubString(6, 128));
   }
   veri->Add("*** Verified " + FormatDateTime("MM/DD/YY HH:NN:SS", Now()));
   verified->Text = veri->Text;
   delete veri;
}

//---------------------------------------------------------------------------
//   GPGConsole
//---------------------------------------------------------------------------

__fastcall TGPGConsole::TGPGConsole(TForm *owner)
   : TObject()
{
   Owner = owner;
   hInputRead = NULL;
   hInputWrite = NULL;
   hOutputRead = NULL;
   hOutputWrite = NULL;
   hErrorRead = NULL;
   hErrorWrite = NULL;
   hPassRead = NULL;
   hPassWrite = NULL;
   hAttrRead = NULL;
   hAttrWrite = NULL;
   hStatRead = NULL;
   hStatWrite = NULL;

   bstatus = false;
   battrib = false;

   exitwait = false;
   exitcode = 0;

   attrib = new TStringList();
   status = new TStringList();
   error = new TStringList();
}
//---------------------------------------------------------------------------

__fastcall TGPGConsole::~TGPGConsole()
{
   delete error;
   delete status;
   delete attrib;

   DestroyHandles();
}
//---------------------------------------------------------------------------

void __fastcall TGPGConsole::SetupFileDescriptors(AnsiString &cmdline)
{
   int p;
   AnsiString fd;

   p = cmdline.Pos("%stat%");
   if (p)
   {
      fd = AnsiString((int)hStatWrite);
      setfield("%stat%", fd, cmdline);
   }

   p = cmdline.Pos("%attr%");
   if (p)
   {
      fd = AnsiString((int)hAttrWrite);
      setfield("%attr%", fd, cmdline);
   }

   p = cmdline.Pos("%pass%");
   if (p)
   {
      fd = AnsiString((int)hPassRead);
      setfield("%pass%", fd, cmdline);
   }
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::CreateHandles()
{
   TSecurityAttributes sa;
   HANDLE hOutputReadTmp, hErrorReadTmp, hInputWriteTmp, hPassWriteTmp,
      hAttrReadTmp, hStatReadTmp;

   DestroyHandles();
   sa.nLength = sizeof(sa);
   sa.lpSecurityDescriptor = NULL;
   sa.bInheritHandle = true;

   if (!CreatePipe(&hInputRead, &hInputWriteTmp, &sa, 0))
   {
	  ShowError("Pipe create error: \n");
	  return false;
   }

   if (!CreatePipe(&hOutputReadTmp, &hOutputWrite, &sa, 0))
   {
      ShowError("Pipe create error: \n");
      return false;
   }

   if (!CreatePipe(&hErrorReadTmp, &hErrorWrite, &sa, 0))
   {
      ShowError("Pipe create error: \n");
      return false;
   }

   if (bstatus && !CreatePipe(&hStatReadTmp, &hStatWrite, &sa, 0))
   {
	  ShowError("Pipe create error: \n");
      return false;
   }

   if (battrib && !CreatePipe(&hAttrReadTmp, &hAttrWrite, &sa, 0))
   {
      ShowError("Pipe create error: \n");
	  return false;
   }

   if (bpasswd && !CreatePipe(&hPassRead, &hPassWriteTmp, &sa, 0))
   {
      ShowError("Pipe create error: \n");
      return false;
   }

   if (!DuplicateHandle(GetCurrentProcess(), hInputWriteTmp,
      GetCurrentProcess(), &hInputWrite, 0, FALSE, DUPLICATE_SAME_ACCESS))
   {
	  ShowError("Duplicate handle error: \n");
      return false;
   }

   if (!DuplicateHandle(GetCurrentProcess(), hOutputReadTmp,
      GetCurrentProcess(), &hOutputRead, 0, FALSE, DUPLICATE_SAME_ACCESS))
   {
	  ShowError("Duplicate handle error: \n");
      return false;
   }

   if (!DuplicateHandle(GetCurrentProcess(), hErrorReadTmp,
	  GetCurrentProcess(), &hErrorRead, 0, FALSE, DUPLICATE_SAME_ACCESS))
   {
	  ShowError("Duplicate handle error: \n");
      return false;
   }

   if (bstatus && !DuplicateHandle(GetCurrentProcess(), hStatReadTmp,
      GetCurrentProcess(), &hStatRead, 0, FALSE, DUPLICATE_SAME_ACCESS))
   {
	  ShowError("Duplicate handle error: \n");
      return false;
   }

   if (battrib && !DuplicateHandle(GetCurrentProcess(), hAttrReadTmp,
      GetCurrentProcess(), &hAttrRead, 0, FALSE, DUPLICATE_SAME_ACCESS))
   {
      ShowError("Duplicate handle error: \n");
      return false;
   }

   if (bpasswd && !DuplicateHandle(GetCurrentProcess(), hPassWriteTmp,
      GetCurrentProcess(), &hPassWrite, 0, FALSE, DUPLICATE_SAME_ACCESS))
   {
	  ShowError("Duplicate handle error: \n");
      return false;
   }

   if (battrib) CloseHandle(hAttrReadTmp);
   if (bstatus) CloseHandle(hStatReadTmp);
   if (bpasswd) CloseHandle(hPassWriteTmp);
   CloseHandle(hOutputReadTmp);
   CloseHandle(hErrorReadTmp);
   CloseHandle(hInputWriteTmp);

   return true;
}
//---------------------------------------------------------------------------

void __fastcall TGPGConsole::DestroyHandles()
{
   if (hInputRead) CloseHandle(hInputRead);
   if (hOutputRead) CloseHandle(hOutputRead);
   if (hInputWrite) CloseHandle(hInputWrite);
   if (hOutputWrite) CloseHandle(hOutputWrite);
   if (hErrorRead) CloseHandle(hErrorRead);
   if (hErrorWrite) CloseHandle(hErrorWrite);
   if (hPassRead) CloseHandle(hPassRead);
   if (hPassWrite) CloseHandle(hPassWrite);
   if (hAttrRead) CloseHandle(hAttrRead);
   if (hAttrWrite) CloseHandle(hAttrWrite);
   if (hStatRead) CloseHandle(hStatRead);
   if (hStatWrite) CloseHandle(hStatWrite);

   hInputRead = NULL;
   hInputWrite = NULL;
   hOutputRead = NULL;
   hOutputWrite = NULL;
   hErrorRead = NULL;
   hErrorWrite = NULL;
   hPassRead = NULL;
   hPassWrite = NULL;
   hAttrRead = NULL;
   hAttrWrite = NULL;
   hStatRead = NULL;
   hStatWrite = NULL;
}
//---------------------------------------------------------------------------

void __fastcall TGPGConsole::PrepareStartupInfo()
{
   memset(&startupinfo, 0, sizeof(STARTUPINFO));
   startupinfo.cb = sizeof (STARTUPINFO);
   startupinfo.lpReserved = NULL;
   startupinfo.lpDesktop = NULL;
   startupinfo.lpTitle = NULL;
   startupinfo.lpReserved2 = NULL;
   startupinfo.cbReserved2 = 0;
   startupinfo.dwFlags = STARTF_USESTDHANDLES;

   startupinfo.hStdInput = hInputRead;
   startupinfo.hStdOutput = hOutputWrite;
   startupinfo.hStdError = hErrorWrite;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::LaunchChildProcess(const AnsiString &cmdLine)
{
   if (!CreateProcess(
	  NULL,             // pointer to filename
	  cmdLine.c_str(),  // pointer to command line string
	  NULL, 		// pointer to process security attributes, we don't need to inherit the handle
	  NULL, 		// pointer to thread security attributes
	  true,  		// handle inheritance flag
	  CREATE_NO_WINDOW, // creation flags
	  NULL, 		// pointer to new environment block
	  NULL, 		// pointer to current directory name
	 &startupinfo,      // pointer to STARTUPINFO
	 &processinfo       // pointer to PROCESS_INFORMATION
   ))
   {
	  ShowError("CreateProcess error: \n");
	  return false;
   }

   CloseHandle(processinfo.hThread);

   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::SendPassphrase(char *passphrase)
{
   DWORD byteswrote;
   if (!WriteFile(hPassWrite, passphrase, strlen(passphrase), &byteswrote, NULL))
   {
	  ShowError("Error sending the passphrase to GPG: \n");
	  return false;
   }

   CloseHandle(hPassWrite);
   hPassWrite = NULL;
   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::Run(AnsiString cmdLine, char *Passphrase,
   TStrings *Input, TStrings *Output)
{
   bstatus = cmdLine.Pos("%stat%");
   battrib = cmdLine.Pos("%attr%");
   bpasswd = cmdLine.Pos("%pass%");

   if (!CreateHandles())
	  return false;

   SetupFileDescriptors(cmdLine);
   PrepareStartupInfo();

   if (!LaunchChildProcess(cmdLine))
	  return false;

   if (Passphrase && !SendPassphrase(Passphrase))
	  return false;

   if (Input && !SendInput(Input))
	  return false;

   if (!ReadOutput(Output))
	  return false;

   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::Run(AnsiString cmdLine, AnsiString &Passphrase,
	  const AnsiString &Input, TStream *Output)
{
   assert(Passphrase.Length() > 0);
   bstatus = cmdLine.Pos("%stat%");
   battrib = cmdLine.Pos("%attr%");
   bpasswd = cmdLine.Pos("%pass%");

   if (!CreateHandles())
	  return false;

   SetupFileDescriptors(cmdLine);
   PrepareStartupInfo();

   if (!LaunchChildProcess(cmdLine))
	  return false;

   if (!SendPassphrase(Passphrase.c_str()))
	  return false;

   if (!SendInput(Input))
	  return false;

   if (!ReadOutput(Output))
	  return false;

   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::Run(AnsiString cmdLine, AnsiString &Passphrase,
	  TStream *Input, AnsiString &Output)
{
   assert(Passphrase.Length() > 0);
   bstatus = cmdLine.Pos("%stat%");
   battrib = cmdLine.Pos("%attr%");
   bpasswd = cmdLine.Pos("%pass%");
 //ZeroString(Output);

   if (!CreateHandles())
	  return false;

   SetupFileDescriptors(cmdLine);
   PrepareStartupInfo();

   if (!LaunchChildProcess(cmdLine))
	  return false;

   if (!SendPassphrase(Passphrase.c_str()))
	  return false;

   if (!SendInput(Input))
	  return false;

   if (!ReadOutput(Output))
	  return false;

   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::SendInput(const AnsiString &input)
{
   DWORD byteswrote;
//sm(IntToStr(input.Length())+"  "+input);
   AnsiString line = input + "\r\n";
   if (!WriteFile(hInputWrite, line.c_str(), line.Length(), &byteswrote, NULL))
   {
	  ShowError("Error sending the data to GPG: \n");
	  return false;
   }

   CloseHandle(hInputWrite);
   hInputWrite = NULL;

   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::SendInput(TStream *input)
{
   char buf[4096];
   DWORD bufsize = sizeof(buf), byteswrote, bytesread;// = bufsize;

   if (input)
   {
	  assert(input->Size > 0);
	  input->Seek(0, soFromBeginning);
	  while ((bytesread = input->Read(buf, bufsize)) > 0)
	  {
		 if (!WriteFile(hInputWrite, buf, bytesread, &byteswrote, NULL))
		 {
			ShowError("Error sending the data to GPG: \n");
			return false;
		 }
	  }

	  CloseHandle(hInputWrite);
	  hInputWrite = NULL;
   }

   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::SendInput(TStrings *input)
{
   DWORD byteswrote;
   if (input)
   {
	  int i = 0;
	  while (i < input->Count)
	  {
		 AnsiString line = input->Strings[i] + "\r\n";
		 if (!WriteFile(hInputWrite, line.c_str(), line.Length(), &byteswrote, NULL))
		 {
			ShowError("Error sending the data to GPG: \n");
			return false;
		 }
		 i++;
	  }
	  CloseHandle(hInputWrite);
	  hInputWrite = NULL;
   }

   return true;
}
//---------------------------------------------------------------------------
// exitwait: waitforsingleobject: WAIT_ABANDON, WAIT_OBJECT_0, WAIT_TIMEOUT
// exitcode: this is the exec'd programs exit code

bool __fastcall TGPGConsole::ReadOutput(TStrings *output)
{
   DWORD bt;
   int ms = 100;
   time_t marktime = time(NULL);

   while (time(NULL) < marktime + 30)
   {
	  exitwait = WaitForSingleObject(processinfo.hProcess, ms);
	  if (exitwait == WAIT_FAILED)
	  {
		 ShowError("Error occured while waiting for GPG operation to finish: \n");
		 return false;
	  }

	  while (PeekNamedPipe(hOutputRead, NULL, NULL, NULL, &bt, NULL) && bt)
	  {
		 if (!GetWaiting(hOutputRead, output))
		 {
			ShowError("Error reading GPG output: \n");
			return false;
		 }
	  }

	  while (PeekNamedPipe(hErrorRead, NULL, NULL, NULL, &bt, NULL) && bt)
	  {
		 if (!GetWaiting(hErrorRead, error))
		 {
			ShowError("Error reading GPG error output: \n");
			return false;
		 }
	  }

	  if (bstatus)
	  while (PeekNamedPipe(hStatRead, NULL, NULL, NULL, &bt, NULL) && bt)
	  {
		 if (!GetWaiting(hStatRead, status))
		 {
			ShowError("Error reading GPG status output: \n");
			return false;
		 }
	  }

	  if (battrib)
	  while (PeekNamedPipe(hAttrRead, NULL, NULL, NULL, &bt, NULL) && bt)
	  {
		 if (!GetWaiting(hAttrRead, attrib))
		 {
			ShowError("Error reading GPG logger output: \n");
			return false;
		 }
	  }

	  if (!GetExitCodeProcess(processinfo.hProcess, &exitcode))
	  {
		 ShowError("Error getting GPG exit code: \n");
		 return false;
	  }
	  if (exitwait != WAIT_TIMEOUT)
		 break;

	  marktime = time(NULL);
   }

   if (exitcode)
   {
	  // 1 = GPG processing error
	  // 2 = GPG command error
	  return false;
   }

   return true;
}
//---------------------------------------------------------------------------
// exitwait: waitforsingleobject: WAIT_ABANDON, WAIT_OBJECT_0, WAIT_TIMEOUT
// exitcode: this is the exec'd programs exit code

bool __fastcall TGPGConsole::ReadOutput(AnsiString &output)
{
   DWORD bt;
   int ms = 100;
   time_t marktime = time(NULL);

   while (time(NULL) < marktime + 30)
   {
	  exitwait = WaitForSingleObject(processinfo.hProcess, ms);
	  if (exitwait == WAIT_FAILED)
	  {
		 ShowError("Error occured while waiting for GPG operation to finish: \n");
		 return false;
	  }

	  while (PeekNamedPipe(hOutputRead, NULL, NULL, NULL, &bt, NULL) && bt)
	  {
		 if (!GetWaiting(hOutputRead, output))
		 {
			ShowError("Error reading GPG output: \n");
			return false;
		 }
	  }

	  while (PeekNamedPipe(hErrorRead, NULL, NULL, NULL, &bt, NULL) && bt)
	  {
		 if (!GetWaiting(hErrorRead, error))
		 {
			ShowError("Error reading GPG error output: \n");
			return false;
		 }
	  }

	  if (bstatus)
	  while (PeekNamedPipe(hStatRead, NULL, NULL, NULL, &bt, NULL) && bt)
	  {
		 if (!GetWaiting(hStatRead, status))
		 {
			ShowError("Error reading GPG status output: \n");
			return false;
		 }
	  }

	  if (battrib)
	  while (PeekNamedPipe(hAttrRead, NULL, NULL, NULL, &bt, NULL) && bt)
      {
         if (!GetWaiting(hAttrRead, attrib))
		 {
			ShowError("Error reading GPG logger output: \n");
			return false;
		 }
	  }

	  if (!GetExitCodeProcess(processinfo.hProcess, &exitcode))
	  {
		 ShowError("Error getting GPG exit code: \n");
		 return false;
	  }

	  if (exitwait != WAIT_TIMEOUT)
		 break;

	  marktime = time(NULL);
   }

   while (output.Length() && (output[output.Length()] == '\r' || output[output.Length()] == '\n'))
	  output.Delete(output.Length(), 1);

   if (exitcode)
   {
      // 1 = GPG processing error
      // 2 = GPG command error
      return false;
   }

   return true;
}
//---------------------------------------------------------------------------
// exitwait: waitforsingleobject: WAIT_ABANDON, WAIT_OBJECT_0, WAIT_TIMEOUT
// exitcode: this is the exec'd programs exit code

bool __fastcall TGPGConsole::ReadOutput(TStream *output)
{
   DWORD bt;
   int ms = 100;
   time_t marktime = time(NULL);

   while (time(NULL) < marktime + 30)
   {
      exitwait = WaitForSingleObject(processinfo.hProcess, ms);
      if (exitwait == WAIT_FAILED)
      {
         ShowError("Error occured while waiting for GPG operation to finish: \n");
         return false;
      }

      while (PeekNamedPipe(hOutputRead, NULL, NULL, NULL, &bt, NULL) && bt)
      {
         if (!GetWaiting(hOutputRead, output))
         {
            ShowError("Error reading GPG output: \n");
            return false;
         }
      }

      while (PeekNamedPipe(hErrorRead, NULL, NULL, NULL, &bt, NULL) && bt)
      {
         if (!GetWaiting(hErrorRead, error))
         {
            ShowError("Error reading GPG error output: \n");
            return false;
         }
      }

      if (status)
      while (PeekNamedPipe(hStatRead, NULL, NULL, NULL, &bt, NULL) && bt)
      {
         if (!GetWaiting(hStatRead, status))
         {
            ShowError("Error reading GPG status output: \n");
            return false;
         }
      }

      if (attrib)
      while (PeekNamedPipe(hAttrRead, NULL, NULL, NULL, &bt, NULL) && bt)
      {
         if (!GetWaiting(hAttrRead, attrib))
         {
            ShowError("Error reading GPG logger output: \n");
            return false;
         }
      }

      if (!GetExitCodeProcess(processinfo.hProcess, &exitcode))
      {
         ShowError("Error getting GPG exit code: \n");
         return false;
      }

      if (exitwait != WAIT_TIMEOUT)
         break;

      marktime = time(NULL);
   }

   if (exitcode)
   {
      // 1 = GPG processing error
      // 2 = GPG command error
      return false;
   }
   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::GetWaiting(HANDLE hReadHandle, TStrings *output)
{
   static char buf[1024];
   static AnsiString incomingline;
   unsigned long bytesread, bufsize = sizeof(buf);

   if (!ReadFile(hReadHandle, buf, bufsize, &bytesread, NULL))
	  return false;

   if (bytesread > 0)
   {
	  for (unsigned i = 0; i < bytesread; i++)
	  {
		 if (buf[i] != '\0' && buf[i] != '\n' && buf[i] != '\r')
			incomingline += String(buf[i]);
		 if (buf[i] == '\n')
		 {
			output->Add(incomingline);
			incomingline.SetLength(0);
		 }
	  }
   }
   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::GetWaiting(HANDLE hReadHandle, TStream *output)
{
   static char buf[32768];
   unsigned long bytesread, bufsize = sizeof(buf);

   if (!ReadFile(hReadHandle, buf, bufsize, &bytesread, NULL))
      return false;

   if (bytesread > 0)
   {
      output->Write(buf, bytesread);
   }

   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::GetWaiting(HANDLE hReadHandle, AnsiString &output)
{
   static char buf[32768] = { '\0' };
   unsigned long bytesread, bufsize = sizeof(buf);

   output = "";
   if (!ReadFile(hReadHandle, buf, bufsize, &bytesread, NULL))
	  return false;

   if (bytesread > 0)
   {
	  output = buf;
   }

   return true;
}
//---------------------------------------------------------------------------

void __fastcall TGPGConsole::ShowError(const AnsiString &errormsg)
{
   MsgBox(Owner, errormsg + GetLastErrorStr(), MT_ERROR, MB_OK);
}

//---------------------------------------------------------------------------
// GPGerrDlg
//---------------------------------------------------------------------------

void __fastcall ShowGPGError(TForm *Owner, TStrings *Error, TStrings *Status)
{
   TGPGerrDlg *errdlg = new TGPGerrDlg(Owner, Error, Status);
   errdlg->ShowModal();
   delete errdlg;
}
//---------------------------------------------------------------------------

void __fastcall ShowGPGError(TForm *Owner, TStrings *ErrorStatus)
{
   bool inerror = true;
   TStringList *error = new TStringList();
   TStringList *status = new TStringList();
   for (int i = 0; i < ErrorStatus->Count; i++)
   {
	  AnsiString line = ErrorStatus->Strings[i];
	  if (inerror && line == "-----")
	  {
		 inerror = false;
		 continue;
	  }
	  if (!inerror)
		 status->Add(line);
	  else
		 error->Add(line);
   }
   ShowGPGError(Owner, error, status);
   delete status;
   delete error;
}
//---------------------------------------------------------------------------

__fastcall TGPGerrDlg::TGPGerrDlg(TComponent* Owner)
		: TForm(Owner)
{
   ErrorBox->Clear();
   TIniFile *ini = new TIniFile(AppIni());
   AssignFont(Font, ini->ReadString("settings", "dialogfont", OptsDefDialogFont));
   AssignFont(ErrorBox->Font, ini->ReadString("settings", "editorfont",
	  OptsDefEditorFont));
   delete ini;
}
//---------------------------------------------------------------------------

__fastcall TGPGerrDlg::TGPGerrDlg(TComponent* Owner, TStrings *GPGerror,
   TStrings *GPGstatus) : TForm(Owner)
{
   PopupParent = (TForm*)Owner;
   ErrorBox->Clear();
   ErrorBox->Lines->Text = GPGerror->Text;
   TIniFile *ini = new TIniFile(AppIni());
   AssignFont(Font, ini->ReadString("settings", "dialogfont", OptsDefDialogFont));
   AssignFont(ErrorBox->Font, ini->ReadString("settings", "editorfont",
	  OptsDefEditorFont));
   delete ini;

   status = new TStringList();
   status->Text = GPGstatus->Text;
   error = new TStringList();
   error->Text = GPGerror->Text;

   StatusBtn->Visible = GPGstatus->Count;
}
//---------------------------------------------------------------------------

__fastcall TGPGerrDlg::~TGPGerrDlg()
{
   delete status;
}
//---------------------------------------------------------------------------

void __fastcall TGPGerrDlg::FormShow(TObject *Sender)
{
   HWND DC;
   TCanvas *canvas = new TCanvas();
   canvas->Handle = GetDeviceContext(DC);
   canvas->Font->Assign(ErrorBox->Font);

   int high = canvas->TextHeight("yY") * (ErrorBox->Lines->Count + 2);
   ClientHeight = high + (ClientHeight - ErrorBox->Height);
   int w, wide = 0;
   for (int i = 0; i < ErrorBox->Lines->Count; i++)
   {
      w = canvas->TextWidth(ErrorBox->Lines->Strings[i]);
      if (w > wide)
         wide = w;
   }
   wide = wide + (ClientWidth - ErrorBox->Width) + 40;
   ClientWidth = wide < 400 ? 400 : wide;

   OkBtn->Left = (Width / 2) - (OkBtn->Width / 2);

   delete canvas;
}

//---------------------------------------------------------------------------

void __fastcall TGPGerrDlg::StatusBtnClick(TObject *Sender)
{
   if (StatusBtn->Caption == "More...")
   {
      ErrorBox->Lines->Add("");
      ErrorBox->Lines->Add("-------------------");
      ErrorBox->Lines->Add("");
      ErrorBox->Lines->AddStrings(status);
      StatusBtn->Caption = "Less...";
   }
   else
   {
      ErrorBox->Text = error->Text;
      StatusBtn->Caption = "More...";
   }
   FormShow(NULL);
}
//---------------------------------------------------------------------------

void __fastcall TGPGerrDlg::PopupMenu1Popup(TObject *Sender)
{
   Copy1->Enabled = ErrorBox->SelLength > 0;
}
//---------------------------------------------------------------------------

void __fastcall TGPGerrDlg::Copy1Click(TObject *Sender)
{
   ErrorBox->CopyToClipboard();        
}
//---------------------------------------------------------------------------

void __fastcall TGPGerrDlg::Selectall1Click(TObject *Sender)
{
    ErrorBox->SelectAll();
}
//---------------------------------------------------------------------------

