/*
*  mainwins.cpp
*
*  Copyright (C) 2012 Richard Christman
*
*  This file is part of QuickSilver AAM.
*
*  QuickSilver AAM is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by the
*  Free Software Foundation; either version 3 of the License, or (at your
*  option) any later version.
*
*  This program is distributed in the hope that it will be useful, but
*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
*  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
*  for more details.
*
*  You should have received a copy of the GNU General Public License along
*  with this program; if not, see <http://www.gnu.org/licenses>.
*
*  Additional permission under GNU GPL version 3 section 7
*
*  If you modify QuickSilver AAM, or any covered work, by linking or
*  combining it with CodeGear VCL (or a modified version of that library),
*  containing parts covered by the terms of CodeGear's license, the licensors
*  of QuickSilver AAM grant you additional permission to convey the
*  resulting work.
*
*/

//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include <inifiles.hpp>
#include <clipbrd.hpp>
#include <stdlib.h>
#include "mainwins.h"
#include "utility.h"
#include "global.h"
#include "msgmans.h"
#include "sha256.h"
#include "qsaver.h"
#include "qsadbs.h"
#include "paramsdlgs.h"
#include "gnupg.h"
#include "proxymandlgs.h"
#include "nntpmandlgs.h"
#include "pgpdpassdlgs2.h"
#include "pgpdlgs.h"
#include "msgwins.h"
#include "optionsdlgs.h"
#include "subjectsmandlgs.h"
#include "esub.h"
#include "logdlgs.h"
#include "pgpcache.h"
#include "cancelnntpdlgs.h"
#include "nntperrdlgs.h"
#include "dbpackers.h"
#include "help.h"
#include "eraser.h"
#include "dsadlgs.h"
#include "keyringdlgs.h"
#include "welcomedlgs.h"
#include "nymsrvmandlgs.h"
#include "gpg4.h"
#include "cfgmsgdlgs.h"
#include "messagepipes.h"
#include "nntpmods2.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TQSAMainWin *QSAMainWin;

const int lvWho = 0;
//const int lvNym = 1;
const int lvSubject = 1;
const int lvSize = 2;
const int imgNotRead = 0;
const int imgRead = 1;
const int statusRecordsWidth = 100;
const int statusProgressWidth = 120;
const int statusStatusPanel = 0;
const int statusProgressPanel = 1;
const int statusRecordsPanel = 2;
const TColor clPBar = clLime;
const AnsiString MessageTab = "Message";
const AnsiString StatusTab = "Status";
const AnsiString LogTab = "Log";
const AnsiString CancelCap = " Cancel  ";
const AnsiString GetBtnCap = " Get Mail  ";
const AnsiString ChkBtnCap = " Chk Mail  ";
const AnsiString CpyBtnCap = " Reply ";

const int blutrayicon = 0;
const int envtrayicon = 1;
const int errtrayicon = 2;

const AnsiString nymserverhdrs = ":from:date:";
const AnsiString messagehdrs = ":comment:from:to:cc:subject:date:";
const AnsiString indentchars = " \t";
//---------------------------------------------------------------------------
// main window
//---------------------------------------------------------------------------

void __fastcall CheckFixNymserverDefinitions()
{
   bool save = false;
   TStringList *server = new TStringList();
   TStringList *servers = new TStringList();
   DB->LoadNymservers(servers);
   for (int i = 0; i < servers->Count; i++)
   {
	  AnsiString name = servers->Names[i], type = servers->Values[name].LowerCase();
	  if (type == "zax" || type == "ghio")
	  {
		 server->Clear();
		 server->Values["server"] = name;
		 server->Values["type"] = type;
		 if (type == "ghio")
			server->Values["enckey"] = "3DES";
		 servers->Values[name] = server->CommaText;
		 save = true;
	  }
   }
   if (save)
	  DB->SaveNymservers(servers);

   delete servers;
   delete server;
}
//---------------------------------------------------------------------------

__fastcall TQSAMainWin::TQSAMainWin(TComponent* Owner)
	: TForm(Owner)
{
   Caption = QuickSilverAAMVersion(); cap = Caption;
   SetApplicationDataDirectory();
   ChDir(AppDir());

   if (!FileExists(AppDir()+"gnupg.nls\\en@quot.mo"))
   {
	  MsgBox(NULL,
		 "Problem! The file: 'gnupg.nls\\en@quot.mo'   \n"
		 "is missing from the QSL home directory. The file is required.   \n"
		 "Reinstalling the program should solve the problem.", MT_ERROR, MB_OK);
	  Application->ShowMainForm = false;
	  PostMessage(Handle, WM_QUIT, 0, 0);
	  return;
   }

   runminimized = false;
   for (int i = 1; i <= ParamCount(); i++)
   {
	  if (ParamStr(i).LowerCase().Pos("-minimize") == 1)
	  {
		 runminimized = true;
		 break;
	  }
   }

   firstrun = !FileExists(AppIni()) && !FileExists(AppDB());
   if (firstrun)
   {
	  if (!WelcomeToQSA(this))
	  {
		 Application->ShowMainForm = false;
		 PostMessage(Handle, WM_QUIT, 0, 0);
		 bailout = true;
		 return;
	  }
   }

   if (!CheckAllFilesPresent())
   {
	  Application->ShowMainForm = false;
	  PostMessage(Handle, WM_QUIT, 0, 0);
	  bailout = true;
	  return;
   }

   if (!InitESUB32Library())
   {
	  Application->ShowMainForm = false;
	  PostMessage(Handle, WM_QUIT, 0, 0);
	  bailout = true;
	  return;
   }

   if (!InitEraserLibrary())
   {
	  Application->ShowMainForm = false;
	  PostMessage(Handle, WM_QUIT, 0, 0);
	  bailout = true;
	  return;
   }

   CheckFixIdeaDllInGPGConf();
   profile = new TStringList();
   LoadIni(true);

   Message->Clear();
   Status->Clear();
   Log->Clear();
   Message->Align = alClient;
   Status->Align = alClient;
   Log->Align = alClient;
   SetPage(StatusTab);

   DB = OpenQSADB(AppDB());
   if (!DB)
   {
	  MsgBox(this,
		 "Failed to initialize mail database!    \n"
		 "Quitting!", MT_ERROR, MB_OK);
	  Application->ShowMainForm = false;
	  PostMessage(Handle, WM_QUIT, 0, 0);
	  delete profile;
	  profile = NULL;
	  return;
   }

   //TStringList *data = new TStringList();
   //DB->GetText("log", data);
   //look(data);
   //delete data;

   if (firstrun)
   {
	  //setup
	  AnsiString name;
	  TIniFile *ini = new TIniFile(AppIni());
	  TIniFile *def = new TIniFile(AppDefsIni());
	  TStringList *data = new TStringList();

	  // various settings
	  def->ReadSectionValues("settings", data);
	  for (int i = 0; i < data->Count; i++)
	  {
		 name = data->Names[i]; // 'account' 'defanonto' 'defoptions'
		 ini->WriteString("settings", name, data->Values[name]);
	  }
	  // nntp hosts
	  def->ReadSectionValues("nntphosts", data);
	  DB->SaveAccounts(data);
	  // nymservers
	  def->ReadSectionValues("nymservers", data);
	  DB->SaveNymservers(data);
	  // save modified DB
	  DB->SaveToFile();

	  delete data;
	  delete def;
	  delete ini;
   }

   if (!CheckInstallNymservers())
	  CheckFixNymserverDefinitions();

   // fix filename change
   if (FileExists(AppDir()+"qsadefs.ini") && FileExists(AppDir()+"qsa.defs.txt"))
	  DeleteFile(AppDir()+"qsadefs.ini");

   trayicon = NULL; //new TIcon();
   trayiconidx = blutrayicon;
   data = new TStringList();
   mail = new TStringList();
   error = new TStringList();
   LoadHtmlHelp();

   PGPCache = new TPGPPassphraseCache();

   ListView->DoubleBuffered = true;
   SetupToolbar();
   SetupStatusBar();
   LoadProfile(account, profile);
   LoadProxyBox();
   LoadServerBox();
   LoadListView();
   randomize();

   DB->LoadStatus(Status->Lines);
   Status->Modified = false;
   Status->SelStart = 0;
   Status->Perform(EM_SCROLLCARET, 0, 0);
   ActiveControl = ToolBar;
   checkminselapsed = 0;
   Bevel1->Visible = Win32MajorVersion < 6;
   if (!HasQSL())
	  CopyMsgBtn->Hint = " Copy message | Copy message to clipboard ";

   if (firstrun)
   {
	  if (runminimized)
		 runminimized = false;
	  Timer2->Enabled = true;
	  return;
   }

   Application->OnMessage = AppMessage;
   if (runminimized)
   {
	  if (!checkmail)
	  {
		 MsgBox(this,
			"Problem! You've set the program to minimize at startup,  \n"
			"but you don't have mail checking turned on! I'll open   \n"
			"the program. Go to the Options Dialog, General Tab.",
			 MT_INFO, MB_OK);
		 runminimized = false;
		 return;
	  }
	  SendMessage(Handle, WM_SYSCOMMAND, SC_MINIMIZE, 0);
	  if (!usetray)
		 Show();
   }

   if (checkmail && !CheckMailTimer->Enabled)
	  StartTimer->Enabled = true;
   if (!checkmail && CheckMailTimer->Enabled)
	  CheckMailTimer->Enabled = false;
//sm((int)this);
}
//---------------------------------------------------------------------------

__fastcall TQSAMainWin::~TQSAMainWin()
{
   if (bailout)
	  return;

   if (trayed)
	  TrayIcon_Delete();
   if (trayicon)
	  delete trayicon;

   ListView->OnSelectItem = NULL;

   AnsiString curwin = IntToStr(Left)+IntToStr(Top)+IntToStr(Width)+IntToStr(Height);
   for (int i = 0; i < ListView->Columns->Count; i++)
	  curwin += IntToStr(ListView->Columns->Items[i]->Width);
   curwin += IntToStr(ListView->Height);
   if (curwin != orgwin)
   {
	  TIniFile *ini = new TIniFile(AppIni());
	  SaveWindowPosition(this, "window", ListView, ini);
	  delete ini;
   }

   delete PGPCache;
   UnloadHtmlHelp();
   FreeESUB32Library();
   FreeEraserLibrary();

   delete error;
   if (mail)
	  for (int i = 0; i < mail->Count; i++)
		 delete mail->Objects[i]; //(TStringList*)mail->Objects[i];
   delete mail;
   mail = NULL;
   delete data;
   delete profile;

   CloseQSADB(DB);
}
//---------------------------------------------------------------------------
//deprecated
void __fastcall TQSAMainWin::Timer6Timer(TObject *Sender)
{
   Timer6->Enabled = false;
   if (usetray)
   {
	  TrayIcon_Create(blutrayicon);
	  trayed = true;
   }
   else
   {
   }
   Snooze(500);
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::FormCloseQuery(TObject *Sender, bool &CanClose)
{
   TIniFile *ini = new TIniFile(AppIni());
   bool checkingmail = ini->ReadBool("settings", "checkmail", false);
   delete ini;
   if (checkingmail && usetray && !trayed)
   {
	  switch(MsgBox(this,
		 "Minimize to the system tray?   ",
		  MT_CONFIRM, MB_YESNOCANCEL))
	  {
		  case IDYES:
			 Perform(WM_SYSCOMMAND, SC_MINIMIZE, 0);
			 goterror = false;
		  case IDCANCEL:
			 CanClose = false;
			 break;
	  }
	  return;
   }

   if (NNTPModHandle)
   {
	  if (!CancelNNTPRetrieve(this))
	  {
		 CanClose = false;
		 return;
	  }
	  if (NNTPModHandle)
	  {
		 delete NNTPThrd;
		 NNTPModHandle = NULL;
		 //NNTPMod = NULL;
	  }
	  Snooze(200);
   }

   if (Status->Modified)
	  DB->SaveStatus(Status->Lines);

   if (needspack)
   {
	  ShowStatus("packing database...");
	  PackDatabase(this);
	  needspack = false;
	  Snooze(200);
   }

   if (DB->modified && !DB->Save())
   {
	  CanClose = false;
	  return;
   }
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::AppMessage(tagMSG &Msg, bool &Handled)
{
   if (Msg.message == WM_QSAMESSAGE)
   {
	  switch(Msg.wParam)
	  {
		 case BRINGTOFRONT:
			ShowWindow(Handle, SW_SHOWDEFAULT);
			SetForegroundWindow(Handle);
			SetActiveWindow(Handle);
			TrayIcon_Delete();
			trayed = false;
			Handled = true;
			break;
	  }
   }
}
//---------------------------------------------------------------------------

bool __fastcall TQSAMainWin::IsStandAloneProgFilesInstall()
{
   if (FileExists(AppDir()+"qsl.exe") || FileExists(AppDir()+"qs.exe"))
	  return false;

   if (AppDir().Pos(GetEnvironmentVariable("programfiles").LowerCase()) == 1)
   {
	  return true;
   }
   return false;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::LoadIni(bool startup)
{
   TIniFile *ini = new TIniFile(AppIni());
   if (startup)
   {
	  LoadWindowPosition(this, "window", ListView, ini);
	  orgwin = IntToStr(Left)+IntToStr(Top)+IntToStr(Width)+IntToStr(Height);
	  for (int i = 0; i < ListView->Columns->Count; i++)
		 orgwin += IntToStr(ListView->Columns->Items[i]->Width);
	  orgwin += IntToStr(ListView->Height);
   }
   if (PGPCache)
	  PGPCache->SetMinutes(ini->ReadInteger("settings","cachemins", 0));
   hidehost = ini->ReadBool("settings", "hideacctsel", false);
   ServerBox->Visible = !hidehost;
   Label2->Visible = !hidehost;
   hideprxy = ini->ReadBool("settings", "hideprxysel", false);
   ProxyBox->Visible = !hideprxy;
   Label1->Visible = !hideprxy;

   AssignFont(Font, ini->ReadString("settings", "dialogfont", OptsDefDialogFont));
   AssignFont(ListView->Font, ini->ReadString("settings", "windowfont", OptsDefWindowFont));
   AssignFont(Message->Font, ini->ReadString("settings", "editorfont", OptsDefEditorFont));
   Status->Font->Assign(Message->Font);
   Log->Font->Assign(Message->Font);
   ServerBox->Font->Assign(ListView->Font);
   ProxyBox->Font->Assign(ListView->Font);
   nntpproxy = ini->ReadString("settings", "proxy", "none");
   account = ini->ReadString("settings", "account", "none");
   newsgroup = ini->ReadString("settings", "newsgroup", "alt.anonymous.messages");
   usetray = ini->ReadBool("settings", "usetray", false);
   checkmail = ini->ReadBool("settings", "checkmail", false);
   checkmailmins = ini->ReadInteger("settings", "checkmailmins", 30);
   savedecrypted = ini->ReadBool("settings", "savedecrypted", false);
   NNTPTimeoutTimer->Interval = ini->ReadInteger("settings", "nntptimeout", 30) * 1000;
   FewerHeaders->OnClick = NULL;
   FewerHeaders->Checked = ini->ReadBool("settings", "fewerheaders", OptsDefFewerHeaders);
   FewerHeaders->OnClick = FewerHeadersClick;
   delete ini;

   SetCaption();
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::SetupToolbar()
{
   Bevel7->Left = CopyMsgBtn->Left + CopyMsgBtn->Width + 1;
   CopyBtn->Left = Bevel7->Left + 3;
   DecryptBtn->Left = CopyBtn->Left + CopyBtn->Width; // + 1;
   Bevel2->Left = DecryptBtn->Left + DecryptBtn->Width + 1;
   DeleteBtn->Left = Bevel2->Left + 2;
   Bevel3->Left = DeleteBtn->Left + DeleteBtn->Width + 1;
   Subjects->Left = Bevel3->Left + 15;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::SetupStatusBar()
{
   StatusBar->Panels->Items[statusProgressPanel]->Width = statusProgressWidth;
   StatusBar->Panels->Items[statusRecordsPanel]->Width = statusRecordsWidth;
   StatusBar->Panels->Items[statusStatusPanel]->Width = ClientWidth -
	  (statusProgressWidth + statusRecordsWidth);
   ProgressPanel = StatusBar->Panels->Items[statusProgressPanel];
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::SetCaption()
{
   Caption = cap + "  " +newsgroup;
   if (hidehost && account != "none")
	  Caption = Caption + "  " + account;
   if (hideprxy && nntpproxy != "none")
	  Caption = Caption + "  " + nntpproxy;
}
//---------------------------------------------------------------------------

bool __fastcall TQSAMainWin::GetKeyringPathnames(const AnsiString &gpgconf,
   AnsiString &secring, AnsiString &pubring)
{
   secring = "";
   pubring = "";
   if (FileExists(gpgconf))
   {
	  TStringList *lines = new TStringList();
	  lines->LoadFromFile(GPGConf());
	  for (int i = 0; i < lines->Count; i++)
	  {
		 AnsiString line = lines->Strings[i];
		 if (line.LowerCase().TrimLeft().Pos("secret-keyring ") == 1)
			secring = line.SubString(16, line.Length());
		 if (line.LowerCase().TrimLeft().Pos("keyring ") == 1)
			pubring = line.SubString(9, line.Length());
	  }
	  delete lines;
   }
   return secring != "" && pubring != "";
}
//---------------------------------------------------------------------------

bool __fastcall TQSAMainWin::CheckKeyringsDefined()
{
   if (!FileExists(GPGConf()))
	  return false;

   AnsiString secring, pubring;
   return GetKeyringPathnames(GPGConf(), secring, pubring);
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::CheckFixIdeaDllInGPGConf()
{
   if (FileExists(GPGConf()) && FileExists(GPGIdeaDll()))
   {
	  AnsiString line, loadextension = "load-extension "+GPGDir()+"idea.dll";
	  TStringList *conf = new TStringList();
	  conf->LoadFromFile(GPGConf());
	  if (conf->IndexOf(loadextension) < 0)
	  {
		 int i = 0; while (i < conf->Count)
		 {
			line = conf->Strings[i].TrimLeft().LowerCase();
			if (line.Pos("load-extension ") == 1 && line.Pos("idea.dll") == line.Length()-7)
			{
			   conf->Delete(i);
			   break;
			}
			i++;
		 }
		 i = 0; while (conf->Strings[i].Pos("#") == 1)
			i++;
		 conf->Insert(i, loadextension);
		 conf->SaveToFile(GPGConf());
	  }
	  delete conf;
   }
}
//---------------------------------------------------------------------------

bool __fastcall TQSAMainWin::CheckInstallNymservers()
{
   bool ok = false;
   TStringList *servers = new TStringList();
   DB->LoadNymservers(servers);
   if (!servers->Count)
   {
	  servers->CommaText = nymservers;
	  DB->SaveNymservers(servers);
	  ok = true;
   }
   delete servers;
   return ok;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::Timer1Timer(TObject *Sender)
{
   Timer1->Enabled = false;
   GetNNTPMail(this, account, modeGetMail);
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::FileExitItemClick(TObject *Sender)
{
   Close();
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::PopupMenu1Popup(TObject *Sender)
{
   int sels = ListView->SelCount;
   Popup1DecryptItem->Enabled = sels == 1 && WantsDecrypt();
   Popup1CopyMessageToClipboard->Enabled = !Popup1DecryptItem->Enabled;
   if (Popup1CopyMessageToClipboard->Enabled)
	  Popup1CopyMessageToClipboard->Enabled = sels == 1;
   Popup1OpenItem->Enabled = sels;
   Popup1SelectallItem->Enabled = ListView->Items->Count;
   Popup1DeleteItem->Enabled = sels;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::LoadProxyBox()
{
   TIniFile *ini = new TIniFile(AppIni());
   nntpproxy = ini->ReadString("settings", "nntpproxy", "none");
   delete ini;

   TStringList *proxy = new TStringList();
   TStringList *proxies = new TStringList();
   DB->LoadProxies(proxies);

   ProxyBox->Items->Clear();
   ProxyBox->Sorted = true;
   for (int i = 0; i < proxies->Count; i++)
   {
	  ProxyBox->Items->Add(proxies->Names[i]);
   }
   ProxyBox->Sorted = false;

   ProxyBox->Items->Insert(0, "none");
   ProxyBox->ItemIndex = ProxyBox->Items->IndexOf(nntpproxy);
   CheckReportDefaultProxyMissing();

   if (ProxyBox->Text == "none")
	  ProxyBox->Font->Style = TFontStyles() << fsBold;
   else
	  ProxyBox->Font->Style = TFontStyles();

   SetCaption();
   delete proxies;
   delete proxy;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::LoadServerBox()
{
   TIniFile *ini = new TIniFile(AppIni());
   account = ini->ReadString("settings", "account", "none");
   delete ini;

   TStringList *server = new TStringList();
   TStringList *servers = new TStringList();
   DB->GetText("hosts", servers);
   ServerBox->Items->Clear();
   ServerBox->Sorted = true;
   for (int i = 0; i < servers->Count; i++)
   {
	  if (!servers->Names[i].Trim().IsEmpty())
		 ServerBox->Items->Add(servers->Names[i]);
   }
   ServerBox->Sorted = false;

   ServerBox->Items->Insert(0, "none");
   ServerBox->ItemIndex = ServerBox->Items->IndexOf(account);
   CheckReportDefaultServerMissing();

   if (ServerBox->Text == "none")
	  ServerBox->Font->Style = TFontStyles() << fsBold;
   else
	  ServerBox->Font->Style = TFontStyles();

   SetCaption();
   delete servers;
   delete server;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::CheckReportDefaultProxyMissing()
{
   if (ProxyBox->ItemIndex < 0)
   {
	  MsgBox(this,
		 "The NNTP proxy you've been using, " + nntpproxy + ", is missing! You  \n"
		 "don't currently have a proxy assigned.",
		 MT_WARN, MB_OK);
	  ProxyBox->ItemIndex = 0;
	  nntpproxy = "none";
	  TIniFile *ini = new TIniFile(AppIni());
	  ini->WriteString("settings", "nntpproxy", nntpproxy);
	  delete ini;
   }
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::CheckReportDefaultServerMissing()
{
   if (ServerBox->ItemIndex < 0)
   {
	  MsgBox(this,
		 "The NNTP host you've been using, " + account + ", is missing! You  \n"
		 "don't currently have a host assigned.",
		 MT_WARN, MB_OK);
	  ServerBox->ItemIndex = 0;
	  account = "none";
	  TIniFile *ini = new TIniFile(AppIni());
	  ini->WriteString("settings", "account", account);
	  delete ini;
   }
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::ToolsAccountsItemClick(TObject *Sender)
{
   if (ManageNNTPHosts(this))
   {
	  LoadServerBox();
	  LoadProfile(account, profile);
   }
}
//---------------------------------------------------------------------------

int __fastcall DesendSort(TStringList *list, int Index1, int Index2)
{
  if (list->Strings[Index1] < list->Strings[Index2]) return +1;
  if (list->Strings[Index1] > list->Strings[Index2]) return -1;
  return 0;
}

//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::LoadListItem(TListItem *item, const AnsiString &msgkey, TStrings *msg)
{
   item->ImageIndex = ((msgkey[bflags1] & mflRead) == mflRead) ? imgRead:imgNotRead;
   AnsiString dt = GetTok(keyDate, msgkey, DBFSc);
   //nym = GetTok(keyUserID, msgkey, DBFSc);

   item->Caption = "20" + dt.SubString(1, 2) + "-" +
	  dt.SubString(3, 2) + "-" + dt.SubString(5, 2) + " " +
	  dt.SubString(7, 2) + ":" + dt.SubString(9, 2);
   item->SubItems->Add(GetTok(keyWho, msgkey, DBFSc));
   //item->SubItems->Add(GetEAddress(nym));
   item->SubItems->Add(GetTok(keySubjectOrg, msgkey, DBFSc)); // add the orgsubject
   div_t dv = div(GetTok(keySize, msgkey, DBFSc).ToIntDef(0), 1000);
   if (dv.rem > 500 || dv.quot == 0)
	  dv.quot++;
   item->SubItems->Add(withcommas(dv.quot) + " KB");

   AnsiString subject = GetTok(keySubject, msgkey, DBFSc);
   if (subject.IsEmpty())
	  subject = GetSubject(msg);



   item->SubItems->Strings[lvSubject] = subject; // change to hash/encr subject
   if (msg->IndexOf(BeginPGPMessage) < 0)
	   subject += ":  "+GetSecondSubject(msg);

   if (subject.IsEmpty())
	  subject = GetSubject(msg);
   item->SubItems->Strings[lvSubject] = subject; // change to hash/encr subject
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::LoadListView()
{
   ListView->Items->BeginUpdate();
   mail->Clear();
   ListView->Items->Clear();
   DB->LoadMessageIndex(mail);
   mail->CustomSort(DesendSort);
   TStringList *msg = new TStringList();
   for (int i = 0; i < mail->Count; i++)
   {
	  AnsiString msgkey = mail->Strings[i];
	  TListItem *item = ListView->Items->Add();
	//if (GetTok(keySubject, msgkey, DBFSc).IsEmpty())
		 DB->LoadMessage(msgkey, msg);
	  LoadListItem(item, msgkey, msg);
	  if (msg->Count)
		 msg->Clear();
   }
   delete msg;
   ListView->Items->Count = mail->Count;
   ListView->Items->EndUpdate();
   ShowRecords(mail->Count);
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::LoadProfile(const AnsiString &name, TStrings *profile)
{
   assert(name != "");
   assert(profile != NULL);
   if (name == "none")
	  return;

   TStringList *hosts = new TStringList();
   DB->GetText("hosts", hosts);
   profile->CommaText = hosts->Values[name];
   if (!profile->Count)
   {
	  MsgBox(this, "Problem! Your previously active news account, '"+name+"', is   \n"
		"missing. Currently, you don't have an active account. You'll need to   \n"
		"select one, before you can continue.",
		MT_INFO, MB_OK);
	  account = "none";
	  TIniFile *ini = new TIniFile(AppIni());
	  ini->WriteString("settings", "account", account);
	  delete ini;
	  LoadServerBox();
   }
   delete hosts;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::SaveProfile(const AnsiString &name, TStrings *Profile)
{
   assert(name != "");
   assert(profile != NULL);
   if (name == "none")
	  return;

   TStringList *hosts = new TStringList();
   DB->GetText(Hosts, hosts);
   hosts->Values[name] = Profile->CommaText;
   DB->PutText(Hosts, hosts);
   delete hosts;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::SaveMessage(int msgno, TStrings *msg)
{
   assert(msgno > 0);
   assert(msg != NULL);
   assert(msg->Count);
   AnsiString subno, subject, sphrase, kphrase, msgkey, nym, keyid;
   TMsgManager *msgman = new TMsgManager(msg);
   subno = msgman->GetHeader("Subject-No:");
   assert(subno != "");

   TStringList *sub = new TStringList();
   DB->LoadSubject(subno, sub);
   subject = sub->Values["subject"]; //DB->GetSubjectItem(subno, "subject", "");
   keyid = sub->Values["keyid"];
   delete sub;

   nym = GetTok(0, GPGGetPublicKeyStr(keyid), ':');
   assert(nym != "");
   msgkey = msgman->key+DBFS+IntToStr(msgno)+DBFS+subject+DBFS+subno+DBFS+nym;
   DB->SaveMessage(msgkey, msg);
   delete msgman;

   ListView->Items->BeginUpdate();
   mail->InsertObject(0, msgkey, NULL);

   TListItem *item = ListView->Items->Insert(0);
   LoadListItem(item, msgkey, msg);
   ListView->Items->Item[0]->MakeVisible(false);
   ListView->Items->EndUpdate();
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::ShowStatus(const AnsiString &text)
{
   StatusBar->Panels->Items[statusStatusPanel]->Text = " " + text;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::ShowRecords(const int count)
{
   StatusBar->Panels->Items[statusRecordsPanel]->Text = " " +
	  IntToStr(count) + " messages ";
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::FileMenuClick(TObject *Sender)
{
   FileGetMailItem->Caption = Subjects->Checked ? "&Check mail" : "Get mail";
   FileGetMailItem->Enabled = !NNTPModHandle;
   FileSecuremode->Checked = DB->IsSecureMode();
   FileSavedatabase->Enabled = DB->modified;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::SubjectsClick(TObject *Sender)
{
   GetBtn->Caption = Subjects->Checked ? ChkBtnCap : GetBtnCap;
   if (Subjects->Checked)
   {
	  DSABox(this, "chkmail",
		"NOTICE: When you tick this box, you'll find the 'Get Mail' button caption "
		"changes to 'Chk Mail'. In this mode QSA will only download message subjects and "
		"check to see if any are yours. The messages waiting will be reported. It will "
		"still be necessary to download all messages to get those belonging to you.\n\n"
		"IMPORTANT: If you use this feature, don't be tempted to 'catch up' the newsgroup "
		"when you don't have mail. Its a bad idea. Your goal should be to download every "
		"message posted to a.a.m.",
		"Checking Mail", MB_OK, 0, 300);
   }
}
//---------------------------------------------------------------------------

bool __fastcall TQSAMainWin::CheckAllFilesPresent()
{
   AnsiString reinstall = "Reinstalling the program should solve the problem.";
   if (!FileExists(GPGDir()+"gpg.exe"))
   {
	  MsgBox(this,
		"Problem! The file 'gpg.exe' is missing from my home directory!   "+
		 reinstall, MT_ERROR, MB_OK);
	  return false;
   }
   if (!FileExists(GPGDir()+"iconv.dll"))
   {
	  MsgBox(this,
		"Problem! The file 'iconv.dll' is missing from my home directory!   "+
		 reinstall, MT_ERROR, MB_OK);
	  return false;
   }
   if (!FileExists(GPGDir()+"idea.dll"))
   {
	  MsgBox(this,
		"Problem! The file 'idea.dll' is missing from my home directory!   "+
		 reinstall, MT_ERROR, MB_OK);
	  return false;
   }
   if (!FileExists(AppDir()+"eraser.dll"))
   {
	  MsgBox(this,
		"Problem! The file 'eraser.dll' is missing from my home directory!   "+
		 reinstall, MT_ERROR, MB_OK);
	  return false;
   }
   if (!FileExists(AppDir()+"libeay32.dll"))
   {
	  MsgBox(this,
		"Problem! The file 'libeay32.dll' is missing from my home directory!   "+
		 reinstall, MT_ERROR, MB_OK);
	  return false;
   }
   if (!FileExists(AppDir()+"psesub32.dll"))
   {
	  MsgBox(this,
		"Problem! The file 'psesub32.dll' is missing from my home directory!   "+
		 reinstall, MT_ERROR, MB_OK);
	  return false;
   }
   if (!FileExists(AppDir()+"ssleay32.dll"))
   {
	  MsgBox(this,
		"Problem! The file 'ssleay32.dll' is missing from my home directory!   "+
		 reinstall, MT_ERROR, MB_OK);
	  return false;
   }
   if (!FileExists(AppDefsIni()))
   {
	  MsgBox(this,
		"Problem! The file '"+ExtractFileName(AppDefsIni())+
		   "' is missing from my home directory!   "+
		 reinstall, MT_ERROR, MB_OK);
	  return false;
   }
   if (FileExists(AppServersIni()))
	  DeleteFile(AppServersIni());
   //if (!FileExists(AppServersIni()))
   //{
   //	  MsgBox(this,
   //		"Problem! The file '"+ExtractFileName(AppServersIni())+
   //		   "' is missing from my home directory!   "+
   //		 reinstall, MT_ERROR, MB_OK);
   //	  return false;
   //}
   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TQSAMainWin::CheckNoSubjectsDefined()
{
   bool ok = true;
   TStringList *subs = new TStringList();
   DB->LoadSubjects(subs);
   for (int i = 0; i < subs->Count; i++)
   {
	  if (subs->Strings[i].Pos("active=1") > 0)
	  {
		 ok = false;
		 break;
	  }
   }
   delete subs;
   return ok;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::EnableToolbar(bool enable)
{
   Subjects->Enabled = enable;
   GetBtn->Enabled = enable;
   ServerBox->Enabled = enable;
   ProxyBox->Enabled = enable;
   Label1->Enabled = enable;
   Label2->Enabled = enable;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::FileGetMailItemClick(TObject *Sender)
{
   if (NNTPModHandle && GetBtn->Caption == GetBtnCap)
	  return;

   goterror = false;
   GetBtn->Enabled = false;
   Subjects->Enabled = false;
   if (GetBtn->Caption == CancelCap)
   {
	  if (NNTPModHandle)
	  {
		 PostMessage(NNTPModHandle, WM_NNTPCANCEL, 0, 0);
		 GetBtn->Font->Style = TFontStyles() << fsItalic;
		 GetBtn->Caption = GetBtnCap;
		 Snooze(500);
		 if (NNTPModHandle)
		 {
			delete NNTPThrd;
			NNTPModHandle = NULL;
			//NNTPMod = NULL;
		 }
	  }
	  EnableToolbar(true);
	  return;
   }

   if (!CheckKeyringsDefined())
   {
	  MsgBox(this,
		"You need to tell me where your PGP Keyrings are located!   \n"
		"I'll open the dialog for you.",
		 MT_INFO, MB_OK);
	  if (!ManagePGPKeyrings(this))
	  {
		 EnableToolbar(true);
		 return;
	  }
	  Snooze(300);
   }

   if (account == "none")
   {
	  MsgBox(this,
		"You don't have a NEWS account selected!   \n"
		"I'll open the dialog for you.",
		 MT_INFO, MB_OK);
	  if (!ManageNNTPHosts(this))
	  {
		 LoadServerBox();
		 LoadProfile(account, profile);
		 EnableToolbar(true);
		 return;
	  }
	  Snooze(300);
   }

   if (CheckNoSubjectsDefined())
   {
	  MsgBox(this,
		"You don't have any AAM subject lines defined!   \n"
		"I'll open the dialog for you.",
		 MT_INFO, MB_OK);
	  if (!ManageSubjectLines(this))
	  {
		 EnableToolbar(true);
		 return;
	  }
	  Snooze(300);
   }

   Snooze(300);
   GetBtn->Font->Style = TFontStyles();
   GetBtn->Caption = CancelCap;
   GetBtn->Enabled = true;
   ServerBox->Enabled = false;
   ProxyBox->Enabled = false;
   Label1->Enabled = false;
   Label2->Enabled = false;
   Status->Clear();

   bool catchup = false;
   CopyMsgBtn->Enabled = false;
   /*if (Subjects->Checked)
   {
	  MsgBox(this, "Check mail is not implemented yet.", MT_INFO, MB_OK);
	  return;
   }
   if (catchup)
   {
	  MsgBox(this, "Catchup newsgroup is not implemented yet.", MT_INFO, MB_OK);
	  return;
   }*/
   int mode = Subjects->Checked ? modeChkMail : catchup ? modeCatchUp : modeGetMail;
   GetNNTPMail(this, account, mode);
 //GetNNTPMail(this, account, Subjects->Checked, catchup);
}
//--------------------------------------------------------------------------

void __fastcall TQSAMainWin::ClearListViewSelected()
{
   for (int i = 0; i < ListView->Items->Count; i++)
	  ListView->Items->Item[i]->Selected = false;
   Message->Clear();
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::HelpAboutItemClick(TObject *Sender)
{
   MsgBox(this,
	  QuickSilverAAMVersion()+"\n"+ProgramCopyright()+"    ",
	  MT_INFO, MB_OK
   );
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::Openini1Click(TObject *Sender)
{
   EditProgramParameters(this);
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::Popup1DeleteItemClick(TObject *Sender)
{
   AnsiString msg = "Delete the currently selected message";
   msg += ListView->SelCount > 1 ? "s. Are you sure?   " : ". Are you sure?  ";

   if (MsgBox(this, msg, MT_CONFIRM, MB_OKCANCEL) == IDOK)
   {
	  DeleteSelectedMessages();
   }
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::DeleteSelectedMessages()
{
   Message->Clear();
   int i = ListView->Items->Count - 1;
   FewerHeaders->OnClick = NULL;
   while (i >= 0)
   {
	  TListItem *item = ListView->Items->Item[i];
	  if (item->Selected)
	  {
		 item->Selected = false;
		 DB->DelMessage(mail->Strings[item->Index]);
		 delete mail->Objects[item->Index];
		 mail->Delete(item->Index);
		 item->Delete();
		 ShowRecords(mail->Count);
		 needspack = true;
	  }
	  i--;
   }
   FewerHeaders->OnClick = FewerHeadersClick;
   DecryptBtn->Enabled = false;
   CopyMsgBtn->Enabled = false;
   DeleteBtn->Enabled = false;
   ListView->Invalidate();
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::Popup2CopyItemClick(TObject *Sender)
{
   if (IsActiveTab(MessageTab))
   {
	  assert(Message->SelLength);
	  Message->CopyToClipboard();
   }
   if (IsActiveTab(StatusTab))
   {
	  assert(Status->SelLength);
	  Status->CopyToClipboard();
   }
   if (IsActiveTab(LogTab))
   {
	  assert(Log->SelLength);
	  Log->CopyToClipboard();
   }
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::PopupMenu2Popup(TObject *Sender)
{
   if (IsActiveTab(MessageTab))
   {
	  int sels = ListView->SelCount;
	  Popup2DecryptmessageItem->Enabled = NeedsDecryptVerify(Message->Lines);
	  Popup2CopyMessageToClipboard->Enabled = !Popup2DecryptmessageItem->Enabled;
	  if (Popup2CopyMessageToClipboard->Enabled)
		 Popup2CopyMessageToClipboard->Enabled = sels == 1;
	  Popup2DeletemessageItem->Enabled = Message->Lines->Count;
	  Popup2CopyItem->Enabled = Message->SelLength > 0;
	  Popup2SelectallItem->Enabled = Message->Text.Length() > 0;
	  return;
   }
   if (IsActiveTab(StatusTab))
   {
	  Popup2CopyMessageToClipboard->Enabled = false;
	  Popup2DecryptmessageItem->Enabled = false;
	  Popup2DeletemessageItem->Enabled = false;
	  Popup2CopyItem->Enabled = Status->SelLength > 0;
	  Popup2SelectallItem->Enabled = Status->Text.Length() > 0;
	  return;
   }
   if (IsActiveTab(LogTab))
   {
	  Popup2CopyMessageToClipboard->Enabled = false;
	  Popup2DecryptmessageItem->Enabled = false;
	  Popup2DeletemessageItem->Enabled = false;
	  Popup2CopyItem->Enabled = Log->SelLength > 0;
	  Popup2SelectallItem->Enabled = Log->Text.Length() > 0;
	  return;
   }
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::Popup2SelectallItemClick(TObject *Sender)
{
   if (IsActiveTab(MessageTab))
   {
	  Message->SelectAll();
   }
   if (IsActiveTab(StatusTab))
   {
	  Status->SelectAll();
   }
   if (IsActiveTab(LogTab))
   {
	  Log->SelectAll();
   }
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::WMLogDisplay(TMessage &Msg)
{
   if (Msg.WParam == CLEARLOG)
   {
	  Status->Lines->Clear();
	  return;
   }

   char *cstr = (char*)Msg.LParam;
   Status->Lines->Add(cstr);
   delete [] cstr;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::WMShowStatus(TMessage &Msg)
{
   char *cstr = (char*)Msg.LParam;
   ShowStatus(cstr);
   if (!Msg.WParam) // NNTPMod passes strings that need to be deleted
	  delete [] cstr;  // cfgmsgdlgs and messagepipes don't delete
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::WMShowProgress(TMessage &Msg)
{
   char *cstr = (char*)Msg.LParam;
   StatusBar->Panels->Items[statusProgressPanel]->Text = cstr;
   delete [] cstr;

   progresspcnt = Msg.WParam;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::WMMessageBox(TMessage &Msg)
{
   AnsiString msgtext, params;
   char *cstr = (char*)Msg.LParam;
   params = cstr;
   delete [] cstr;

   TMsgBoxTypes boxtype = (TMsgBoxTypes)GetTok(1, params, '|').ToIntDef(0);
   UINT buttons = GetTok(2, params, '|').ToIntDef(0);
   msgtext = GetTok(0, params, '|');
   Msg.Result = MsgBox(this, msgtext, boxtype, buttons);
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::WMNNTPTimeoutTimer(TMessage &Msg)
{
   NNTPTimeoutTimer->Enabled = Msg.WParam;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::NNTPTimeoutTimerTimer(TObject *Sender)
{
   //check and save timeouts
   TIniFile *ini = new TIniFile(AppIni());
   if (ini->ReadBool("settings", "savetimeouts", false))
   {
	  TStringList *timeouts = new TStringList();
	  AnsiString timeoutsfile = AppDataDir()+"timeouts.txt";
	  if (FileExists(timeoutsfile))
		 timeouts->LoadFromFile(timeoutsfile);
	  timeouts->Add(FormatDateTime("yyyy-mm-dd hh:nn:ss", Now()));
	  timeouts->SaveToFile(timeoutsfile);
	  delete timeouts;
   }
   delete ini;

   PostMessage(NNTPModHandle, WM_NNTPTIMEOUTTIMER, TIMERTIMEOUT, 0);
   Snooze(1000);
   //if (NNTPThrd)
   //{
	  delete NNTPThrd;
	  Snooze(1000);
	  if (NNTPThrdHandle)
		 TerminateThread((void*)NNTPThrdHandle, 9999);
	  NNTPThrdHandle = NULL;
	  NNTPModHandle = NULL;
	  NNTPThrd = NULL;
	  if (!Status->Lines->Text.Pos("OPERATION TIMED OUT!"))
	  {
		 Status->Lines->Add("OPERATION TIMED OUT!");
		 if (Status->Lines->Text.Pos("Connection closed!") == 0)
			Status->Lines->Add("Connection closed!");
		 Status->Lines->Add("Failed!");
	  }
   //}
   StatusBar->Panels->Items[1]->Text = "TIMED OUT";
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::WMSaveDatabase(TMessage &Msg)
{
   FileSavedatabase->Click();
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::WMDelNNTPThread(TMessage &Msg)
{
   Snooze(500);
   delete NNTPThrd;
   NNTPThrdHandle = 0;
   NNTPThrd = NULL;
}
//---------------------------------------------------------------------------

//void __fastcall TQSAMainWin::WMSaveProfile(TMessage& Msg)
//{
//   SaveProfile(account, profile);
//}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::WMSaveMessage(TMessage& Msg)
{
   char *cstr = (char*)Msg.LParam;
   TStringList *message = new TStringList();
   message->SetText(cstr);
   assert(message->Count);
   delete [] cstr;

   //int mails = Msg.WParam;
   int curmsg = message->Strings[0].ToIntDef(-1);
   assert(curmsg > 0);
   message->Delete(0);
   SaveMessage(curmsg, message);
   delete message;

   ShowRecords(mail->Count);
   if (trayed)
   {
	  traymails++;
	  AnsiString emails = traymails == 1 ? "email" : "emails";
	  TrayIcon_Modify(envtrayicon, "You have "+IntToStr(traymails)+" new "+emails);
   }
}
//---------------------------------------------------------------------------

int __fastcall TQSAMainWin::FindMessageKey(const AnsiString &subkey)
{
   int idx = -1;
   for (int i = 0; i < mail->Count; i++)
   {
	  if (mail->Strings[i].Pos(subkey) == 1)
	  {
		 idx = i;
		 break;
	  }
   }
   return idx;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::WMFewerHeaders(TMessage &Msg)
{
   FewerHeaders->Checked = Msg.WParam;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::WMSetMessage(TMessage& Msg)
{
   TStrings *objmsg = (TStrings*)Msg.LParam;
   assert(objmsg && objmsg->Count);

   AnsiString from, subject, msgkey;
   msgkey = objmsg->Strings[0].SubString(1,15); objmsg->Delete(0);
   subject = objmsg->Strings[0]; objmsg->Delete(0);
   from = objmsg->Strings[0]; objmsg->Delete(0);
   int idx = FindMessageKey(msgkey);
   assert(idx >= 0);

   ListView->Items->Item[idx]->SubItems->Strings[lvSubject] =
	  ListView->Items->Item[idx]->SubItems->Strings[lvSubject] + ": " + subject;
   ListView->Items->Item[idx]->SubItems->Strings[lvWho] = from;

   if (ListView->SelCount == 1 && ListView->Items->Item[idx]->Selected)
   {
	  assert(objmsg == ObjMessage(idx));
	  TStringList *msg = new TStringList();
	  msg->AddStrings(objmsg);
	  if (FewerHeaders->Checked)
		 CompactMessageHeaders(msg);

	  Message->Clear();
	  Snooze(100);
	  Message->Lines->BeginUpdate();
	  Message->Lines->AddStrings(msg);
	  Message->SelStart = 0;
	  Message->Perform(EM_SCROLLCARET, 0, 0);
	  Snooze(100);
	  Message->Lines->EndUpdate();
	  delete msg;
   }
   if (savedecrypted)
	  SaveMessageDecrypted(objmsg);

   ListView->Invalidate();
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::SaveMessageDecrypted(TStrings *msg)
{
   assert(ListView->Selected != NULL);
   TListItem *item = ListView->Selected;
   curmsgkey = mail->Strings[item->Index];
   AnsiString newkey = curmsgkey, from = item->SubItems->Strings[0];
   newkey = SetTok(3, newkey, DBFSc, from);
   assert(newkey != curmsgkey);
   DB->SaveMessage(newkey, msg);
   DB->DelMessage(curmsgkey);
   mail->Strings[item->Index] = newkey;
   curmsgkey = newkey;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::WMDelMessage(TMessage& Msg)
{
   AnsiString msgkey = (char*)Msg.LParam;
   int idx = FindMessageKey(msgkey.SubString(1, 15));
   assert(idx >= 0);
   msgkey = mail->Strings[idx];
   DB->DelItem("inbox"+DBPS+msgkey);
   ListView->Items->Item[idx]->Delete();
   mail->Delete(idx);
   ListView->Invalidate();
   if (msgkey == curmsgkey)
	  Message->Clear();
   needspack = true;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::WMSetSubject(TMessage& Msg)
{
   AnsiString from, subject, msgkey, keystr = (char*)Msg.LParam;
   msgkey = GetTok(1, keystr, '\254');
   keystr = GetTok(0, keystr, '\254');
   int idx = mail->IndexOf(msgkey);
   assert(idx >= 0);
   TListItem *item = ListView->Items->Item[idx];
   item->SubItems->Strings[lvWho] = GetTok(1, keystr, DBFSc);
   item->SubItems->Strings[lvSubject] = item->SubItems->Strings[lvSubject]+": "+
	  GetTok(0, keystr, DBFSc);
   ListView->Invalidate();
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::WMUpdateComplete(TMessage &Msg)
{
   GetBtn->Caption = " Get Mail  ";
   GetBtn->Font->Style = TFontStyles() << fsItalic;
   //StatusBar->Panels->Items[2]->Text = "";
   ShowRecords(mail->Count);
   if (Subjects->Checked)
	  Subjects->Checked = false;
   Subjects->Enabled = true;
   GetBtn->Enabled = true;
   ServerBox->Enabled = true;
   ProxyBox->Enabled = true;
   Label1->Enabled = true;
   Label2->Enabled = true;
   CopyMsgBtn->Enabled = ListView->Selected; 
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::WMAccountChange(TMessage &Msg)
{
   bool active = true;
   AnsiString oldname, newname, names = (char*)Msg.LParam;
   oldname = GetTok(0, names, '\255');
   newname = GetTok(1, names, '\255');
   if (ServerBox->Text == oldname)
   {
	  ServerBox->ItemIndex = -1;
	  active = true;
   }
   int idx = ServerBox->Items->IndexOf(oldname);
   assert(idx >= 0);
   ServerBox->Items->Strings[idx] = newname;
   if (active)
   {
	  ServerBox->ItemIndex = idx;
	  account = newname;
   }
}
//---------------------------------------------------------------------------

int __fastcall TQSAMainWin::DownloadPercentComplete(WORD next, DWORD total)
{
   if (total > 0)
	  return (int)((next / total) * 100);
   return (int)next;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::Timer4Timer(TObject *Sender)
{
   AnsiString text = StatusBar->Panels->Items[statusProgressPanel]->Text;
   flashcount++;
   if (flashcount > 20)
   {
	  StatusBar->Panels->Items[statusProgressPanel]->Text = text.Trim();
	  StatusBar->Refresh();
	  Timer4->Enabled = false;
	  return;
   }
   text = text.Pos(" ") == 1 ? text.Trim() + "  " : "  " + text.Trim();
   StatusBar->Panels->Items[statusProgressPanel]->Text = text;
   StatusBar->Refresh();
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::WMProgressError(TMessage &Msg)
{
   char *cstr = (char*)Msg.LParam;
   error->Clear();
   error->SetText(cstr);
   delete [] cstr;
   goterror = true;
   flashcount = 0;

   StatusBar->Panels->Items[statusProgressPanel]->Text = "ERROR! Click here!";
   StatusBar->Refresh();

   Timer4->Enabled = true;
   MessageBeep(MB_ICONHAND);
   if (trayed)
   {
	  TrayIcon_Modify(errtrayicon, "Sorry, FATAL error!");
   }
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::StatusBarMouseDown(TObject *Sender,
	  TMouseButton Button, TShiftState Shift, int X, int Y)
{
   if (goterror) //StatusBar->Panels->Items[statusProgressPanel]->Text.LowerCase().Pos("error") > 0)
   {
	  int x = StatusBar->Panels->Items[statusStatusPanel]->Width;
	  int w = StatusBar->Panels->Items[statusProgressPanel]->Width;
	  if (X > x && X < x+w)
	  {
		 StatusBar->Panels->Items[statusProgressPanel]->Text = "";
		 StatusBar->Invalidate();
		 ShowNNTPError(this, error);
		 goterror = false;
	  }
   }
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::StatusBarDrawPanel(TStatusBar *StatusBar,
	  TStatusPanel *Panel, const TRect &Rect)
{
   TCanvas *canvas = StatusBar->Canvas;
   if (Panel == StatusBar->Panels->Items[statusStatusPanel])
   {
	  canvas->TextOut(Rect.Left, Rect.Top-1, Panel->Text);
   }

   if (Panel == StatusBar->Panels->Items[statusProgressPanel])
   {
	  AnsiString text = StatusBar->Panels->Items[statusProgressPanel]->Text;
	  TRect orgrect = Rect, rect = Rect;
	  orgrect.Bottom -= 1;
	  orgrect.Right -= 2;
	  rect = orgrect;
	  TColor sbColor = canvas->Brush->Color;
	  bool iserror = text.LowerCase().Pos("error");
	  int pcnt = (int)(((iserror ? 100 : progresspcnt)*.01) * (Rect.Right - Rect.Left));
	  if (pcnt > 0)
	  {
		 rect.Right = rect.Left + (int)pcnt;
		 canvas->Font->Color = iserror ? clWhite : clWindowText;
		 canvas->Brush->Color = iserror ? clRed : clPBar;
		 canvas->FillRect(rect);
		 canvas->Brush->Color = sbColor;
		 orgrect.Left = rect.Right;
	  }
	  canvas->FillRect(orgrect);
	  canvas->Brush->Style = bsClear;
	  int w = Rect.Right - Rect.Left, tw = canvas->TextWidth(text);
	  canvas->TextOut(Rect.Left+(w/2)-(tw/2), Rect.Top-1, text);
   }

   if (Panel == StatusBar->Panels->Items[statusRecordsPanel])
   {
	  canvas->TextOut(Rect.Left, Rect.Top-1, Panel->Text);
   }
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::SpeedButton1Click(TObject *Sender)
{
   TStringList *msg = new TStringList();
   TStringList *msgs = new TStringList();
   DB->LoadMessageIndex(msgs);
   for (int i = msgs->Count - 1; i > 0; i--)
   {
	  DB->LoadMessage(msgs->Strings[i], msg);
   }
   delete msgs;
   delete msg;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::MessageSelectionChange(TObject *Sender)
{
   CopyBtn->Enabled = Message->SelLength > 0;
   if (Message->Text.IsEmpty())
   {
	  ClearListViewSelected();
	  DecryptBtn->Enabled = false;
      CopyMsgBtn->Enabled = false;
	  DeleteBtn->Enabled = false;
   }
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::StatusSelectionChange(TObject *Sender)
{
   CopyBtn->Enabled = Status->SelLength > 0;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::EditMenuClick(TObject *Sender)
{
   int sels = ListView->SelCount;
   if (ActiveControl == ListView)
   {
	  Copymessagetoclipboard1->Enabled = sels == 1 && Message->Text.Length();
	  EditCopyItem->Enabled = false;
	  EditSelectallItem->Enabled = mail->Count;
   }
   else
   {
	  if (IsActiveTab(MessageTab))
	  {
		 Copymessagetoclipboard1->Enabled = sels == 1 && Message->Text.Length();
		 EditCopyItem->Enabled = Message->SelLength > 0;
		 EditSelectallItem->Enabled = Message->Text.Length() > 0;
		 return;
	  }
	  if (IsActiveTab(StatusTab))
	  {
		 Copymessagetoclipboard1->Enabled = false;
		 EditCopyItem->Enabled = Status->SelLength > 0;
		 EditSelectallItem->Enabled = Status->Text.Length() > 0;
	  }
	  if (IsActiveTab(LogTab))
	  {
		 Copymessagetoclipboard1->Enabled = false;
		 EditCopyItem->Enabled = Log->SelLength > 0;
		 EditSelectallItem->Enabled = Log->Text.Length() > 0;
	  }
   }
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::EditCopyItemClick(TObject *Sender)
{
   if (IsActiveTab(MessageTab))
   {
	  Message->CopyToClipboard();
   }
   if (IsActiveTab(StatusTab))
   {
	  Status->CopyToClipboard();
   }
   if (IsActiveTab(LogTab))
   {
	  Log->CopyToClipboard();
   }
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::EditSelectallItemClick(TObject *Sender)
{
   if (ActiveControl == ListView)
   {
	  ListView->SelectAll();
	  Message->Clear();
	  return;
   }
   if (IsActiveTab(MessageTab))
   {
	  Message->SelectAll();
   }
   if (IsActiveTab(StatusTab))
   {
	  Status->SelectAll();
   }
   if (IsActiveTab(LogTab))
   {
	  Log->SelectAll();
   }
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::Popup1SelectallItemClick(TObject *Sender)
{
   ListView->SelectAll();
   Message->Clear();
}
//---------------------------------------------------------------------------

bool __fastcall TQSAMainWin::WantsDecrypt()
{
   bool wants = false;
   for (int i = 0; i < ListView->Items->Count; i++)
   {
	  TListItem *item = ListView->Items->Item[i];
	  TStrings *objmsg = (TStrings*)mail->Objects[item->Index];// ObjMessage can't handle NULL
	  if (item->Selected && NeedsDecryptVerify(objmsg))
	  {
		 wants = true;
		 break;
	  }
   }
   return wants;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::ToolsMenuClick(TObject *Sender)
{
   Composenewmessage1->Visible = FileExists(AppDir()+"qsl.exe");
   Catchupnewsgroup1->Enabled = ServerBox->Text != "none" && !NNTPModHandle;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::ToolsProxiesItemClick(TObject *Sender)
{
   if (ManageProxies(this))
   {
	  LoadProxyBox();
   }
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::ToolsSubjectsItemClick(TObject *Sender)
{
   ManageSubjectLines(this);
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::Catchupnewsgroup1Click(TObject *Sender)
{
   if (MsgBox(this,
	  "Catching up the newsgroup skips over all messages from the last   \n"
	  "time you downloaded mail to the current last article on the   \n"
	  "server. It's a convenience you can use on occassion, but don't   \n"
	  "over-do it. Your goal is to download every message from a.a.m.   \n\n"
	  "Catch up the newsgroup now?",
	   MT_CONFIRM, MB_OKCANCEL) == IDCANCEL)
		  return;

   Status->Clear();
   //bool catchup = true;
 //GetNNTPMailThrd(this, profile, Status->Lines, Subjects->Checked, catchup, NULL);
 //GetNNTPMail(this, account, Subjects->Checked, catchup);
   GetNNTPMail(this, account, modeCatchUp);
   Subjects->Checked = false;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::ServerBoxSelect(TObject *Sender)
{
   account = ServerBox->Text;
   LoadProfile(account, profile);
   TIniFile *ini = new TIniFile(AppIni());
   ini->WriteString("settings", "account", account);
   delete ini;
   SetCaption();
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::ProxyBoxSelect(TObject *Sender)
{
   nntpproxy = ProxyBox->Text;
   TIniFile *ini = new TIniFile(AppIni());
   ini->WriteString("settings", "nntpproxy", nntpproxy);
   delete ini;
   SetCaption();
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::ComboBoxDropDown(TObject *Sender)
{
   TComboBox *box = dynamic_cast<TComboBox*>(Sender);
   box->Font->Style = TFontStyles();
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::ComboBoxCloseUp(TObject *Sender)
{
   TComboBox *box = dynamic_cast<TComboBox*>(Sender);
   if (box->Text == "none")
	 box->Font->Style = TFontStyles() << fsBold;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::ToolsDeletemessageItemClick(TObject *Sender)
{
   if (!ListView->Items->Count)
   {
	  MsgBox(this, "Nothing to delete!", MT_INFO, MB_OK);
	  return;
   }

   int sels = ListView->SelCount;
   if (!sels)
   {
	  if (MsgBox(this,
		"Delete ALL messages in the database?   ",
		 MT_CONFIRM, MB_OKCANCEL) == IDCANCEL)
			return;
	  if (MsgBox(this,
		"Delete ALL messages. Are you sure?   ",
		 MT_CONFIRM, MB_OKCANCEL) == IDCANCEL)
			return;

	  ListView->SelectAll();
	  DeleteSelectedMessages();
	  return;
   }

   AnsiString msg = "Delete the currently selected message";
   msg += sels > 1 ? "s. Are you sure?   " : ". Are you sure?   ";

   if (MsgBox(this, msg, MT_CONFIRM, MB_OKCANCEL) == IDOK)
   {
	  DeleteSelectedMessages();
   }
}
//---------------------------------------------------------------------------

TStrings* __fastcall TQSAMainWin::ObjMessage(int index)
{
   assert(index >= 0);
   assert(index < ListView->Items->Count);
   assert(mail->Objects[index] != NULL);
   return (TStrings*)mail->Objects[index];
}
//---------------------------------------------------------------------------

bool __fastcall TQSAMainWin::CheckFixMessageNotSelected()
{
   assert(curmsgkey != "");
   // the selected item got inadvertantly unselected
   if (!ListView->SelCount)
   {
	  //AnsiString msgkey = curmsgkey;
	  //if (msgkey.IsEmpty())
	  //	 msgkey = prvmsgkey;
	  int idx = mail->IndexOf(curmsgkey);
	  assert(idx >= 0);
	  //for (int i = 0; i < mail->Count; i++)
	  //{
	  //	 if (mail->Strings[i] == curmsgkey)
	  //	 {
	  ListView->Items->Item[idx]->Selected = true;
	  ListView->Selected = ListView->Items->Item[idx];
	  //		break;
	  //	 }
	  //}
   }
   if (!ListView->Selected)
   {
	  MessageBeep(0);
	  return false;
   }
   return true;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::ToolsDecryptmessageItemClick(TObject *Sender)
{
   if (!CheckFixMessageNotSelected())
	  return;

   assert(ListView->SelCount == 1);
   if (!WantsDecrypt())
   {
	  MsgBox(this, "Nothing to decrypt!", MT_INFO, MB_OK);
	  return;
   }

   TStringList *msg = new TStringList();
   assert(ListView->Selected);
   TListItem *item = ListView->Selected;
   AnsiString msgkey = mail->Strings[item->Index];
   TStrings *objmsg = ObjMessage(item->Index);

   assert(msgkey == curmsgkey);
   assert(objmsg != NULL);

   if (GPGHasPGPSignatureVerification(objmsg))
   {
	  if (DecryptMessage(this, objmsg))
	  {
		 msg->Clear();
		 msg->AddStrings(objmsg);
		 if (FewerHeaders->Checked)
			CompactMessageHeaders(msg);

		 Message->Clear();
		 Message->Lines->BeginUpdate();
		 Message->Lines->AddStrings(msg);
		 Message->SelStart = 0;
		 Message->Perform(EM_SCROLLCARET, 0, 0);
		 Snooze(100);
		 Message->Lines->EndUpdate();
		 DecryptBtn->Enabled = NeedsDecryptVerify(objmsg);
		 CopyMsgBtn->Enabled = !DecryptBtn->Enabled;
		 ListView->Invalidate();

		 if (savedecrypted)
			SaveMessageDecrypted(objmsg);

		 ActiveControl = Message;
		 delete msg;
		 return;
	  }
	  return;
   }

   objmsg->Insert(0, GetTok(keySubNo, msgkey, DBFSc));
   AnsiString hdr = GetNymAndTypeHeader(msgkey);
   if (hdr.Pos("typeerror") || hdr.Pos("keyerror"))
	  return;
   objmsg->Insert(1, hdr);

   if (DecryptNymMessage(this, objmsg))
   {
	  AnsiString subject = GetSubject(objmsg), from = GetFrom(objmsg);
	  item->SubItems->Strings[lvSubject] = item->SubItems->Strings[lvSubject] +
		 ":  " + GetSecondSubject(objmsg);
	  if (!from.IsEmpty())
		 item->SubItems->Strings[lvWho] = from;

	  msg->Clear();
	  msg->AddStrings(objmsg);
	  if (FewerHeaders->Checked)
		 CompactMessageHeaders(msg);

	  Message->Clear();
	  Message->Lines->BeginUpdate();
	  Message->Lines->AddStrings(msg);
	  Message->SelStart = 0;
	  Message->Perform(EM_SCROLLCARET, 0, 0);
	  Snooze(100);
	  Message->Lines->EndUpdate();
	  DecryptBtn->Enabled =  NeedsDecryptVerify(objmsg);
	  CopyMsgBtn->Enabled = !DecryptBtn->Enabled;
	  if (GPGHasPGPSignatureVerification(objmsg))
		 MarkMessageRead(item->Index);
	  CopyMsgBtn->Enabled = true;
	  ListView->Invalidate();
	  ActiveControl = Message;

	  if (GPGIsPGPMessage(objmsg))
	  {
		 if (!GPGIsPGPSignedMessage(objmsg))
		 {
			Snooze(500);
			MsgBox(this,
			  "Nym message successfuly decrypted!   \n\n"
			  "I see there's another layer of encryption. That layer must     \n"
			  "have been added by the sender.",
			   MT_INFO, MB_OK);
		 }
		 if (DecryptMessage(this, objmsg))
		 {
			msg->Clear();
			msg->AddStrings(objmsg);
			if (FewerHeaders->Checked)
			   CompactMessageHeaders(msg);

			Message->Clear();
			Message->Lines->BeginUpdate();
			Message->Lines->AddStrings(msg);
			Message->SelStart = 0;
			Message->Perform(EM_SCROLLCARET, 0, 0);
			Snooze(100);
			Message->Lines->EndUpdate();

		 }
		 DecryptBtn->Enabled = NeedsDecryptVerify(objmsg);
		 CopyMsgBtn->Enabled = !DecryptBtn->Enabled;
		 if (GPGHasPGPSignatureVerification(objmsg))
			MarkMessageRead(item->Index);
		 CopyMsgBtn->Enabled = true;
		 ListView->Invalidate();
		 ActiveControl = Message;
	  }

	  if (savedecrypted)
		 SaveMessageDecrypted(objmsg);
   }
   else
   {
	  DecryptBtn->Enabled = NeedsDecryptVerify(objmsg);
	  CopyMsgBtn->Enabled = !DecryptBtn->Enabled;
	  if (GPGHasPGPSignatureVerification(objmsg))
		 MarkMessageRead(item->Index);
	  CopyMsgBtn->Enabled = true;
	  ListView->Invalidate();
	  ActiveControl = Message;
   }

   delete msg;
}
//---------------------------------------------------------------------------

/*void __fastcall TQSAMainWin::ToolsDecryptmessageItemClick(TObject *Sender)
{
   if (!CheckFixMessageNotSelected())
	  return;

   assert(ListView->SelCount == 1);
   if (!WantsDecrypt())
   {
	  MsgBox(this, "Nothing to decrypt!", MT_INFO, MB_OK);
	  return;
   }

   assert(ListView->Selected);
   TListItem *item = ListView->Selected;
   AnsiString msgkey = mail->Strings[item->Index];
   TStrings *msg = ObjMessage(item->Index);
   assert(msgkey == curmsgkey);
   assert(msg != NULL);

   if (GPGHasPGPSignatureVerification(msg))
   {
	  DecryptMessage(this, msg);
	  fullmsg->Clear();
	  fullmsg->AddStrings(msg);
	  if (FewerHeaders->Checked)
		 CompactMessageHeaders(msg);

	  Message->Clear();
	  Message->Lines->BeginUpdate();
	  Message->Lines->AddStrings(msg);
	  Message->SelStart = 0;
	  Message->Perform(EM_SCROLLCARET, 0, 0);
	  Snooze(100);
	  Message->Lines->EndUpdate();
	  DecryptBtn->Enabled = NeedsDecryptVerify(ObjMessage(item->Index));
	  CopyMsgBtn->Enabled = !DecryptBtn->Enabled;
	  ListView->Invalidate();
	  ActiveControl = Message;
	  return;
   }

   msg->Insert(0, GetTok(keySubNo, msgkey, DBFSc));
   AnsiString hdr = GetNymAndTypeHeader(msgkey);
   if (hdr.Pos("typeerror") || hdr.Pos("keyerror"))
	  return;
   msg->Insert(1, hdr);
   if (DecryptNymMessage(this, msg))
   {
	  fullmsg->Clear();
	  fullmsg->AddStrings(msg);
	  if (FewerHeaders->Checked)
		 CompactMessageHeaders(msg);

	  AnsiString subject = GetSubject(msg), from = GetFrom(msg);
	  item->SubItems->Strings[lvSubject] = item->SubItems->Strings[lvSubject] +
		 ":  " + GetSecondSubject(msg);
	  //item->SubItems->Strings[lvSubject] = GetSecondSubject(msg);
	  if (!from.IsEmpty())
		 item->SubItems->Strings[lvWho] = from;
	  Message->Clear();
	  Message->Lines->BeginUpdate();
	  Message->Lines->AddStrings(msg);
	  Message->SelStart = 0;
	  Message->Perform(EM_SCROLLCARET, 0, 0);
	  Snooze(100);
	  Message->Lines->EndUpdate();
	  DecryptBtn->Enabled =  NeedsDecryptVerify(ObjMessage(item->Index));
	  CopyMsgBtn->Enabled = !DecryptBtn->Enabled;
	  if (GPGHasPGPSignatureVerification(msg))
		 MarkMessageRead(item->Index);
	  CopyMsgBtn->Enabled = true;
	  ListView->Invalidate();
	  ActiveControl = Message;

	  if (GPGIsPGPMessage(msg))
	  {

		 if (!GPGIsPGPSignedMessage(msg))
		 {
			Snooze(500);
			MsgBox(this,
			  "Nym message successfuly decrypted!   \n\n"
			  "I see there's another layer of encryption. That layer must     \n"
			  "have been added by the sender.",
			   MT_INFO, MB_OK);
		 }
		 if (DecryptMessage(this, msg))
		 {
			fullmsg->Clear();
			fullmsg->AddStrings(msg);
			if (FewerHeaders->Checked)
			   CompactMessageHeaders(msg);

			Message->Clear();
			Message->Lines->BeginUpdate();
			Message->Lines->AddStrings(msg);
			Message->SelStart = 0;
			Message->Perform(EM_SCROLLCARET, 0, 0);
			Snooze(100);
			Message->Lines->EndUpdate();
		 }
		 DecryptBtn->Enabled = NeedsDecryptVerify(ObjMessage(item->Index));
		 CopyMsgBtn->Enabled = !DecryptBtn->Enabled;
		 if (GPGHasPGPSignatureVerification(msg))
			MarkMessageRead(item->Index);
		 CopyMsgBtn->Enabled = true;
		 ListView->Invalidate();
		 ActiveControl = Message;
	  }
   }
   else
   {
	  Message->Clear();
	  Message->Lines->BeginUpdate();
	  Message->Lines->AddStrings(msg);
	  Message->SelStart = 0;
	  Message->Perform(EM_SCROLLCARET, 0, 0);
	  Snooze(100);
	  Message->Lines->EndUpdate();
	  DecryptBtn->Enabled = NeedsDecryptVerify(ObjMessage(item->Index));
	  CopyMsgBtn->Enabled = !DecryptBtn->Enabled;
	  if (GPGHasPGPSignatureVerification(msg))
		 MarkMessageRead(item->Index);
	  CopyMsgBtn->Enabled = true;
	  ListView->Invalidate();
	  ActiveControl = Message;
   }
   if (savedecrypted)
	  SaveMessageDecrypted(fullmsg);//Message->Lines);
}*/
//---------------------------------------------------------------------------

AnsiString __fastcall TQSAMainWin::GetSubject(TStrings *msg)
{
   TMsgManager *msgman = new TMsgManager(msg);
   AnsiString subject = msgman->GetHeader("Hash-Subject:");
   if (subject.IsEmpty())
   {
	  subject = msgman->GetHeader("Encrypt-Subject:");
	  if (!subject.IsEmpty())
		 msgman->DelHeader("Encrypt-Subject:");
   }
   else
	  msgman->DelHeader("Hash-Subject:");
   delete msgman;
   return subject;
}
//---------------------------------------------------------------------------

AnsiString __fastcall TQSAMainWin::GetSecondSubject(TStrings *msg)
{
   int c = 0;
   AnsiString sub;
   for (int i = 0; i < msg->Count; i++)
   {
	  AnsiString line = msg->Strings[i];
	  if (line.Trim().IsEmpty())
	  {
		 c++;
		 if (c > 2)
			break;
	  }
	  if (line.Pos("Subject:") == 1)
		 sub = line.SubString(9, line.Length());
   }
   return sub.Trim();

}
//---------------------------------------------------------------------------

AnsiString __fastcall TQSAMainWin::GetFrom(TStrings *msg)
{
   int x = 0;
   AnsiString from;
   for (int i = 0; i < msg->Count; i++)
   {
	  AnsiString line = msg->Strings[i];
	  if (line.Pos("From:") == 1)
	  {
		 from = GetEAddress(line);
		 if (++x == 2)
			break;
	  }
   }
   return from;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::FormResize(TObject *Sender)
{
   SetupStatusBar();
   SetEditRect(Message);
   SetEditRect(Status);
   SetEditRect(Log);
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::Options1Click(TObject *Sender)
{
   if (ManageProgramOptions(this))
   {
	  LoadIni(false);
	  if (checkmail && !CheckMailTimer->Enabled)
		 StartTimer->Enabled = true;
	  if (!checkmail && CheckMailTimer->Enabled)
		 CheckMailTimer->Enabled = false;
   }
}
//---------------------------------------------------------------------------

bool __fastcall TQSAMainWin::IsMessageRead(const AnsiString &msgkey)
{
   return ((msgkey[bflags1] & mflRead) == mflRead);
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::WMMarkMessageRead(TMessage &Msg)
{
   AnsiString msgkey = (char*)Msg.LParam;
   assert(msgkey != "");
   int idx = FindMessageKey(msgkey.SubString(1, 15));
   assert(idx >= 0);
   MarkMessageRead(idx);
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::MarkMessageRead(const int index)
{
   AnsiString newkey, oldkey;
   assert(index >= 0);
   oldkey = mail->Strings[index];
   newkey = oldkey;
   newkey[bflags1] |= mflRead;
   DB->SwapKey("inbox"+DBPS+oldkey, newkey);
   ListView->Items->Item[index]->ImageIndex = imgRead;
   mail->Strings[index] = newkey;
   curmsgkey = newkey;
   ListView->Invalidate();
}
//---------------------------------------------------------------------------

bool __fastcall CheckNotifyQSLClipboardOrCopyData(TForm *Owner, TStrings *msg)
{
   TIniFile *ini = new TIniFile(AppIni());
   bool notify = ini->ReadBool("settings", "notifyqsl", true);
   delete ini;
   if (!notify)
   {
	  Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"Message NOT copied...");
	  return false;
   }
   HWND hWnd = NULL;
   hWnd = FindWindow("TQS2MainWin", NULL);
   if (!hWnd)
	  hWnd = FindWindow("TQSLMainWin", NULL);
   if (!hWnd)
   {
	  Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"Message copied to clipboard...");
	  Clipboard()->AsText = msg->Text;
	  return true;
   }

   int len = msg->Text.Length();
   COPYDATASTRUCT cdata;
   char *text = new char[len + 1];
   strncpy(text, msg->Text.c_str(), len);
   cdata.dwData = QSCOPYDATA;
   cdata.cbData = len;
   cdata.lpData = text;
   bool ok = SendMessage(hWnd, WM_COPYDATA, (WPARAM)Owner->Handle, (LPARAM)&cdata);
   if (ok)
	  Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"Message copied to QSL...");
   else
   {
	  Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"Message copied to clipboard...");
	  Clipboard()->AsText = msg->Text;
	  MsgBox(Owner,
		 "Message copy to QSL failed. Copied to clipboard!    ",
		  MT_INFO, MB_OK);
   }
   delete [] text;
   return ok;
}
//---------------------------------------------------------------------------

/*bool __fastcall TQSAMainWin::CheckNotifyQSLClipboardOrCopyData(TStrings *msg)
{
   TIniFile *ini = new TIniFile(AppIni());
   bool notify = ini->ReadBool("settings", "notifyqsl", true);
   delete ini;
   if (!notify)
   {
	  ShowStatus("Message NOT copied...");
	  return false;
   }
   HWND hWnd = NULL;
   hWnd = FindWindow("TQS2MainWin", NULL);
   if (!hWnd)
	  hWnd = FindWindow("TQSLMainWin", NULL);
   if (!hWnd)
   {
	  ShowStatus("Message copied to clipboard...");
	  Clipboard()->AsText = msg->Text;
	  return true;
   }

   int len = msg->Text.Length();
   COPYDATASTRUCT cdata;
   char *text = new char[len + 1];
   strncpy(text, msg->Text.c_str(), len);
   cdata.dwData = QSCOPYDATA;
   cdata.cbData = len;
   cdata.lpData = text;
   bool ok = SendMessage(hWnd, WM_COPYDATA, (WPARAM)Handle, (LPARAM)&cdata);
   if (ok)
	  ShowStatus("Message copied to QSL...");
   else
   {
	  ShowStatus("Message copied to clipboard...");
	  Clipboard()->AsText = msg->Text;
	  MsgBox(this,
		 "Message copy to QSL failed. Copied to clipboard!    ",
		  MT_INFO, MB_OK);
   }
   delete [] text;
   return ok;
}*/
//---------------------------------------------------------------------------

/*bool __fastcall CheckNotifyQSLClipboard(TForm *Owner, TStrings *msg)
{
   Clipboard()->AsText = msg->Text;

   bool ok = false;
   //TIniFile *ini = new TIniFile(AppIni());
   //if (ini->ReadBool("settings", "notifyqsl", true))
   {
	  HWND hWnd;
	  if ((hWnd = FindWindow("TQS2MainWin", NULL)) > 0)
	  {
		 Snooze(300);
		 PostMessage(hWnd, WM_QS2MESSAGE, CLIPBOARDREPLY, 0);
		 ok = true;
	  }
	  if (!ok)
	  {
		 if ((hWnd = FindWindow("TQSLMainWin", NULL)) > 0)
		 {
			Snooze(300);
			PostMessage(hWnd, WM_QSLMESSAGE, CLIPBOARDREPLY, 0);
			ok = true;
		 }
	  }
   }
   //delete ini;
   return ok;
}*/
//---------------------------------------------------------------------------

/*bool __fastcall TQSAMainWin::CheckNotifyQSLClipboard(TStrings *msg)
{
   Clipboard()->AsText = msg->Text;

   bool ok = false;
   TIniFile *ini = new TIniFile(AppIni());
   if (ini->ReadBool("settings", "notifyqsl", true))
   {
	  HWND hWnd;
	  if ((hWnd = FindWindow("TQS2MainWin", NULL)) > 0)
	  {
		 Snooze(300);
		 PostMessage(hWnd, WM_QS2MESSAGE, CLIPBOARDREPLY, 0);
		 ok = true;
	  }
	  if (!ok)
	  {
		 if ((hWnd = FindWindow("TQSLMainWin", NULL)) > 0)
		 {
			Snooze(300);
			PostMessage(hWnd, WM_QSLMESSAGE, CLIPBOARDREPLY, 0);
			ok = true;
		 }
	  }
   }
   delete ini;
   return ok;
}*/
//---------------------------------------------------------------------------

AnsiString __fastcall TQSAMainWin::GetNymAndTypeHeader(const AnsiString &MsgKey)
{
   AnsiString keystr, keyid, subno, eaddr, domain, type;
   subno = GetTok(keySubNo, MsgKey, DBFSc);
   assert(subno != "");
   keyid = DB->GetSubjectItem(subno, "keyid", "");
   if (keyid.IsEmpty())
   {
	  MsgBox(this,
		 "Problem: there isn't a PGP key for this nym.   \n"
		 "Possibly you deleted it...",
		  MT_INFO, MB_OK);
	  return "keyerror";
   }
   keystr = GPGGetPublicKeyStr(keyid);
   eaddr = GetEAddress(keystr);
   domain = GetDomain(eaddr);
   type = DB->GetNymserverType(domain);
   if (type.IsEmpty())
   {
	  MsgBox(this,
		 "PROBLEM! I don't have a nymserver type for: " + domain + "   \n"
		 "That won't work. Are you sure it's a nymserver? Maybe you need    \n"
		 "to added it to the list of nymservers.   \n\n"
		 "I can't prepare the message. Copy failed!.",
		  MT_ERROR, MB_OK);
	  type = "typeerror";
   }
   return "Nym: " + GetTok(0, keystr, ':') + "; keyid=" + keyid + "; type=" + type + ";";
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::WMReplyMessage(TMessage &Msg)
{
   TForm *win = (TForm*)Msg.LParam;
   Copymessagetoclipboard1Click(NULL);
   win->Perform(WM_REPLYMESSAGE, 0, 0);
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::Copymessagetoclipboard1Click(TObject *Sender)
{
   CopyMsgBtn->Enabled = false;
   CopyMsgBtn->Caption = "copying";
   AnsiString cap = Caption;
   Caption = " Copying... ";
   Snooze(200);
   assert(Message->Lines->Count);
   AnsiString subno = GetTok(keySubNo, curmsgkey, DBFSc);
   int idx = mail->IndexOf(curmsgkey);
   assert(idx >= 0);
   TStringList *lines = new TStringList();
   lines->AddStrings(ObjMessage(idx));
   if (lines->IndexOf(BeginPGPMessage) >= 0)
   {
	  if (DB->GetSubjectItem(subno, "ekey", "0") == "1")
	  {
		 DSABox(this, "symmetric",
			"I see this message has symmetric encryption. Since I have the "
			"password(s), I'll decrypt the symmetric layer(s) before copying.",
			"Symmetric Encryption", MB_OK);
		 AnsiString symkeys = DB->GetSubjectItem(subno, "kphrase", "");
		 assert(symkeys != "");

		 if (!DecryptSymmetric(this, lines, symkeys))
		 {
			CopyMsgBtn->Caption = CpyBtnCap;
			CopyMsgBtn->Enabled = true;
			Caption = cap;
			delete lines;
			return;
		 }
		 StripSymmetricDecryptionMarkers(lines);
	  }
   }
   ShowStatus("Copying sender's message... ");
   AnsiString nymhdr = GetNymAndTypeHeader(curmsgkey);
   if (nymhdr.Pos("typeerror") || nymhdr.Pos("keyerror"))
   {
	  ShowStatus("");
	  CopyMsgBtn->Caption = CpyBtnCap;
	  CopyMsgBtn->Enabled = true;
	  Caption = cap;
	  return;
   }
   lines->Insert(0, nymhdr);

   HandleMessageTransfer(this, lines, QSCOPYDATA);

   delete lines;
   //Snooze(200);
   ShowStatus("");
   CopyMsgBtn->Caption = CpyBtnCap;
   CopyMsgBtn->Enabled = true;
   Caption = cap;
   ActiveControl = Message;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::WndProc(Messages::TMessage &Msg)
{
   if (Msg.Msg == WM_DRAWITEM)
   {
	  DRAWITEMSTRUCT *dis = (DRAWITEMSTRUCT*)Msg.LParam;
	  if (dis->CtlType == ODT_MENU && Menu && dis->hwndItem == StatusBar->Handle)
			dis->CtlType = ODT_STATIC;
	  //return;
   }
   if (Msg.Msg == WM_SYSCOMMAND)
   {
	  if ((Msg.WParam & 0xFFF0) == SC_MINIMIZE)
	  {
		 if (usetray && !trayed)
		 {
			Application->ShowMainForm = false;
			ShowWindow(Handle, SW_HIDE);
			TrayIcon_Create(blutrayicon);
			traymails = 0;
			trayed = true;
			Snooze(500);
			return;
		 }
	  }
   }
   if (Msg.Msg == WM_TASKBARCREATED && trayed)
   {
	  TrayIcon_Create(trayiconidx);
   }

   TForm::WndProc(Msg);
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::TrayIcon_Create(int iconidx)
{
   if (trayicon)
	  delete trayicon;
   trayicon = new TIcon();
   TrayIcons->GetIcon(iconidx, trayicon);
   trayiconidx = iconidx;
   // add the program's icon to the tray
   AnsiString tipText = QuickSilverAAMVersion();
   memset(&iconinfo,0,sizeof(iconinfo));
   iconinfo.cbSize           = sizeof(iconinfo) ;
   iconinfo.hWnd             = Handle;                            // form handle gets the notification message
   iconinfo.uID              = WM_TRAYICON_ID;                    // id of tray icon setup
   iconinfo.hIcon            = trayicon->Handle;                  // icon to display
   iconinfo.uCallbackMessage = WM_TRAYICONMESSAGE;                // our notification message
   iconinfo.uFlags           = NIF_MESSAGE | NIF_ICON | NIF_TIP;  // indicate modifications (all)
   strncpy (iconinfo.szTip, tipText.c_str(), tipText.Length());            // set tool tip
   iconinfo.szTip[tipText.Length()] = '\0';
   Shell_NotifyIcon(NIM_ADD, &iconinfo) ;                         // add it
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::TrayIcon_Modify(int iconidx, const AnsiString &tipText)
{
   if (trayicon)
	  delete trayicon;
   trayicon = new TIcon();
   TrayIcons->GetIcon(iconidx, trayicon);
   trayiconidx = iconidx;
   iconinfo.cbSize           = sizeof(iconinfo);
   iconinfo.hWnd             = Handle;                            // form handle gets the notification message
   iconinfo.uID              = WM_TRAYICON_ID;                    // id of tray icon setup
   iconinfo.hIcon            = trayicon->Handle;                  // icon to display
   iconinfo.uFlags           = NIF_ICON | NIF_TIP;                // indicate modifications (all)
   strncpy (iconinfo.szTip, tipText.c_str(), tipText.Length());   // set tool tip
   iconinfo.szTip[tipText.Length()] = '\0';
   Shell_NotifyIcon(NIM_MODIFY, &iconinfo) ;                      // modify it
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::TrayIcon_Delete()
{
   // remove the icon from the tray
   memset(&iconinfo, 0, sizeof(iconinfo));
   iconinfo.cbSize   = sizeof(iconinfo);
   iconinfo.hWnd     = Handle;
   iconinfo.uID      = WM_TRAYICON_ID;
   Shell_NotifyIcon(NIM_DELETE, &iconinfo);
   if (trayicon)
   {
	  delete trayicon;
	  trayicon = NULL;
   }
   traymails = 0;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::WMTrayIconMessage(TMessage &Msg)
{
   POINT point;
   switch (Msg.LParam)
   {
	 case WM_LBUTTONDBLCLK:
		TrayMenuRestoreItemClick(NULL);
		break;
	 case WM_RBUTTONDOWN:
	  ::GetCursorPos(&point);
		SetForegroundWindow(Handle);
		TrayMenu->Popup(point.x,point.y);
		break;
  }
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::TrayMenuRestoreItemClick(TObject *Sender)
{
   if (WindowState == wsMinimized)
	  WindowState = wsNormal;
   ShowWindow(Handle, SW_RESTORE);
   SetForegroundWindow(Handle);
   Show();
   TrayIcon_Delete();
   trayed = false;
   traymails = 0;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::TrayMenuExitItemClick(TObject *Sender)
{
   Close();
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::HelpHistoryItemClick(TObject *Sender)
{
   assert(FileExists(AppDir()+"qsa.history.txt"));
   Exec32("notepad.exe "+AppDir()+"qsa.history.txt");
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::HelpMenuClick(TObject *Sender)
{
   HelpHistoryItem->Enabled = FileExists(AppDir()+"qsa.history.txt");
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::Messages1Click(TObject *Sender)
{
   int sels = ListView->SelCount;
   ToolsDecryptmessageItem->Enabled = sels == 1 && WantsDecrypt();
   ToolsDeletemessageItem->Enabled = sels;
   FileOpenItem->Enabled = sels;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::Nntplog1Click(TObject *Sender)
{
   bool firstrun = false;
   if (!HasTab(LogTab))
   {
	  TabControl->Tabs->Add(LogTab);
	  firstrun = true;
   }
   if (!firstrun)
   {
	  Log->Clear();
	  Snooze(200);
   }
   Log->Lines->BeginUpdate();
   DB->GetText("log", Log->Lines);
   Log->SelStart = Log->Text.Length();
   Log->Perform(EM_SCROLLCARET, 0, 0);
   Log->Lines->EndUpdate();
   SetPage(LogTab);
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::Applicationdirectory1Click(TObject *Sender)
{
   ExploreFolder(AppDir());
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::Applicationdatadirectory1Click(TObject *Sender)
{
   ExploreFolder(AppDataDir());
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::Commandprompt1Click(TObject *Sender)
{
   ChDir(AppDir());
   Exec32("cmd.exe", SW_SHOWNORMAL);
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::Open1Click(TObject *Sender)
{
   int sels = ListView->SelCount;
   Selectedmessages1->Enabled = sels > 0;
   Selectedmessages1->Caption = sels == 1 ? "&Selected message":"&Selected messages";
   Applicationdatadirectory1->Visible = AppDataDir() != AppDir();
}
//---------------------------------------------------------------------------

bool __fastcall NeedsDecryptVerify(TStrings *msg)
{
   return msg->IndexOf(BeginPGPMessage) >= 0 ||
	  msg->IndexOf(BeginPGPSignedMessage) >= 0;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::TabControlChange(TObject *Sender)
{
   int sels = ListView->SelCount;
   AnsiString TabText = TabControl->Tabs->Strings[TabControl->TabIndex];
   if (TabText == StatusTab)
   {
	   Status->BringToFront();
	   CopyBtn->Enabled = Status->SelLength > 0;
	   DecryptBtn->Enabled = false;
	   CopyMsgBtn->Enabled = false;
	   DeleteBtn->Enabled = false;
	   FewerHeaders->Enabled = false;
	   return;
	}
	if (TabText == MessageTab)
	{
	   Message->BringToFront();
	   CopyBtn->Enabled = Message->SelLength > 0;
	   DecryptBtn->Enabled = sels == 1 && WantsDecrypt();
	   CopyMsgBtn->Enabled = !DecryptBtn->Enabled;
	   if (CopyMsgBtn->Enabled)
		  CopyMsgBtn->Enabled == Message->Text.Length();
	   DeleteBtn->Enabled = sels > 0;
	   FewerHeaders->Enabled = true;
	   return;
	}
	if (TabText == LogTab)
	{
	   Log->BringToFront();
	   CopyBtn->Enabled = Log->SelLength > 0;
	   DecryptBtn->Enabled = false;
	   CopyMsgBtn->Enabled = false;
	   DeleteBtn->Enabled = false;
	   FewerHeaders->Enabled = false;
	   return;
   }
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::SetPage(const AnsiString &TabText)
{
   if (TabControl->Tabs->IndexOf(TabText) < 0)
	  TabControl->Tabs->Add(TabText);
   TabControl->TabIndex = TabControl->Tabs->IndexOf(TabText);
   TabControlChange(TabControl);

   if (TabText == StatusTab)
	  Status->BringToFront();
   else if (TabText == MessageTab)
	  Message->BringToFront();
   else if (TabText == LogTab)
	  Log->BringToFront();
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::Copy1Click(TObject *Sender)
{
   AnsiString TabText = TabControl->Tabs->Strings[TabControl->TabIndex];
   if (TabText == StatusTab)
	  Status->CopyToClipboard();
   else
	  Log->CopyToClipboard();
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::Selectall1Click(TObject *Sender)
{
   AnsiString TabText = TabControl->Tabs->Strings[TabControl->TabIndex];
   if (TabText == StatusTab)
	  Status->SelectAll();
   else
	  Log->SelectAll();
}
//---------------------------------------------------------------------------

bool __fastcall TQSAMainWin::HasTab(const AnsiString &TabText)
{
   return TabControl->Tabs->IndexOf(TabText) >= 0;
}
//---------------------------------------------------------------------------

bool __fastcall TQSAMainWin::IsActiveTab(const AnsiString &TabText)
{
   return TabControl->Tabs->Strings[TabControl->TabIndex] == TabText;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::StartTimerTimer(TObject *Sender)
{
   StartTimer->Enabled = false;
   CheckMailTimer->Enabled = true;
   checkminselapsed = 0;
   GetBtn->Click();
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::CheckMailTimerTimer(TObject *Sender)
{
   if (!checkmail)
   {
	  CheckMailTimer->Enabled = false;
	  return;
   }

   checkminselapsed++;
   //assert(checkminselapsed <= checkmailmins);
   if (!goterror && !NNTPModHandle && checkminselapsed >= checkmailmins)
   {
	  checkminselapsed = 0;
	  if (goterror)
		 return;
	  GetBtn->Click();
   }
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::HelpTopicsItemClick(TObject *Sender)
{
   DisplayHtmlHelpTopic(this, ">qsahelp");
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::FileSecuremodeClick(TObject *Sender)
{
   DB->ManageSecureMode(this);
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::PGPkeyrings1Click(TObject *Sender)
{
   ManagePGPKeyrings(this);
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::Timer2Timer(TObject *Sender)
{
   Timer2->Enabled = false;
   DisplayHtmlHelpTopic(this, "::/html/start.htm>qsahelp");
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::FileSavedatabaseClick(TObject *Sender)
{
   if (!DB->modified)
	  return;

   ShowStatus("");
   Snooze(200);
   ShowStatus(" Saving database... ");
   if (needspack)
   {
	  //ShowStatus(" saving database...");
	  Snooze(200);
	  PackDatabase(this); // saves the file after packing
	  needspack = false;
   }
   else
   {
	  //ShowStatus(" saving database...");
	  DB->Save();
   }
   Snooze(200);
   ShowStatus("");
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::Nymservers1Click(TObject *Sender)
{
   ManageNymservers(this);
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::ListViewSelectItem(TObject *Sender,
	  TListItem *Item, bool Selected)
{
   if (!Item->Selected || !Selected)
	  return;

   int sels = ListView->SelCount;
 //CopyMsgBtn->Enabled = sels == 1;
   DeleteBtn->Enabled = sels > 0;
   if (sels == 0 || sels > 1)
   {
	  if (HasTab(MessageTab))
		 Message->Clear();

	  DecryptBtn->Enabled = false;
	  CopyMsgBtn->Enabled = false;
	  prvmsgkey = curmsgkey;
	  curmsgkey = "";
	  return;
   }

   assert(ListView->Selected);
   TListItem *item = ListView->Selected;
   AnsiString msgkey = mail->Strings[item->Index];

   if (HasTab(MessageTab) && msgkey == curmsgkey)
   {
	  SetPage(MessageTab);
	  return;
   }

   if (!HasTab(MessageTab))
	  TabControl->Tabs->Add(MessageTab);

   curmsgkey = msgkey;
   TStrings *objmsg = dynamic_cast<TStrings*>(mail->Objects[item->Index]);
   if (objmsg == NULL) // ^ ObjMessage can't handle NULL
   {
	  objmsg = new TStringList();
	  DB->LoadMessage(msgkey, objmsg);
	  mail->Objects[item->Index] = objmsg;
	  TMsgManager *msgman = new TMsgManager(objmsg);
	  msgman->DelHeader("Subject-No:");
	  delete msgman;
   }

   TStringList *msg = new TStringList();
   msg->AddStrings(objmsg);
   if (FewerHeaders->Checked)
	  CompactMessageHeaders(msg);

   Message->Lines->BeginUpdate();
   Message->Lines->Clear();
   Message->Lines->AddStrings(msg);
   Message->SelStart = 0;
   Message->Perform(EM_SCROLLCARET, 0, 0);
   Message->Lines->EndUpdate();
   Snooze(50);
   bool needs = NeedsDecryptVerify(objmsg);
   DecryptBtn->Enabled = needs;
   CopyMsgBtn->Enabled = !DecryptBtn->Enabled;
   if (!needs && ((curmsgkey[bflags1] & mflRead) != mflRead))
	  MarkMessageRead(item->Index);
   SetPage(MessageTab);
   delete msg;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::ListViewDblClick(TObject *Sender)
{
   assert(ListView->Selected);
   TListItem *item = ListView->Selected;
   curmsgkey = mail->Strings[item->Index];
   TStrings *objmsg = ObjMessage(item->Index);
   if (!NeedsDecryptVerify(objmsg) && !IsMessageRead(curmsgkey))
	  MarkMessageRead(item->Index);
   ViewMessage(this, curmsgkey, objmsg);
}
//---------------------------------------------------------------------------

/*void __fastcall TQSAMainWin::ListViewDblClick(TObject *Sender)
{
   assert(ListView->Selected);
   TListItem *item = ListView->Selected;
   curmsgkey = mail->Strings[item->Index];
   TStringList *objmsg = new TStringList();
   DB->LoadMessage(curmsgkey, objmsg);
   assert(objmsg->Count);
   mail->Objects[item->Index] = objmsg;
   if (!NeedsDecryptVerify(objmsg) && !IsMessageRead(curmsgkey))
	  MarkMessageRead(item->Index);
   ViewMessage(this, curmsgkey, objmsg);
   delete objmsg;
}*/
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::ListViewKeyDown(TObject *Sender, WORD &Key,
	  TShiftState Shift)
{
   if (Key == VK_RETURN)
   {
	  ListViewDblClick(NULL);
   }
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::Save1Click(TObject *Sender)
{
   assert(curmsgkey != "");
   int index = mail->IndexOf(curmsgkey);
   assert(index >= 0);
   DB->SaveMessage(curmsgkey, ObjMessage(index));
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::ToolsConfigmessagesClick(TObject *Sender)
{
   ManageNymConfigMessages(this);
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::qsaserverstxt1Click(TObject *Sender)
{
   Exec32("notepad "+AppDir()+"qsa.servers.txt");
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::Composenewmessage1Click(TObject *Sender)
{
   if (FileExists(AppDir()+"qsl.exe"))
   {
	  TIniFile *ini = new TIniFile(AppDir()+"qsl.ini");
	  AnsiString qslver = ini->ReadString("settings", "version", "");
	  delete ini;
	  if (qslver < "1.0.1")
	  {
		 MsgBox(this, "Sorry, I require QSL version 1.0.1 or greater   \n"
		 "to handle this feature!   ",
		 MT_INFO, MB_OK);
		 return;
	  }
   }
   HWND hWnd = FindWindow("TQSLMainWin", NULL);
   if (!hWnd)
   {
	  if (FileExists(AppDir()+"qsl.exe"))
	  {
		 Exec32(AppDir()+"qsl.exe -nowin");
		 for (int i = 0; i < 10; i++)
		 {
			Snooze(1000);
			hWnd = FindWindow("TQSLMainWin", NULL);
			if (hWnd)
			   break;
		 }
		 if (!hWnd)
		 {
			MsgBox(this, "Sorry, I failed to start QSL.   ", MT_ERROR, MB_OK);
			return;
		 }
	  }
	  if (!hWnd)
	  {
		 MsgBox(this, "Sorry, QSL is not installed.   ", MT_ERROR, MB_OK);
		 return;
	  }
   }
   PostMessage(hWnd, WM_QSLMESSAGE, COMPOSENEWMESSAGE, 0);
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::View1Click(TObject *Sender)
{
   Fewerheaders1->Checked = FewerHeaders->Checked;
   Fewerheaders1->Enabled = FewerHeaders->Enabled;
}
//---------------------------------------------------------------------------

void __fastcall AddCompactSig(int &i, TStrings *msg, TStrings *sig)
{
   for (int x = 0; x < sig->Count-1; x++)
   {
	  AnsiString line = sig->Strings[x].LowerCase();;
	  if (line.Pos("signature") && !line.Pos("binary"))
	  {
		 msg->Insert(i++, sig->Strings[x]);
		 continue;
	  }
	  if (line.Pos("verified"))
	  {
		 msg->Insert(i++, sig->Strings[x]);
		 continue;
	  }
   }
}
//---------------------------------------------------------------------------

void __fastcall CompactMessageHeaders(TStrings *msg)
{
   // compact nymserver headers
   int i = 0; while (i < msg->Count)
   {
	  AnsiString hdr, line = msg->Strings[i];
	  if (line.Trim().IsEmpty())
		 break;
	  if (!line.Pos(":"))
		 break;

	  //assert(line.Pos(":"));
	  hdr = ":"+line.SubString(1, line.Pos(":")).LowerCase();
	  if (!nymserverhdrs.Pos(hdr))
	  {
		 msg->Delete(i);
		 if (i < msg->Count && msg->Strings[i].Trim().IsEmpty())
			break;

		 while (i < msg->Count && indentchars.Pos(msg->Strings[i].SubString(1,1)) > 0)
			msg->Delete(i);  // get folded portion
		 continue;
	  }
	  else
	  {
		 while (i+1 < msg->Count && indentchars.Pos(msg->Strings[i+1].SubString(1,1)) > 0)
			i++; // advanced past folded portion
	  }
	  i++;
   }
   if (NeedsDecryptVerify(msg))
	  return;

   i++;
   TStringList *sig = new TStringList();
   // compact nymserver sig verification
   while (i < msg->Count && msg->Strings[i].Pos("***") == 1)
   {
	  sig->Add(msg->Strings[i]);
	  msg->Delete(i);
   }
   AddCompactSig(i, msg, sig);
   i++;
   // compact message headers
   while (i < msg->Count)
   {
	  AnsiString hdr, line = msg->Strings[i];
	  if (line.Trim().IsEmpty())
		 break;
	  if (!line.Pos(":"))
		 break;

	  //assert(line.Pos(":"));
	  hdr = ":"+line.SubString(1, line.Pos(":")).LowerCase();
	  if (!messagehdrs.Pos(hdr))
	  {
		 msg->Delete(i);
		 if (i < msg->Count && msg->Strings[i].Trim().IsEmpty())
			break;

		 while (i < msg->Count && indentchars.Pos(msg->Strings[i].SubString(1,1)) > 0)
			msg->Delete(i);  // get folded portion
		 continue;
	  }
	  else
	  {
		 while (i+1 < msg->Count && indentchars.Pos(msg->Strings[i+1].SubString(1,1)) > 0)
			i++;  // advance past folded portion
	  }
	  i++;
   }
   i++;
   sig->Clear();
   // compact message sig verification--if present
   while (i < msg->Count && msg->Strings[i].Pos("***") == 1)
   {
	  sig->Add(msg->Strings[i]);
	  msg->Delete(i);
   }
   if (sig->Count)
   {
	  AddCompactSig(i, msg, sig);
   }
   delete sig;
   // delete trailing pgp lines
   while ((i = msg->IndexOf(EndPGPDecryptedVerifiedMessage)) >= 0)
	  msg->Delete(i);
   while ((i = msg->IndexOf(EndPGPDecryptedMessage)) >= 0)
	  msg->Delete(i);
   while ((i = msg->IndexOf(EndPGPVerifiedMessage)) >= 0)
	  msg->Delete(i);
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::FewerHeadersClick(TObject *Sender)
{
   if (HasTab(MessageTab))
   {
	  assert(curmsgkey != "");
	  int index = mail->IndexOf(curmsgkey);
	  assert(index >= 0);
	  TStringList *msg = new TStringList();
	  msg->AddStrings(ObjMessage(index));
	  Message->Lines->BeginUpdate();
	  Message->Clear();
	  if (FewerHeaders->Checked)
		 CompactMessageHeaders(msg);
	  Message->Lines->AddStrings(msg);
	  Message->SelStart = 0;
	  Message->Perform(EM_SCROLLCARET, 0, 0);
	  Snooze(10);
	  Message->Lines->EndUpdate();
	  delete msg;
   }
   TIniFile *ini = new TIniFile(AppIni());
   ini->WriteBool("settings", "fewerheaders", FewerHeaders->Checked);
   delete ini;
}
//---------------------------------------------------------------------------

void __fastcall TQSAMainWin::Fewerheaders1Click(TObject *Sender)
{
   FewerHeaders->Checked = Fewerheaders1->Checked;
}
//---------------------------------------------------------------------------

