/*
*  msgmans.cpp 
*
*  Copyright (C) 2012 Richard Christman 
*
*  This file is part of QuickSilver AAM.
*
*  QuickSilver AAM is free software; you can redistribute it and/or modify 
*  it under the terms of the GNU General Public License as published by the 
*  Free Software Foundation; either version 3 of the License, or (at your 
*  option) any later version.
* 
*  This program is distributed in the hope that it will be useful, but 
*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
*  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
*  for more details.
* 
*  You should have received a copy of the GNU General Public License along 
*  with this program; if not, see <http://www.gnu.org/licenses>.
* 
*  Additional permission under GNU GPL version 3 section 7
* 
*  If you modify QuickSilver AAM, or any covered work, by linking or 
*  combining it with CodeGear VCL (or a modified version of that library), 
*  containing parts covered by the terms of CodeGear's license, the licensors 
*  of QuickSilver AAM grant you additional permission to convey the 
*  resulting work.
*
*/

//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "msgmans.h"
#include "utility.h"
#include "dbs.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)

//---------------------------------------------------------------------------
// THeaderManager
//---------------------------------------------------------------------------

__fastcall TMsgHeader::TMsgHeader(TStrings *Message)
{
   Lines = Message;
}
//---------------------------------------------------------------------------

void __fastcall TMsgHeader::Insert(int Index, const AnsiString &Text)
{
   Lines->Insert(Index, Text);
}
//---------------------------------------------------------------------------

void __fastcall TMsgHeader::Delete(int Index)
{
   Lines->Delete(Index);
}
//---------------------------------------------------------------------------

int __fastcall TMsgHeader::Length()
{
   int l = 0, c = GetCount();
   for (int i = 0; i < c; i++)
      l += Strings[i].Length() + 2;
   return l;
}
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgHeader::GetLine(int Index)
{
   return Lines->Strings[Index];
}
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgHeader::GetText()
{
   AnsiString text;
   for (int i = 0; i < Count; i++)
   {
      if (i > 0)
		 text += "\n\r";
      text += Lines->Strings[i];
   }
   return text;
}
//---------------------------------------------------------------------------

int __fastcall TMsgHeader::GetCount()
{
   int c = -1; while (++c < Lines->Count)
   {
      if (Lines->Strings[c].Trim().IsEmpty())
      {
		 if (c < Lines->Count - 1 && Lines->Strings[c + 1] == "##")
			continue;
		 break;
	  }
   }
   return c;
}

//---------------------------------------------------------------------------
// TBodyManager
//---------------------------------------------------------------------------

__fastcall TMsgBody::TMsgBody(TStrings *Message)
{
   Lines = Message;
}
//---------------------------------------------------------------------------

void __fastcall TMsgBody::Insert(int Index, const AnsiString &Text)
{
   Lines->Insert(GetFirstBody() + Index, Text);
}
//---------------------------------------------------------------------------

void __fastcall TMsgBody::Delete(int Index)
{
   Lines->Delete(GetFirstBody() + Index);
}
//---------------------------------------------------------------------------

int __fastcall TMsgBody::Length()
{
   int l = 0, c = GetCount();
   for (int i = 0; i < c; i++)
	  l += Strings[i].Length() + 2;
   return l;
}
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgBody::GetLine(int Index)
{
   return Lines->Strings[GetFirstBody() + Index];
}
//---------------------------------------------------------------------------

int __fastcall TMsgBody::GetCount()
{
   return Lines->Count - GetFirstBody();
}
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgBody::GetText()
{
   AnsiString text;
   for (int i = GetFirstBody(); i < Lines->Count; i++)
   {
	  if (i > GetFirstBody())
		 text += "\n\r";
	  text += Lines->Strings[i];
   }
   return text;
}
//---------------------------------------------------------------------------

int __fastcall TMsgBody::GetFirstBody()
{
   int f = -1; while (++f < Lines->Count)
   {
	  if (Lines->Strings[f].Trim().IsEmpty())
	  {
		 if (f < Lines->Count - 1 && Lines->Strings[f + 1] == "##")
			continue;
		 break;
	  }
   }
   return f + 1;
}

//---------------------------------------------------------------------------
// TMsgManager
//---------------------------------------------------------------------------

__fastcall TMsgManager::TMsgManager(TStrings *Msg)
   : TObject()
{
   headers = new TMsgHeader(Msg);
   body = new TMsgBody(Msg);
//sm(body->Count);
   Lines = Msg;
}
//---------------------------------------------------------------------------

__fastcall TMsgManager::TMsgManager(TForm *Owner, TStrings *Msg)
   : TObject()
{
   owner = Owner;
   headers = new TMsgHeader(Msg);
   body = new TMsgBody(Msg);
   Lines = Msg;
}
//---------------------------------------------------------------------------

__fastcall TMsgManager::~TMsgManager()
{
   delete headers;
   delete body;
}
//---------------------------------------------------------------------------

bool __fastcall TMsgManager::HasPaste()
{
   return IndexOfHeader("##") >= 0;
}
//---------------------------------------------------------------------------

bool __fastcall TMsgManager::HasHeader(const AnsiString &Header)
{
   bool ok = false;
   AnsiString line;

   AnsiString header = Header.LowerCase();
   int i = -1; while (++i < Lines->Count)
   {
	  line = Lines->Strings[i].LowerCase().Trim();
	  if (line.IsEmpty())
	  {
		 if (i < Lines->Count - 1 && Lines->Strings[i + 1] == "##")
			continue;
		 break;
	  }

	  if (line.Pos(header) == 1)
	  {
		 ok = true;
		 break;
	  }
   }
   return ok;
}
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgManager::GetHeader(const AnsiString &Header)
{
   AnsiString hdr, line;

   AnsiString header = Header.LowerCase();
   int i = -1; while (++i < Lines->Count)
   {
	  line = Lines->Strings[i];
	  if (line.Trim().IsEmpty())
	  {
		 if (i < Lines->Count - 1 && Lines->Strings[i + 1] == "##")
			continue;
		 break;
	  }

	  if (line.LowerCase().Pos(header) == 1)
	  {
		 hdr = line;
		 while (++i < Lines->Count)
		 {
			line = Lines->Strings[i];
			if (line.IsEmpty() || line[1] != ' ' && line[1] != '\t')
			   break;
			hdr += "\r\n"+line;
		 }
		 break;
	  }
   }

   if (!hdr.IsEmpty())
	  hdr = hdr.SubString(hdr.Pos(":") + 1, hdr.Length()).Trim();
   return hdr;
}
//---------------------------------------------------------------------------

int __fastcall TMsgManager::DelHeader(const AnsiString &Header)
{
   int result = -1;
   AnsiString line, header = Header.LowerCase();
   int x = 0; while (x < Lines->Count)
   {
	  line = Lines->Strings[x].LowerCase();
	  if (line.Trim().IsEmpty())
	  {
		 if (x < Lines->Count - 1 && Lines->Strings[x + 1] == "##")
			continue;
		 break;
	  }

	  if (line.Pos(header) == 1)
	  {
		 Lines->Delete(x);
		 while (x < Lines->Count)
		 {
			line = Lines->Strings[x];
			if (line.IsEmpty())
			   break;

			if (line[1]==' ' || line[1]=='\t')
			{
			   Lines->Delete(x);
			   continue;
			}

			break;
		 }
		 result = x;
		 break;
	  }
	  x++;
   }

   return result;
}
//---------------------------------------------------------------------------

void __fastcall TMsgManager::AppendHeader(const AnsiString &HeaderAndValue)
{
   AppHeader(HeaderAndValue);
}

void __fastcall TMsgManager::AppHeader(const AnsiString &HeaderAndValue)
{
   for (int i = 0; i < Lines->Count; i++)
   {
	  if (Lines->Strings[i].Trim().IsEmpty())
	  {
		 Lines->Insert(i, HeaderAndValue);
		 break;
	  }
   }
}
//---------------------------------------------------------------------------

int __fastcall TMsgManager::IndexOfHeader(const AnsiString &Header)
{
   int eoh = EOH(), x = -1;
   for (int i = 0; i < eoh; i++)
   {
	  if (Lines->Strings[i].LowerCase().Pos(Header.LowerCase()) == 1)
	  {
		 x = i;
		 break;
	  }
   }
   return x;
}
//---------------------------------------------------------------------------

int __fastcall TMsgManager::IndexOfHeaderAfter(const AnsiString &Header)
{
   int x = IndexOfHeader(Header);
   if (x >= 0)
   {
	  while (x++ < Lines->Count)
	  {
		 AnsiString line = Lines->Strings[x];
		 if (line.SubString(1,1) == " " || line.SubString(1,1) == "\t")
			continue;
		 break;
	  }
   }
   return x;
}
//---------------------------------------------------------------------------

/*AnsiString __fastcall TMsgManager::GetAddress(const AnsiString &fromline)
{
   int f, l, at;
   AnsiString addr;
   char Terms[7] = { ' ', ',', '<', '>', '=', 10, 13 };

   at = fromline.Pos("@");
   if (at)
   {
	  sm("'"+fromline+"'");
	  f = at-1; l = at+1;
	  while (f >= 1 && !strchr(Terms, fromline[f]))
		 f--;
	  while (l <= fromline.Length() && !strchr(Terms, fromline[l]))
		 l++;
	  addr = fromline.SubString(f, l - (f-1));
	  sm("'"+addr+"' \n"+IntToStr(addr.Length()));
   }
   return addr.Trim();
}*/
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgManager::GetAddress(const AnsiString &fromline)
{
   int f, l, at;
   AnsiString addr;
   //char Terms[7] = { ' ', ',', '<', '>', '=', 10, 13 };
   AnsiString terms = " ;<,>=\n\r";

   at = fromline.Pos("@");
   if (at)
   {
	  f = at; l = at;
	  while (f > 1 && !terms.Pos(fromline[f - 1]))
		 f--;
	  while (l <= fromline.Length() && !terms.Pos(fromline[l]))
		 l++;
	  addr = fromline.SubString(f, l - f);
   }

   return addr.Trim();
}
//---------------------------------------------------------------------------

/*AnsiString __fastcall TMsgManager::GetAddress(const AnsiString &fromline)
{
   char *at;
   UINT f,l;
   AnsiString addr;
   char Terms[7] = { ' ', ',', '<', '>', '=', 10, 13 };

   char *tmp = new char[fromline.Length() + 1];
   strcpy(tmp, fromline.c_str());
   at = strchr(tmp, '@');
   if (at)
   {
	  l = f = at - tmp;
	  while (f > 0 && !strchr(Terms, tmp[f - 1]))
		 f--;
	  while (l < strlen(tmp) && !strchr(Terms, tmp[l]))
		 l++;
	  tmp[l] = '\0';
	  addr = AnsiString(&tmp[f]);
   }
   delete tmp;
   return addr.Trim();
}*/
//---------------------------------------------------------------------------

int __fastcall TMsgManager::GetAddresses(const AnsiString &Header, TStrings *Addresses)
{
   int p = 1;
   AnsiString addr, buf = GetHeader(Header);

   while (p > 0)
   {
	  addr = GetAddress(buf).Trim();
	  if (!addr.IsEmpty() && Addresses->IndexOf(addr) == -1)
		 Addresses->Add(addr);

	  p = buf.Pos(",");
	  buf = buf.SubString(p + 1,1024);
   };

   return Addresses->Count;
}
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgManager::GetRecipient()
{
   //if (HasHeader("Post:"))
   //   return GetHeader("Post:");
   TStringList *addresses = new TStringList();
   GetRecipients(addresses);
   AnsiString recip = addresses->Count ? addresses->Strings[0]:AnsiString("");
   delete addresses;
   return recip;
}
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgManager::GetRecipients()
{
   TStringList *addresses = new TStringList();
   GetRecipients(addresses);
   AnsiString recips = addresses->Text;
   delete addresses;
   return recips;
}
//---------------------------------------------------------------------------

int __fastcall TMsgManager::GetRecipients(TStrings *Addresses)
{
   Addresses->Clear();
   GetAddresses("To:", Addresses);
   GetAddresses("Cc:", Addresses);
   GetAddresses("Bcc:", Addresses);
   //assert(Addresses->Count);
   return Addresses->Count;
}
//---------------------------------------------------------------------------
// get complete email addresses
int __fastcall TMsgManager::GetFullAddresses(TStrings *Recipients)
{
   Recipients->Clear();
   GetFullAddresses("To:", Recipients);
   GetFullAddresses("Cc:", Recipients);
   GetFullAddresses("Bcc:", Recipients);
   return Recipients->Count;
}
//---------------------------------------------------------------------------

int __fastcall TMsgManager::GetFullAddresses(const AnsiString &Header,
   TStrings *Recipients)
{
   AnsiString recip, recips = GetHeader(Header);
   if (!recips.IsEmpty())
   {
      TStringList *list = new TStringList();
	  int x;
	  while ((x = recips.Pos("\n")) > 0)
         recips.Delete(x, 1);
      while ((x = recips.Pos("\r")) > 0)
         recips.Delete(x, 1);
      while ((x = recips.Pos("\t")) > 0)
         recips.Delete(x, 1);

	  char div = ';';
      x = recips.Pos(div);
      if (x == 0)
      {
		 div = ',';
         x = recips.Pos(div);
      }
      if (x > 0)
      {
         x = 0; while ((recip = GetTok(x, recips, div)) != "")
         {
            Recipients->Add(recip.Trim());
            x++;
         }
	  }
      else
      {
         Recipients->Add(recips);
	  }
      delete list;
   }
   return Recipients->Count;
}
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgManager::GetNewsgroups()
{
   int pos;
   AnsiString grps = GetHeader("Newsgroups:").Trim();
   if (grps.IsEmpty())
	  grps = GetHeader("Post:");

   while ((pos = grps.Pos("\t")) > 0)
	  grps.Delete(pos,1);
   while ((pos = grps.Pos("\r")) > 0)
	  grps.Delete(pos,1);
   while ((pos = grps.Pos("\n")) > 0)
	  grps.Delete(pos,1);
   while ((pos = grps.Pos(" ")) > 0)
	  grps.Delete(pos,1);

   return grps;
}
//---------------------------------------------------------------------------

bool __fastcall TMsgManager::IsPost()
{
	return HasHeader("Post:") || HasHeader("Newsgroups:");
}
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgManager::GetSubject()
{
   int p; AnsiString sub = GetHeader("Subject:");
   while ((p = sub.Pos('\r')) > 0)
	  sub.Delete(p, 1);
   while ((p = sub.Pos('\n')) > 0)
	  sub.Delete(p, 1);
   while ((p = sub.Pos('\t')) > 0)
	  sub[p] = ' ';
   return sub;
}
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgManager::GetFrom()
{
   int p; AnsiString frm = GetHeader("From:");
   while ((p = frm.Pos('\r')) > 0)
	  frm.Delete(p, 1);
   while ((p = frm.Pos('\n')) > 0)
	  frm.Delete(p, 1);
   while ((p = frm.Pos('\t')) > 0)
	  frm[p] = ' ';
   return frm;
}
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgManager::GetPost()
{
   int p; AnsiString pst = GetHeader("Post:");
   while ((p = pst.Pos('\r')) > 0)
	  pst.Delete(p, 1);
   while ((p = pst.Pos('\n')) > 0)
	  pst.Delete(p, 1);
   while ((p = pst.Pos('\t')) > 0)
	  pst.Delete(p, 1);
   while ((p = pst.Pos(" ")) > 0)
	  pst.Delete(p, 1);
   return pst;
}
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgManager::GetChain()
{
   return GetHeader("Chain:");
}
//---------------------------------------------------------------------------

bool __fastcall TMsgManager::ValidAddress(const AnsiString &Address)
{
   int at, lf, cr;
   AnsiString addr = Address;

   at = addr.Pos("@");
   if (at < 2)
	  return false;

   lf = addr.Pos("\n");
   if (lf)
	  addr = addr.SubString(1, lf - 1);

   cr = addr.Pos("\r");
   if (cr)
	  addr = addr.SubString(1, cr - 1);

   if (addr.SubString(at + 1, addr.Length()).Pos(".") ==  0)
	  return false;

   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TMsgManager::ValidHost(const AnsiString &Host)
{
   return (Host == "localhost" || (Host.Pos("@") == 0 && Host.Pos(".") > 0));
}
//---------------------------------------------------------------------------

/*bool __fastcall TMsgManager::ValidAddress(const AnsiString &Address)
{
   int at = Address.Pos("@"), dot = Address.Pos("."), cr = Address.Pos("\r"),
	  lf = Address.Pos("\n"), len = Address.Length();
   if (lf && !cr)
	  cr = lf;
   if (cr && len >= cr)
	  len = cr - 1;
   return at > 1 && dot > at + 1 && dot < len;
}*/
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgManager::GetKey()
{
   assert(Lines->Count > 0);
   /*AnsiString recip = recipient;
   if (recip.IsEmpty())
	  recip = newsgroups;
   if (recip.IsEmpty())
	  recip = "unknown";*/

   //char flags1 = '', flags2 = '';
   //flags1 |= mflOutgoing;
   //if (HasHeader("Attachment:"))
   //	  flags1 |= mflAttachment;

   AnsiString date = GetHeader("Date:"), msgno = GetHeader("Message-No:");
   AnsiString key = FormatDateTime("yymmddhhnnsszzz",Now())+""+msgno+DBFS+
	  subject+DBFS+IntToStr(GetLength())+DBFS+GetHeader("From:");

   return key;
}
//---------------------------------------------------------------------------

/*AnsiString __fastcall TMsgManager::GetReceiveKey()
{
   assert(Lines->Count > 0);

   AnsiString Decoded, Sender, Subject;
   Sender = from;
   if (Sender.IsEmpty())
	  Sender = replyto;
   if (Sender.IsEmpty())
	  Sender = "unknown";

   //byte flags1 = 128, flags2 = 128;
   //return FormatDateTime("yymmddhhnnsszzz", Now()) + char(flags1) + char(flags2) + DBFS +
   //	  Sender + DBFS + subject + DBFS + IntToStr(GetLength());

   return FormatDateTime("yymmddhhnnsszzz", Now())+""+DBFS+Sender+DBFS+
	  subject+DBFS+IntToStr(GetLength());
}*/
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgManager::GetReplyTo()
{
   return GetHeader("Reply-To:");
}
//---------------------------------------------------------------------------

int __fastcall TMsgManager::EOH()
{
   int eoh = -1; while (++eoh < Lines->Count)
   {
	  if (Lines->Strings[eoh].Trim().IsEmpty())
      {
		 if (eoh < Lines->Count - 1 && Lines->Strings[eoh + 1] == "##")
            continue;
         break;
	  }
   }
   return eoh;
}
//---------------------------------------------------------------------------

int __fastcall TMsgManager::GetLength()
{
   return Lines->Text.Length();
}
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgManager::GetHost()
{
   return GetHeader("Host:");
}
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgManager::GetSignOn()
{
   return GetHeader("Sign-On:");
}
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgManager::GetReferences(bool AddDummy)
{
   AnsiString refs = GetHeader("References:").Trim();
   if (!refs.IsEmpty())
   {
	  int pos;
	  while ((pos = refs.Pos("\t")) > 0)
		 refs.Delete(pos, 1);
	  while ((pos = refs.Pos("\r")) > 0)
		 refs.Delete(pos, 1);
	  while ((pos = refs.Pos("\n")) > 0)
		 refs.Delete(pos, 1);
	  while ((pos = refs.Pos(" ")) > 0)
		 refs.Delete(pos, 1);
	  while ((pos = refs.Pos("><")) > 0)
		 refs.Insert("\r\n ", pos + 1);

	  TStringList *lines = new TStringList();
	  lines->Text = refs;

	  if (AddDummy && HasHeader("Chain:") && lines->Count > 0)
		 refs += "\r\n <>";
	  delete lines;
   }

   return refs;
}
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgManager::GetLastReference()
{
   TStringList *refs = new TStringList();
   refs->Text = GetReferences(false);
   if (!refs->Count)
	  return "";

   AnsiString ref = refs->Strings[refs->Count - 1];
   return GetTok(0, GetTok(1, ref, '<'), '>');
}
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgManager::GetReferences4Reply()
{
   AnsiString mid = GetHeader("Message-ID:"),
	  refs = GetHeader("References:").Trim();
   if (!mid.IsEmpty())
	  refs += "\r\n " + mid;

   if (!refs.IsEmpty())
   {
	  int pos;
	  while ((pos = refs.Pos("\t")) > 0)
		 refs.Delete(pos, 1);
	  while ((pos = refs.Pos("\r")) > 0)
		 refs.Delete(pos, 1);
	  while ((pos = refs.Pos("\n")) > 0)
		 refs.Delete(pos, 1);
	  while ((pos = refs.Pos(" ")) > 0)
		 refs.Delete(pos, 1);
	  while ((pos = refs.Pos("><")) > 0)
		 refs.Insert("\r\n ", pos + 1);

	  TStringList *lines = new TStringList();
	  lines->Text = refs;
	  delete lines;
   }

   return refs;
}
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgManager::Get4Reply(const AnsiString &header)
{
   int p, i;
   AnsiString tmp;
   TStringList *hdr = new TStringList();

   if (header.Trim() == "References:")
   {
	  hdr->Text = GetReferences4Reply();
	  if (!hdr->Text.IsEmpty())
		 hdr->Text = " " + hdr->Text;
	  hdr->Text = "References:" + hdr->Text;
   }
   else if (header.Trim() == "Newsgroups:")
   {
	  tmp = "Newsgroups: " + GetNewsgroups();
	  i = 1; while (i <= tmp.Length())
	  {
		 if (tmp[i] == ',')
			tmp.Insert("\r\n ", i + 1);
		 i++;
	  }
	  hdr->Text = tmp;
   }
   else if (header.Trim() == "Post:")
   {
	  tmp = "Post: " + GetNewsgroups();
	  i = 1; while (i <= tmp.Length())
	  {
		 if (tmp[i] == ',')
			tmp.Insert("\r\n ", i + 1);
		 i++;
	  }
	  hdr->Text = tmp;
   }
   else if (header.Trim() == "To:")
   {
	  tmp = GetHeader("Reply-To:");
	  if (tmp.IsEmpty())
		 tmp = GetHeader("From:");
	  if (GetAddress(tmp) == "Use-Author-Address-Header@[127.1]")
	  {
		 tmp = GetHeader("Author-Address:")+" ";
		 if ((p = tmp.Pos(" <AT> ")) > 0)
		 {
			tmp.Delete(p, 6);
			tmp.Insert("@", p);
		 }
		 while ((p = tmp.Pos(" <DOT> ")) > 0)
		 {
			tmp.Delete(p, 7);
			tmp.Insert(".", p);
		 }
	  }
	  if (!tmp.IsEmpty())
		 tmp = "To: "+tmp;
	  hdr->Text = tmp;
   }
   else if (header.Trim() == "Subject:")
   {
	  tmp = GetHeader("Subject:");
	  hdr->Text =  "Subject: " + (tmp.LowerCase().Pos("re: ") == 1 ? tmp : "Re: " + tmp);
   }

   AnsiString Hdr = hdr->Text;
   delete hdr;
   return Hdr;
}
//---------------------------------------------------------------------------

/*void __fastcall TMsgManager::MassageReferences()
{
   AnsiString refs = GetReferences(false);
   if (!refs.IsEmpty())
   {
      int idx = DelHeader("References:");
      assert(idx < EOH());
      assert(idx >= 0);
	  headers->Insert(idx, "References: " + refs);
   }
}*/
//---------------------------------------------------------------------------

void __fastcall TMsgManager::CutMessage()
{
   int cut = 0;
   AnsiString line;
   for (int i = body->Count - 1; i > 0; i--)
   {
      line = body->Strings[i];
      if (line == "~~")
	  {
         cut = i;
         break;
      }
   }
   if (cut)
   {
      while (body->Count > cut)
         body->Delete(body->Count -1);
   }
}
//---------------------------------------------------------------------------

int __fastcall TMsgManager::GetHeaderNames(TStrings *names)
{
   names->Clear();

   for (int i = 0; i < Lines->Count; i++)
   {
	  AnsiString name, line = Lines->Strings[i];
	  if (line.Trim().IsEmpty())
	  {
		 if (i < Lines->Count - 1 && Lines->Strings[i + 1] == "##")
			continue;
		 break;
	  }

	  if (line[1] == ' ' || line[1] == '\t' || line == "##")
		 continue;

	  name = line.SubString(1, line.Pos(":"));
	  if (name.IsEmpty())
		 continue;

	  names->Add(name);
   }

   return names->Count;
}
//---------------------------------------------------------------------------

void __fastcall TMsgManager::Insert(int Index, const AnsiString &line)
{
   assert(Index >= 0);
   assert(Index < Lines->Count);
   Lines->Insert(Index, line);
}
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgManager::Get4Paste(const AnsiString &header)
{
   AnsiString hdr = GetHeader(header).Trim();
   if (hdr.IsEmpty())
	  return "";

   int p;
   if (header.LowerCase() == "subject:")
   {
	  while ((p = hdr.Pos("\r")) > 0)
		 hdr.Delete(p, 1);
	  while ((p = hdr.Pos("\n")) > 0)
		 hdr.Delete(p, 1);
	  while ((p = hdr.Pos("\t")) > 0)
		 hdr[p] = ' ';
	  return "Subject: " + hdr;
   }

   if (header.LowerCase() == "references:")
   {
	  while ((p = hdr.Pos("\r")) > 0)
		 hdr.Delete(p, 1);
	  while ((p = hdr.Pos("\n")) > 0)
		 hdr.Delete(p, 1);
	  while ((p = hdr.Pos("\t")) > 0)
		 hdr.Delete(p, 1);
	  while ((p = hdr.Pos("> ")) > 0)
		 hdr.Delete(p + 1, 1);
	  while ((p = hdr.Pos(" <")) > 0)
		 hdr.Delete(p, 1);
	  while ((p = hdr.Pos("><")) > 0)
		 hdr.Insert(" ", p + 1);
	  while ((p = hdr.Pos("<>")) > 0)
		 hdr.Delete(p, 2);
	  return "References: " + hdr.Trim();
   }

   if (header.LowerCase() == "newsgroups:")
   {
	  while ((p = hdr.Pos("\r")) > 0)
		 hdr.Delete(p, 1);
	  while ((p = hdr.Pos("\n")) > 0)
		 hdr.Delete(p, 1);
	  while ((p = hdr.Pos("\t")) > 0)
		 hdr.Delete(p, 1);
	  while ((p = hdr.Pos(" " )) > 0)
		 hdr.Delete(p, 1);
	  return "Newsgroups: " + hdr;
   }

   if (header.LowerCase() == "post:")
   {
	  while ((p = hdr.Pos("\r")) > 0)
		 hdr.Delete(p, 1);
	  while ((p = hdr.Pos("\n")) > 0)
		 hdr.Delete(p, 1);
	  while ((p = hdr.Pos("\t")) > 0)
		 hdr.Delete(p, 1);
	  while ((p = hdr.Pos(" " )) > 0)
		 hdr.Delete(p, 1);
	  return "Newsgroups: " + hdr;
   }

   while ((p = hdr.Pos("\r")) > 0)
	  hdr.Delete(p, 1);
   while ((p = hdr.Pos("\n")) > 0)
	  hdr.Delete(p, 1);
   while ((p = hdr.Pos("\t")) > 0)
	  hdr[p] = ' ';
   return header + " " + hdr;
}
//---------------------------------------------------------------------------

void __fastcall AddString(const AnsiString &String, TStrings *Strings,
   int WordWrapAt)
{
   AnsiString s = String;
   int wwAt = WordWrapAt, wwActual = wwAt - 1;
   if (wwAt > 0 && wwAt < 16)
	  wwAt = 16;

   while (wwAt > 0 && s.Length() > wwAt)
   {
	  for (int i = wwAt; i > 0; i--)
	  {
		 if (s[i]==' ' || s[i]=='-')
		 {
			wwActual = i;
			break;
		 }
	  }
	  Strings->Add(s.SubString(1, wwActual));
	  s = s.SubString(wwActual + 1, 4096);
   }

   Strings->Add(s);
}
//---------------------------------------------------------------------------

void __fastcall TMsgManager::WrapMessage(int AtCol, const AnsiString &QuoteMarks)
{
   if (!AtCol)
	  return;

   AnsiString line, marks = QuoteMarks.TrimRight();
   bool notquoted = true; // don't wrap quotes lines--of course

   TStringList *mail = new TStringList();

   int x = GetFirstBody();
   while (Lines->Count > x)
   {
	  int col=AtCol;
	  line = Lines->Strings[x];
	  if (line != "-- ")
		 line = line.TrimRight();
	  if (notquoted && line.Pos(marks))
		 col = 0;
	  AddString(line, mail, col);
	  Lines->Delete(x);
   }

   Lines->AddStrings(mail);
   delete mail;
}
//---------------------------------------------------------------------------

/*void __fastcall TMsgManager::WrapMessage(int AtCol, const AnsiString &QuoteMarks)
{
   if (!AtCol)
	  return;

   AnsiString line;
   bool notquoted = true; // don't wrap quotes lines--of course

   bool inhead = true;
   TStringList *mail = new TStringList();

   while (body->Count)
   {
      int col=AtCol;
      line = body->Strings[0].TrimRight();
      if (inhead)
      {
         if (line.IsEmpty() && body->Count > 1 && body->Strings[1] != "##")
            inhead = false;
         mail->Add(line);
      }
      else
	  {
         if (line.Pos(QuoteMarks) == 1 && notquoted)
            col = 0;
         AddString(line, mail, col);
	  }

      body->Delete(0);
   }
   body->Text = mail->Text;

   delete mail;
}*/
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgManager::GetTo()
{
   return GetHeader("To:");
}
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgManager::GetFcc()
{
   return GetHeader("Fcc:");
}
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgManager::GetText()
{
   return Lines->Text;
}
//---------------------------------------------------------------------------

bool __fastcall TMsgManager::HasFcc()
{
   return !GetFcc().IsEmpty();
}
//---------------------------------------------------------------------------

//bool __fastcall TMsgManager::IsPGP()
//{
//   return HasHeader("Pgp:") || HasHeader("PgpBtn:");
//}
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgManager::PopHeader()
{
   assert(Lines->Count > 0);
   AnsiString fld, hdr = Lines->Strings[0].Trim();
   if (hdr.IsEmpty() && Lines->Count > 1 && Lines->Strings[1] == "##")
   {
	  Lines->Delete(0);
	  Lines->Delete(0);
	  hdr = Lines->Strings[0];
   }
   if (hdr.IsEmpty())
	  return "";

   // remove any partial headers (orphan folds)
   while (Lines->Count && (Lines->Strings[0].Pos(" ") == 1 ||
	  Lines->Strings[0].Pos("\t") == 1))
		 Lines->Delete(0);

   // anything left...
   assert(Lines->Count > 0);

   // get header
   hdr = Lines->Strings[0];
   Lines->Delete(0);

   // get any folded lines
   while (Lines->Count)
   {
      fld = Lines->Strings[0];
      if ((fld.Pos(" ") != 1 && fld.Pos("\t") != 1))
         break;

      hdr += "\r\n" + fld;
      Lines->Delete(0);
   }

   return hdr;
}
//---------------------------------------------------------------------------

/*void __fastcall TMsgManager::GetHostSignOnFrom(TStrings *headers)
{
   headers->Clear();
   AnsiString hdr;
   if ((hdr = GetHeader("Host:")) != "")
	  headers->Add("Host: " + hdr);
   if ((hdr = GetHeader("Sign-On:")) != "")
      headers->Add("Sign-On: " + hdr);
   headers->Add("From: " + GetHeader("From:"));
}*/
//---------------------------------------------------------------------------

int __fastcall TMsgManager::GetHeaders(const AnsiString &CommaNames,
   TStrings *headers)
{
   headers->Clear();
   TStringList *names = new TStringList();
   names->CommaText = CommaNames;
   for (int i = 0; i < names->Count; i++)
   {
	  AnsiString name, value;
	  name = names->Strings[i].Trim();
	  value = GetHeader(name);
	  if (!value.IsEmpty())
		 headers->Add(name + " " + value);
   }
   delete names;
   return headers->Count;
}
//---------------------------------------------------------------------------

int __fastcall TMsgManager::GetAttachments(TStrings *Names)
{
   int x;
   AnsiString name;
   Names->Clear();
   AnsiString attaches = GetHeader("Attachment:");
   while ((x = attaches.Pos("\n")) > 0)
	  attaches.Delete(x, 1);
   while ((x = attaches.Pos("\r")) > 0)
	  attaches.Delete(x, 1);
   x = -1; while ((name = GetTok(++x, attaches, ';')) != "")
	  Names->Add(name.Trim());
   return Names->Count;
}
//---------------------------------------------------------------------------
#define mimehdrsize  (33 + 2 + 45 + 2 + 34 + 2 + 2)

int __fastcall TMsgManager::GuestimateFinalMessageSize()
{
   TStringList *files = new TStringList();
   GetAttachments(files);

   int size = Lines->Text.Length();
   for (int i = 0; i < files->Count; i++)
   {
      AnsiString filename = files->Strings[i];
      if (FileExists(filename))
      {
         size += GetFileSize(filename);
         size += mimehdrsize + (ExtractFileName(filename).Length() * 2);
      }
   }

   size += 200; // just for good measure
   delete files;
   return size;
}
//---------------------------------------------------------------------------

int __fastcall TMsgManager::GetFirstBody()
{
   return body->GetFirstBody();
}
//---------------------------------------------------------------------------

