/*
*  msgwins.cpp 
*
*  Copyright (C) 2012 Richard Christman 
*
*  This file is part of QuickSilver AAM.
*
*  QuickSilver AAM is free software; you can redistribute it and/or modify 
*  it under the terms of the GNU General Public License as published by the 
*  Free Software Foundation; either version 3 of the License, or (at your 
*  option) any later version.
* 
*  This program is distributed in the hope that it will be useful, but 
*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
*  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
*  for more details.
* 
*  You should have received a copy of the GNU General Public License along 
*  with this program; if not, see <http://www.gnu.org/licenses>.
*
*  Additional permission under GNU GPL version 3 section 7
*
*  If you modify QuickSilver AAM, or any covered work, by linking or
*  combining it with CodeGear VCL (or a modified version of that library),
*  containing parts covered by the terms of CodeGear's license, the licensors
*  of QuickSilver AAM grant you additional permission to convey the
*  resulting work.
*
*/

//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop
#include <clipbrd.hpp>
#include <inifiles.hpp>

#include "msgwins.h"
#include "mainwins.h"
#include "msgmans.h"
#include "global.h"
#include "utility.h"
#include "gnupg.h"
#include "qsadbs.h"
#include "pgpdpassdlgs2.h"
#include "pgpdlgs.h"
#include "gpg4.h"
#include "dsadlgs.h"
#include "messagepipes.h"
#include "cfgmsgdlgs.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TMsgWin *MsgWin;
const AnsiString DecryptVerifying = " Decrypting... ";
//---------------------------------------------------------------------------

void __fastcall ViewMessage(TForm* Owner, const AnsiString &Key, TStrings *Msg)
{
   TMsgWin *msgwin = new TMsgWin(Owner, Key, Msg);
   msgwin->Show();
}
//---------------------------------------------------------------------------

__fastcall TMsgWin::TMsgWin(TComponent* Owner)
	: TForm(Owner)
{
}
//---------------------------------------------------------------------------

void __fastcall FormatMessage(TStrings *msg)
{
   while (msg->Count && msg->Strings[0].Trim() != "")
	  msg->Delete(0);
   msg->Delete(0);
   if (msg->Strings[0].Pos("***") == 1)
   {
	  TStringList *sig = new TStringList();
	  while (msg->Count)
	  {
		 AnsiString line = msg->Strings[0];
		 if (line.Pos("***") == 0)
			break;
		  sig->Add(line);
		  msg->Delete(0);
	  }

	  sig->Add(msg->Strings[0]);
	  msg->Delete(0);
	  int x = 0;
	  while (msg->Strings[x].Trim() != "")
		 x++;
	  x++;
	  for (int i = 0; i < sig->Count; i++)
		 msg->Insert(x + i, sig->Strings[i]);
   }
}
//---------------------------------------------------------------------------

__fastcall TMsgWin::TMsgWin(TComponent* Owner, const AnsiString &Key, TStrings *Msg)
	: TForm(Owner) //(TComponent*)NULL)
{
   assert(Key != "");
   assert(Msg->Count > 0);
   TIniFile *ini = new TIniFile(AppIni());
   LoadWindowPosition(this, "msgwin", ini);
   orgwin = IntToStr(Width)+IntToStr(Height);
   AssignFont(Font, ini->ReadString("settings", "dialogfont", OptsDefDialogFont));
   AssignFont(Viewer->Font, ini->ReadString("settings", "editorfont", OptsDefEditorFont));
   FewerHeaders->OnClick = NULL;
   FewerHeaders->Checked = ini->ReadBool("settings", "fewerheaders", OptsDefFewerHeaders);
   FewerHeaders->OnClick = FewerHeadersClick;
   delete ini;
   objmsg = Msg;
   msgkey = Key;
   GetNextWindowXY(this);
   Caption = GetFrom(Msg) + "  " + GetTok(1, msgkey, DBFSc);
   TStringList *msg = new TStringList();
   msg->AddStrings(objmsg);
   TMsgManager *msgman = new TMsgManager(objmsg);
   msgman->DelHeader("Subject-No:");
   delete msgman;
   if (FewerHeaders->Checked)
	  CompactMessageHeaders(msg);
   Viewer->Clear();
   Viewer->Lines->AddStrings(msg);
   Viewer->SelStart = 0;
   Viewer->Perform(EM_SCROLLCARET, 0, 0);
   CopyMsgBtn->Enabled = !DecryptBtn->Enabled;
   if (NeedsDecryptVerify(objmsg))
   {
	  DecryptBtn->Enabled = NeedsDecryptVerify(objmsg);
	  DecryptTimer->Enabled = true;
   }
   SetEditRect(Viewer);
   delete msg;
}
//---------------------------------------------------------------------------

__fastcall TMsgWin::~TMsgWin()
{
   AnsiString curwin = IntToStr(Width)+IntToStr(Height);
   if (curwin != orgwin)
   {
	  TIniFile *ini = new TIniFile(AppIni());
	  SaveWindowPosition(this, "msgwin", ini);
	  delete ini;
   }
}
//---------------------------------------------------------------------------

void __fastcall TMsgWin::CreateParams( TCreateParams &Params )
{
   TForm::CreateParams( Params );
   Params.ExStyle |= WS_EX_APPWINDOW;
   Params.WndParent = ::GetDesktopWindow();
}
//---------------------------------------------------------------------------

void __fastcall TMsgWin::DecryptTimerTimer(TObject *Sender)
{
   DecryptTimer->Enabled = false;
   DecryptBtn->Click();
}
//---------------------------------------------------------------------------

void __fastcall TMsgWin::FormResize(TObject *Sender)
{
   SetEditRect(Viewer);
}
//---------------------------------------------------------------------------

void __fastcall TMsgWin::ViewerSelectionChange(TObject *Sender)
{
   CopyBtn->Enabled = Viewer->SelLength > 0;
}
//---------------------------------------------------------------------------

void __fastcall TMsgWin::CopyBtnClick(TObject *Sender)
{
   StatusBar->SimpleText = " copying... ";
   Clipboard()->AsText = Viewer->SelText;
   Snooze(300);
   StatusBar->SimpleText = "";
}
//---------------------------------------------------------------------------

void __fastcall TMsgWin::FormClose(TObject *Sender, TCloseAction &Action)
{
   Action = caFree;
}
//---------------------------------------------------------------------------

void __fastcall TMsgWin::Timer1Timer(TObject *Sender)
{
   if (StatusBar->SimpleText != DecryptVerifying)
	  StatusBar->SimpleText = DecryptVerifying;
   else
	  StatusBar->SimpleText = "";
}
//---------------------------------------------------------------------------

void __fastcall TMsgWin::WMFlashDecrypt(TMessage& Msg)
{
   switch (Msg.WParam)
   {
	  case 0:
		 StatusBar->SimpleText = "";
		 Timer1->Enabled = false;
		 break;
	  case 1:
		 Timer1->Enabled = true;
		 StatusBar->SimpleText = DecryptVerifying;
   }
}
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgWin::GetSubject(TStrings *msg)
{
   int c = 0;
   AnsiString sub;
   for (int i = 0; i < msg->Count; i++)
   {
	  AnsiString line = msg->Strings[i];
	  if (line.Trim().IsEmpty())
	  {
		 c++;
		 if (c > 2)
			break;
	  }
	  if (line.Pos("Subject:") == 1)
		 sub = line.SubString(9, line.Length());
   }
   return sub;
}
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgWin::GetFrom(TStrings *msg)
{
   int x = 0;
   AnsiString from;
   for (int i = 0; i < msg->Count; i++)
   {
	  AnsiString line = msg->Strings[i];
	  if (line.Pos("From:") == 1)
	  {
		 from = GetEAddress(line);
		 if (++x == 2)
			break;
	  }
   }
   return from;
}
//---------------------------------------------------------------------------

AnsiString __fastcall TMsgWin::GetNymAndTypeHeader(const AnsiString &MsgKey)
{
   AnsiString keystr, keyid, subno, eaddr, domain, type;
   subno = GetTok(keySubNo, MsgKey, DBFSc);
   keyid = DB->GetSubjectItem(subno, "keyid", "");
   if (keyid.IsEmpty())
   {
	  MsgBox(this,
		 "Problem: there isn't a PGP key for this nym.   \n"
		 "Possibly you deleted it...",
		  MT_INFO, MB_OK);
	  return "keyerror";
   }
   keystr = GPGGetPublicKeyStr(keyid);
   eaddr = GetEAddress(keystr);
   domain = GetDomain(eaddr);
   type = DB->GetNymserverType(domain);
   return "Nym: " + GetTok(0, keystr, ':') + "; keyid=" + keyid + "; type=" + type + ";";
}
//---------------------------------------------------------------------------

void __fastcall TMsgWin::ShowStatus(const AnsiString &msg)
{
   StatusBar->SimpleText = " " +msg;
}
//---------------------------------------------------------------------------

void __fastcall TMsgWin::WMReplyMessage(TMessage& Msg)
{
   //ShowStatus("Copy completed successfully!");
   CopyMsgBtn->Caption = " Reply ";
   CopyMsgBtn->Enabled = true;
   Caption = cap;
}
//---------------------------------------------------------------------------

void __fastcall TMsgWin::CopyMsgBtnClick(TObject *Sender)
{
   CopyMsgBtn->Enabled = false;
   CopyMsgBtn->Caption = "copying";
   cap = Caption;
   Caption = " Copying... ";
   OwnerPerform(WM_REPLYMESSAGE, 0, (LPARAM)this);
   /*assert(Viewer->Lines->Count);
   AnsiString subno = GetTok(keySubNo, msgkey, DBFSc);
   TStringList *lines = new TStringList();
   lines->AddStrings(Viewer->Lines);
   if (lines->IndexOf(BeginPGPMessage) >= 0)
   {
	  if (DB->GetSubjectItem(subno, "ekey", "0") == "1")
	  {
		 DSABox(this, "symmetric",
			"I see this message has symmetric encryption. Since I have the "
			"password(s), I'll decrypt the symmetric layer(s) before copying.",
			"Symmetric Encryption", MB_OK);
		 AnsiString symkeys = DB->GetSubjectItem(subno, "kphrase", "");
		 assert(symkeys != "");

		 if (!DecryptSymmetric(this, lines, symkeys))
		 {
			CopyMsgBtn->Caption = "Reply";
			CopyMsgBtn->Enabled = true;
			Caption = cap;
			delete lines;
			return;
		 }
		 StripSymmetricDecryptionMarkers(lines);
	  }
   }
   ShowStatus("Copying sender's message... ");
   AnsiString nymhdr = GetNymAndTypeHeader(msgkey);
   if (nymhdr.Pos("typeerror") || nymhdr.Pos("keyerror"))
   {
	  CopyMsgBtn->Caption = "Reply";
	  CopyMsgBtn->Enabled = true;
	  Caption = cap;
	  return;
   }
   lines->Insert(0, nymhdr);
   AnsiString subdesc = DB->LoadSubject(subno);
   assert(subdesc != "");
   lines->Insert(1, subdesc);

   HandleMessageTransfer(this, lines, QSCOPYDATA_NT);

   delete lines;
   Snooze(200);*/
   ActiveControl = Viewer;
}
//---------------------------------------------------------------------------

void __fastcall TMsgWin::Exit1Click(TObject *Sender)
{
   Close();
}
//---------------------------------------------------------------------------

void __fastcall TMsgWin::Selectall1Click(TObject *Sender)
{
   Viewer->SelectAll();
}
//---------------------------------------------------------------------------

//void __fastcall TMsgWin::WMNewMessageKey(TMessage &Msg)
//{
//   msgkey = (char*)Msg.LParam;
//}
//---------------------------------------------------------------------------

void __fastcall TMsgWin::MarkMessageRead(AnsiString &msgkey)
{
   OwnerPerform(WM_MARKMESSAGEREAD, (UINT)Handle, (LPARAM)msgkey.c_str());
   //msgkey[bflags1] |= mflRead;
}
//---------------------------------------------------------------------------

void __fastcall TMsgWin::Decrypt1Click(TObject *Sender)
{
   assert(GPGIsPGPMessage(objmsg));
   TStringList *msg = new TStringList();

   if (GPGHasPGPSignatureVerification(msg))
   {
	  DecryptMessage(this, objmsg);
	  msg->AddStrings(objmsg);
	  if (FewerHeaders->Checked)
		 CompactMessageHeaders(msg);
	  Viewer->Clear();
	  Viewer->Lines->BeginUpdate();
	  Viewer->Lines->AddStrings(msg);
	  Viewer->SelStart = 0;
	  Viewer->Perform(EM_SCROLLCARET, 0, 0);
	  Snooze(100);
	  Viewer->Lines->EndUpdate();
	  DecryptBtn->Enabled = NeedsDecryptVerify(msg);
	  CopyMsgBtn->Enabled = !DecryptBtn->Enabled;
	  ActiveControl = Viewer;
	  delete msg;
	  return;
   }

   objmsg->Insert(0, GetTok(keySubNo, msgkey, DBFSc));
   AnsiString hdr = GetNymAndTypeHeader(msgkey);
   if (hdr.Pos("typeerror") || hdr.Pos("keyerror"))
	  return;
   objmsg->Insert(1, hdr);

   if (DecryptNymMessage(this, objmsg))
   {
	  AnsiString subject = GetSubject(objmsg), from = GetFrom(objmsg);
	  msg->AddStrings(objmsg);
	  if (FewerHeaders->Checked)
		 CompactMessageHeaders(msg);
	  Viewer->Clear();
	  Viewer->Lines->BeginUpdate();
	  Viewer->Lines->AddStrings(msg);
	  Viewer->SelStart = 0;
	  Viewer->Perform(EM_SCROLLCARET, 0, 0);
	  Snooze(100);
	  Viewer->Lines->EndUpdate();
	  DecryptBtn->Enabled = NeedsDecryptVerify(objmsg);
	  if (GPGHasPGPSignatureVerification(objmsg))
		 MarkMessageRead(msgkey);
	  CopyMsgBtn->Enabled = !DecryptBtn->Enabled;
	  ActiveControl = Viewer;

	  if (GPGIsPGPMessage(objmsg))
	  {
		 if (!GPGIsPGPSignedMessage(objmsg))
		 {
			Snooze(500);
			MsgBox(this,
			  "Nym message successfuly decrypted!   \n\n"
			  "I see there's another layer of encryption. That layer must     \n"
			  "have been added by the sender.",
			   MT_INFO, MB_OK);
		 }
		 DecryptMessage(this, objmsg);
		 msg->AddStrings(objmsg);
		 if (FewerHeaders->Checked)
			CompactMessageHeaders(msg);
		 Viewer->Clear();
		 Viewer->Lines->BeginUpdate();
		 Viewer->Lines->AddStrings(msg);
		 Viewer->SelStart = 0;
		 Viewer->Perform(EM_SCROLLCARET, 0, 0);
		 Snooze(100);
		 Viewer->Lines->EndUpdate();
		 DecryptBtn->Enabled = NeedsDecryptVerify(objmsg);
		 if (GPGHasPGPSignatureVerification(objmsg))
			MarkMessageRead(msgkey);
		 CopyMsgBtn->Enabled = !DecryptBtn->Enabled;
		 ActiveControl = Viewer;
	  }

	  objmsg->Insert(0, msgkey);
	  objmsg->Insert(1, subject);
	  objmsg->Insert(2, from);
	  OwnerPerform(WM_SETMESSAGE, 0, (LPARAM)objmsg);
	  bool needs = NeedsDecryptVerify(objmsg);
	  DecryptBtn->Enabled =  needs;
	  CopyMsgBtn->Enabled = !needs;
   }
   else
   {
	  //Viewer->Clear();
	  //Viewer->Lines->BeginUpdate();
	  //Viewer->Lines->AddStrings(msg);
	  //Viewer->SelStart = 0;
	  //Viewer->Perform(EM_SCROLLCARET, 0, 0);
	  //Snooze(100);
	  //Viewer->Lines->EndUpdate();
	  DecryptBtn->Enabled = NeedsDecryptVerify(objmsg);
	  if (GPGHasPGPSignatureVerification(objmsg))
		 MarkMessageRead(msgkey);
	  CopyMsgBtn->Enabled = !DecryptBtn->Enabled;
	  ActiveControl = Viewer;
   }
   delete msg;
}
/*void __fastcall TMsgWin::Decrypt1Click(TObject *Sender)
{
   assert(GPGIsPGPMessage(objmsg));
   TStringList *msg = new TStringList();
   msg->AddStrings(objmsg);

   if (GPGHasPGPSignatureVerification(msg))
   {
	  DecryptMessage(this, msg);
	  Viewer->Clear();
	  Viewer->Lines->BeginUpdate();
	  Viewer->Lines->AddStrings(msg);
	  Viewer->SelStart = 0;
	  Viewer->Perform(EM_SCROLLCARET, 0, 0);
	  Snooze(100);
	  Viewer->Lines->EndUpdate();
	  DecryptBtn->Enabled = NeedsDecryptVerify(msg);
	  CopyMsgBtn->Enabled = !DecryptBtn->Enabled;
	  ActiveControl = Viewer;
	  return;
   }

   msg->Insert(0, GetTok(keySubNo, msgkey, DBFSc));
   AnsiString hdr = GetNymAndTypeHeader(msgkey);
   if (hdr.Pos("typeerror") || hdr.Pos("keyerror"))
	  return;
   msg->Insert(1, hdr);

   if (DecryptNymMessage(this, msg))
   {
	  AnsiString subject = GetSubject(msg), from = GetFrom(msg);
	  Viewer->Clear();
	  Viewer->Lines->BeginUpdate();
	  Viewer->Lines->AddStrings(msg);
	  Viewer->SelStart = 0;
	  Viewer->Perform(EM_SCROLLCARET, 0, 0);
	  Snooze(100);
	  Viewer->Lines->EndUpdate();
	  DecryptBtn->Enabled = NeedsDecryptVerify(msg);
	  if (GPGHasPGPSignatureVerification(msg))
		 MarkMessageRead(msgkey);
	  CopyMsgBtn->Enabled = !DecryptBtn->Enabled;
	  ActiveControl = Viewer;

	  if (GPGIsPGPMessage(msg))
	  {
		 if (!GPGIsPGPSignedMessage(msg))
		 {
			Snooze(500);
			MsgBox(this,
			  "Nym message successfuly decrypted!   \n\n"
			  "I see there's another layer of encryption. That layer must     \n"
			  "have been added by the sender.",
			   MT_INFO, MB_OK);
		 }
		 DecryptMessage(this, msg);
		 Viewer->Clear();
		 Viewer->Lines->BeginUpdate();
		 Viewer->Lines->AddStrings(msg);
		 Viewer->SelStart = 0;
		 Viewer->Perform(EM_SCROLLCARET, 0, 0);
		 Snooze(100);
		 Viewer->Lines->EndUpdate();
		 DecryptBtn->Enabled = NeedsDecryptVerify(msg);
		 if (GPGHasPGPSignatureVerification(msg))
			MarkMessageRead(msgkey);
		 CopyMsgBtn->Enabled = !DecryptBtn->Enabled;
		 ActiveControl = Viewer;
	  }

	  msg->Insert(0, msgkey);
	  msg->Insert(1, subject);
	  msg->Insert(2, from);
	  OwnerPerform(WM_SETMESSAGE, 0, (LPARAM)msg);
	  bool needs = NeedsDecryptVerify(msg);
	  //if (!needs)
	  //{
	  //	 OwnerPerform(WM_MARKMESSAGEREAD, 0, (LPARAM)msgkey.c_str());
	  //	 msgkey[bflags1] |= mflRead;
	  //}
	  DecryptBtn->Enabled =  needs;
	  CopyMsgBtn->Enabled = !needs;
   }
   else
   {
	  Viewer->Clear();
	  Viewer->Lines->BeginUpdate();
	  Viewer->Lines->AddStrings(msg);
	  Viewer->SelStart = 0;
	  Viewer->Perform(EM_SCROLLCARET, 0, 0);
	  Snooze(100);
	  Viewer->Lines->EndUpdate();
	  DecryptBtn->Enabled = NeedsDecryptVerify(msg);
	  if (GPGHasPGPSignatureVerification(msg))
		 MarkMessageRead(msgkey);
	  CopyMsgBtn->Enabled = !DecryptBtn->Enabled;
	  ActiveControl = Viewer;
   }
}*/
//---------------------------------------------------------------------------

void __fastcall TMsgWin::DeleteBtnClick(TObject *Sender)
{
   if (MsgBox(this, "Delete this message?    \n", MT_CONFIRM, MB_OKCANCEL) == IDCANCEL)
	  return;

   OwnerPerform(WM_DELMESSAGE, 0, (LPARAM)msgkey.c_str());
   Close();
}
//---------------------------------------------------------------------------

int __fastcall TMsgWin::OwnerPerform(UINT msg, WPARAM wParam, LPARAM lParam)
{
   return ((TForm*)Owner)->Perform(msg, wParam, lParam);
}
//---------------------------------------------------------------------------

void __fastcall TMsgWin::Timer2Timer(TObject *Sender)
{
   Timer2->Enabled = false;
   Decrypt1->Click();
}
//---------------------------------------------------------------------------

bool __fastcall TMsgWin::NeedsDecryptVerify(TStrings *msg)
{
   return msg->IndexOf(BeginPGPMessage) >= 0 ||
	  msg->IndexOf(BeginPGPSignedMessage) >= 0;
}
//---------------------------------------------------------------------------

void __fastcall TMsgWin::Edit1Click(TObject *Sender)
{
	Copytoclipboard1->Enabled = !DecryptBtn->Enabled;
}
//---------------------------------------------------------------------------

void __fastcall TMsgWin::PopupMenu1Popup(TObject *Sender)
{
   Copymessagetoclipnoard1->Enabled = !DecryptBtn->Enabled;
}
//---------------------------------------------------------------------------

void __fastcall TMsgWin::FewerHeadersClick(TObject *Sender)
{
   TStringList *msg = new TStringList();
   msg->AddStrings(objmsg);
   Viewer->Lines->BeginUpdate();
   Viewer->Lines->Clear();
   if (FewerHeaders->Checked)
	  CompactMessageHeaders(msg);
   Viewer->Lines->AddStrings(msg);
   Snooze(10);
   Viewer->Lines->EndUpdate();
   delete msg;
   TIniFile *ini = new TIniFile(AppIni());
   ini->WriteBool("settings", "fewerheaders", FewerHeaders->Checked);
   delete ini;
   OwnerPerform(WM_FEWERHEADERS, (WORD)FewerHeaders->Checked, 0);
}
//---------------------------------------------------------------------------

